/** @file
 * @brief Front-ends (C and XML) for PadicoConnector
 */

#include <Padico/Puk.h>
#include <Padico/Module.h>
#include <Padico/NetSelector.h>
#include <Padico/PadicoControl.h>

#include "Connector.h"

PADICO_MODULE_DECLARE(Connector);
PADICO_MODULE_INIT(padico_connector_init);
PADICO_MODULE_FINALIZE(padico_connector_finalize);


/* XML parsing handlers */
static void padico_connector_new_handler(puk_parse_entity_t e);
static void padico_connector_connect_handler(puk_parse_entity_t e);

/* ********************************************************* */

padico_rc_t padico_control_connector_new(const void**addr, size_t*addr_len)
{
  padico_na_host_t host = NULL;
  const puk_component_t component = padico_ns_host_selector(host, NULL, puk_iface_PadicoConnector(), 0, 0);
  if(component == NULL)
    {
      return padico_rc_error("Cannot create connector: no component found\n");
    }
  puk_instance_t instance = puk_component_instantiate(component);
  struct puk_receptacle_PadicoConnector_s r;
  puk_instance_indirect_PadicoConnector(instance, NULL, &r);
  if(r.driver == NULL)
    {
      return padico_rc_error("Cannot create connector: component=%s has no 'PadicoConnnector' facet.", component->name);
    }
  int rc = (*r.driver->create)(r._status, addr, addr_len);
  if(rc)
    {
      return padico_rc_error("Error %d while invoking ConnectorNew on component %s.", rc, component->name);
    }
  return padico_rc_ok();
}

padico_rc_t padico_control_connector_connect(const void*addr, size_t addr_len)
{
  padico_na_host_t host = NULL;
  const puk_component_t component = padico_ns_host_selector(host, NULL, puk_iface_PadicoConnector(), 0 /* port */, 0 /* IPPROTO_TCP */);
  if(component == NULL)
    {
      return padico_rc_error("Cannot create connector: no component found\n");
    }
  puk_instance_t instance = puk_component_instantiate(component);
  struct puk_receptacle_PadicoConnector_s r;
  puk_instance_indirect_PadicoConnector(instance, NULL, &r);
  if(r.driver == NULL)
    {
      return padico_rc_error("Cannot create connector: component=%s has no 'PadicoConnnector' facet.", component->name);
    }
  padico_req_t req = padico_tm_req_new(NULL, NULL);
  int rc = (*r.driver->connect)(r._status, addr, addr_len, req);
  if(rc)
    {
      return padico_rc_error("Error %d while invoking ConnectorConnect on component %s.", rc, component->name);
    }
  padico_rc_t req_rc = padico_tm_req_wait(req);
  return req_rc;
}

/* ********************************************************* */

static void padico_connector_new_handler(puk_parse_entity_t e)
{
  size_t addr_len = 0;
  const void*addr = NULL;
  padico_rc_t rc = padico_control_connector_new(&addr, &addr_len);
  if(!padico_rc_iserror(rc))
    {
      const char*bytes = addr;
      int size = addr_len;
      char*text = puk_hex_encode(bytes, &size, NULL);
      padico_out(20, "encoded=*%s* (len=%d)\n", text, size);
      puk_parse_set_rc(e, padico_rc_msg("%s", text));
      padico_free(text);
    }
  else
    {
      puk_parse_set_rc(e, rc);
    }
}

static void padico_connector_connect_handler(puk_parse_entity_t e)
{
  const char*addr = puk_parse_getattr(e, "address");
  int len = strlen(addr);
  assert(addr != NULL && len < 1024 && len > 0);
  char*buffer = puk_hex_decode(addr, &len, NULL);
  assert(buffer != NULL);
  padico_control_connector_connect(buffer, len);
  padico_free(buffer);
}


/* ********************************************************* */


static int padico_connector_init(void)
{
  puk_iface_register("PadicoConnector");

  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "Connector:New",
    .start_handler  = &padico_connector_new_handler,
    .required_level = PUK_TRUST_OUTSIDE
  });
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "Connector:Connect",
    .start_handler  = &padico_connector_connect_handler,
    .required_level = PUK_TRUST_OUTSIDE
  });

  return 0;
}

static void padico_connector_finalize(void)
{
  puk_xml_delete_action("Connector:New");
  puk_xml_delete_action("Connector:Connect");
}

