
#include <Padico/Puk.h>
#include <Padico/PadicoTM.h>
#include <Padico/Module.h>

PADICO_MODULE_DECLARE(ControlBootstrapPMI2);
PADICO_MODULE_INIT(control_bootstrap_pmi2_init);
PADICO_MODULE_FINALIZE(control_bootstrap_pmi2_finalize);

#ifdef HAVE_PMI2

#include <Padico/NetSelector.h>
#include <Padico/Topology.h>
#include <Padico/PadicoControl.h>
#include <Padico/PMI2.h>

PUK_VECT_TYPE(puk_instance, puk_instance_t);

static struct
{
  puk_component_t component;           /**< component used for listening */
  puk_instance_vect_t instances;       /**< listening bootstrap instances */ // only master have multiple instances
  struct
  {
    marcel_mutex_t lock;
    marcel_cond_t ready;
    padico_na_node_vect_t nodes;
    int nodes_ready;
  } sync;
} control_bootstrap =
  { .component = NULL };

static puk_instance_t padico_bootstrap_indirect(padico_na_node_t remote_node, int rank,
                                                struct puk_receptacle_PadicoConnector_s*r)
{
  /* get remote node topology befor using NetSelector */
  padico_string_t s_key = padico_string_new();
  padico_string_printf(s_key, "bootstrap_topology_%d", rank);
  char*remote_topo = NULL;
  padico_pmi2_kvs_lookup_multipart(padico_string_get(s_key), &remote_topo);
  padico_control_event_disable();
  struct puk_parse_entity_s e = puk_xml_parse_buffer(remote_topo, strlen(remote_topo), PUK_TRUST_CLUSTER);
  padico_rc_t rc2 = puk_parse_get_rc(&e);
  if(padico_rc_iserror(rc2))
    {
      padico_rc_show(rc2);
    }
  padico_control_event_enable();
  padico_free(remote_topo);

  control_bootstrap.component =
    padico_ns_serial_selector(remote_node, NULL, puk_iface_PadicoConnector());
  if(control_bootstrap.component == NULL)
    {
      padico_fatal("ControlBootstrap: no PadicoConnector assembly found.\n");
    }
  if(puk_component_get_driver_PadicoConnector(control_bootstrap.component, NULL) == NULL &&
     puk_component_get_driver_PadicoControl(control_bootstrap.component, NULL) == NULL)
    {
      padico_fatal("ControlBootstrap: bootstrap component do not provide required interface\
		   (PadicoConnector, PadicoControl).\n");
    }
  puk_instance_t instance = puk_component_instantiate(control_bootstrap.component);
  puk_instance_indirect_PadicoConnector(instance, NULL, r);
  padico_out(50, "init bootstrap served by assembly %s\n", instance->component->name);
  return instance;
}

static void control_bootstrap_pmi2_ready_handler(puk_parse_entity_t e)
{
  marcel_mutex_lock(&control_bootstrap.sync.lock);
  control_bootstrap.sync.nodes_ready++;
  marcel_cond_signal(&control_bootstrap.sync.ready);
  marcel_mutex_unlock(&control_bootstrap.sync.lock);
}

static int control_bootstrap_pmi2_init(void)
{
  const char*c_rdv_host = padico_getattr("PADICO_BOOTSTRAP_HOST");
  control_bootstrap.instances = puk_instance_vect_new();
  marcel_mutex_init(&control_bootstrap.sync.lock, NULL);
  marcel_cond_init(&control_bootstrap.sync.ready, NULL);
  control_bootstrap.sync.nodes = padico_na_node_vect_new();
  control_bootstrap.sync.nodes_ready = 0;
  puk_xml_add_action((struct puk_tag_action_s){
      	     .xml_tag        = "ControlBootstrap:ready",
      	     .start_handler  = NULL,
      	     .end_handler    = &control_bootstrap_pmi2_ready_handler,
      	     .required_level = PUK_TRUST_CONTROL
      	     });

  const int rank = atoi(padico_getattr("PADICO_BOOT_RANK")?:"0");
  const int size = atoi(padico_getattr("PADICO_BOOT_SIZE")?:"1");
  /* Publish topology & fence */
  padico_string_t s_key = padico_string_new();
  padico_string_printf(s_key, "bootstrap_topology_%d", rank);
  padico_string_t s_localhost = padico_na_host_serialize(padico_na_getlocalhost());
  padico_pmi2_kvs_publish_multipart(padico_string_get(s_key), padico_string_get(s_localhost));
  padico_pmi2_kvs_fence();
  padico_string_delete(s_key);
  padico_string_delete(s_localhost);

  if(rank == 0)
    {
      for(int i = 1; i < size; i++)
        {
          // Instantiate control-minidriver
          struct puk_receptacle_PadicoConnector_s r;
          puk_instance_t instance = padico_bootstrap_indirect(NULL, i, &r);
          puk_instance_vect_push_back(control_bootstrap.instances, instance);
        }

      // Init control-minidriver
      struct puk_receptacle_PadicoConnector_s r;
      puk_instance_indirect_PadicoConnector(*puk_instance_vect_begin(control_bootstrap.instances), NULL, &r);
      const void*local_addr;
      size_t addr_size;
      (*r.driver->create)(r._status, &local_addr, &addr_size);

      puk_instance_vect_itor_t itor;
      int i = 1;
      puk_vect_foreach(itor, puk_instance, control_bootstrap.instances)
        {
          struct puk_receptacle_PadicoConnector_s r;
          puk_instance_indirect_PadicoConnector(*itor, NULL, &r);
          //Connect
          (*r.driver->connect)(r._status, local_addr, addr_size, NULL);
          i++;
        }
    }
  else
    {

      // Instantiate control-minidriver
      struct puk_receptacle_PadicoConnector_s r;
      const padico_na_host_t host = padico_na_host_resolvebyname(c_rdv_host);
      puk_instance_t instance = padico_bootstrap_indirect(host, 0, &r);
      puk_instance_vect_push_back(control_bootstrap.instances, instance);

      const void*local_addr;
      size_t addr_size;
      // Init control-minidriver
      (*r.driver->create)(r._status, &local_addr, &addr_size);

      //Connect
      (*r.driver->connect)(r._status, local_addr, addr_size, NULL);
    }

  int wait_nodes = size - 1;
  marcel_mutex_lock(&control_bootstrap.sync.lock);
  padico_topo_network_t session = padico_na_session_getnetwork(padico_na_node_getsession(padico_na_getlocalnode()));
  while(padico_na_network_size(session) < size)
    {
      //marcel_cond_wait(&control_bootstrap.sync.ready, &control_bootstrap.sync.lock); //TODO
    }
  if(rank == 0)
    {
      /* wait all nodes to acknowledge they are ready */
      padico_topo_network_t session = padico_na_session_getnetwork(padico_na_node_getsession(padico_na_getlocalnode()));
      while((control_bootstrap.sync.nodes_ready < wait_nodes) ||
            padico_na_network_size(session) < size)
        {
          marcel_cond_wait(&control_bootstrap.sync.ready, &control_bootstrap.sync.lock);
        }
    }
  else
    {
      padico_na_node_t session_0 =
        padico_na_session_getnodebyrank(padico_na_node_getsession(padico_na_getlocalnode()), 0);
      padico_string_t msg = padico_string_new();
      padico_string_t self_node = padico_na_node_serialize(padico_na_getlocalnode());
      padico_string_printf(msg, "<ControlBootstrap:ready>%s</ControlBootstrap:ready>\n", padico_string_get(self_node));
      padico_string_delete(self_node);
      padico_control_send_oneway(session_0, padico_string_get(msg));
      padico_string_delete(msg);
    }

    marcel_mutex_unlock(&control_bootstrap.sync.lock);
    padico_print("ControlBootstrapPmi2 %d: IB control channel is ready.\n", rank);
    return 0;

}

static void control_bootstrap_pmi2_finalize(void)
{
  while(!puk_instance_vect_empty(control_bootstrap.instances))
    {
      puk_instance_t inst = puk_instance_vect_pop_back(control_bootstrap.instances);
      puk_instance_destroy(inst);
    }
  puk_instance_vect_delete(control_bootstrap.instances);
  control_bootstrap.component = NULL;
  padico_na_node_vect_delete(control_bootstrap.sync.nodes);
  puk_xml_delete_action("ControlBootstrap:ready");
}

#else /* HAVE_PMI2 */

static int control_bootstrap_pmi2_init(void)
{
  padico_fatal("ControlBootstrapPmi2: PadicoTM was compiled without PMI2 support.\n");
}

static void control_bootstrap_pmi2_finalize(void)
{
  /* do nothing */
}

#endif /* HAVE_PMI2 */
