/** @file
 * @brief NetSelector with lazy name-resolution and lazy component loading.
 * @ingroup NetSelector
 */


#include "NetSelector.h"

#include <Padico/Module.h>
#include <Padico/Puk.h>

PADICO_MODULE_DECLARE(NetSelector);
PADICO_MODULE_INIT(ns_netselector_init);
PADICO_MODULE_FINALIZE(ns_netselector_finalize);

/* ********************************************************* */

const static char default_profile[] = "default";

typedef const struct padico_netselector_driver_s*padico_netselector_driver_t;
/** describes an entry in the selectors hashtable
 */
struct netselector_entry_s
{
  const struct padico_netselector_driver_s*selector; /**< the driver for the given selector */
  const char*name; /**< the name for this entry- 
		    * @note this field may be different from selector->name (because of aliases)
		    */
};
PUK_VECT_TYPE(netselector_entry, struct netselector_entry_s);

static struct
{
  puk_hashtable_t table; /**< table of know selectors [ key: profile; data: netselector_entry_vect_t ] */
  marcel_mutex_t lock;   /**< lock for accessing the hashtable */
} nsdb = { .table = NULL };

static netselector_entry_vect_t ns_selector_lookup(const char*profile);
static void ns_selector_insert(const struct padico_netselector_driver_s*selector);


/* ********************************************************* */

static int ns_netselector_init(void)
{
  puk_iface_register("NetSelector");
  nsdb.table = puk_hashtable_new_string();
  marcel_mutex_init(&nsdb.lock, NULL);
  return 0;
}

static void ns_netselector_entry_destructor(void*_key, void*_data)
{
  netselector_entry_vect_t v = _data;
  netselector_entry_vect_delete(v);
}
static void ns_netselector_finalize(void)
{
  puk_hashtable_delete(nsdb.table, &ns_netselector_entry_destructor);
}

/* ********************************************************* */

void padico_ns_selector_register(const struct padico_netselector_driver_s*selector)
{
  padico_out(6, "adding selector=%s for profile=%s\n", selector->name, selector->profile?:"*");
  puk_component_declare(selector->name,
			puk_component_provides("NetSelector", "selector", selector));
  ns_selector_insert(selector);
}

puk_component_t padico_ns_serial_selector(padico_na_node_t peer, const char*_profile, 
					  puk_iface_t iface)
{
  const char*profile = _profile?:default_profile;
  const netselector_entry_vect_t v = ns_selector_lookup(profile);
  puk_component_t assembly = NULL;
  netselector_entry_vect_itor_t i;
  padico_out(20, "remote peer node=%s- computing assembly for iface=%s; profile=%s.\n", 
	     peer?padico_na_node_getname(peer):"*", iface->name, profile);
  assert(peer != NULL);
  puk_vect_foreach(i, netselector_entry, v)
    {
      const padico_netselector_driver_t selector = i->selector;
      if(selector && selector->serial)
	assembly = (*selector->serial)(padico_na_getlocalnode(), peer, profile, iface);
      if(assembly)
	break;
    }
  padico_out(10, "profile = %s; assembly = %s\n", profile, assembly->name);
  return assembly;
}

puk_component_t padico_ns_host_selector(padico_na_host_t host, const char*_profile, 
					puk_iface_t iface, uint16_t port, int protocol)
{
  const char*profile = _profile?:default_profile;
  const netselector_entry_vect_t v = ns_selector_lookup(profile);
  puk_component_t assembly = NULL;
  netselector_entry_vect_itor_t i;
  padico_out(20, "remote peer host=%s- computing assembly for iface=%s; profile=%s.\n", 
	     host?padico_na_host_getname(host):"*", iface->name, profile);
  puk_vect_foreach(i, netselector_entry, v)
    {
      const padico_netselector_driver_t selector = i->selector;
      if(selector && selector->host)
	{
	  padico_out(20, "trying selector *%s*\n", selector->name);
	  assembly = (*selector->host)(host, profile, iface, port, protocol);
	  if(assembly)
	    {
	      padico_out(20, "got answer from selector *%s*\n", selector->name);
	      break;
	    }
	}
    }
  padico_out(10, "profile = %s; assembly = %s\n", profile, assembly->name);
  return assembly;
}

/* ********************************************************* */

static void ns_selector_insert(const struct padico_netselector_driver_s*selector)
{
  int already = 0;
  if(nsdb.table)
    {
      netselector_entry_vect_t v = NULL;
      marcel_mutex_lock(&nsdb.lock);
      v = puk_hashtable_lookup(nsdb.table, selector->profile);
      if(!v)
	{
	  v = netselector_entry_vect_new();
	  puk_hashtable_insert(nsdb.table, selector->profile, v);
	}
      else
	{
	  netselector_entry_vect_itor_t i;
	  puk_vect_foreach(i, netselector_entry, v)
	    {
	      if(strcmp(i->name, selector->name) == 0)
		{
		  i->selector = selector;
		  already = 1;
		}
	    }

	}
      if(!already)
	{
	  const struct netselector_entry_s e = 
	    {
	      .name = selector->name,
	      .selector = selector
	    };
	  netselector_entry_vect_push_back(v, e);
	}
      marcel_mutex_unlock(&nsdb.lock);
    }
}

/** returns the known selector entries (vect) for a given profile
 */
static netselector_entry_vect_t ns_selector_lookup(const char*profile)
{
  netselector_entry_vect_t v = NULL;
  marcel_mutex_lock(&nsdb.lock);
  v = puk_hashtable_lookup(nsdb.table, profile);
  marcel_mutex_unlock(&nsdb.lock);
  if(!v)
    {
      padico_string_t s = padico_string_new();
      const char*selector_name = NULL;
      int i = 0;
      padico_out(6, "resolving selectors for profile *%s*\n", profile);
      do
      {
	padico_string_printf(s, "PADICO_NETSELECTOR_%s_%d", profile, i);
	selector_name = padico_getattr(padico_string_get(s));
	if(selector_name)
	  {
	    padico_print("NetSelector: resolving- %s:%d... %s\n", profile, i, selector_name);
	    const puk_component_t component = puk_component_resolve(selector_name);
	    if(!component)
	      {
		padico_warning("NetSelector: invalid component, will not use selector *%s*.\n", selector_name);
	      }
	  }
	i++;
      }
      while(selector_name);
      padico_string_delete(s);
      marcel_mutex_lock(&nsdb.lock);
      v = puk_hashtable_lookup(nsdb.table, profile);
      marcel_mutex_unlock(&nsdb.lock);
    }
  if((v == NULL) || netselector_entry_vect_empty(v))
    {
      padico_fatal("NetSelector: no selector for profile *%s*.\n", profile);
    }
  return v;
}

