#include <Padico/Module.h>
#include <Padico/PadicoControl.h>
#include <Padico/Topology.h>

#include <slurm/pmi2.h>

PADICO_MODULE_DECLARE(PMI2);
PADICO_MODULE_INIT(padico_pmi2_init);
PADICO_MODULE_FINALIZE(padico_pmi2_finalize);

static struct
{
  int rank;
  int size;
} padico_pmi2 = { .rank = -1, .size = -1 };

/* ********************************************************* */

static int padico_pmi2_init(void)
{
  if(PMI2_Initialized())
    return 0;
  int spawned, appnum;
  int found;
  char buf[PMI2_MAX_VALLEN];
  int rc = PMI2_Init(&spawned, &padico_pmi2.size, &padico_pmi2.rank, &appnum);
  if(rc != PMI2_SUCCESS)
    {
      padico_fatal("PMI2: pmi2 initialization failed.\n");
    }
  if(padico_pmi2.size < 0 || padico_pmi2.rank < 0)
    {
      PMI2_Finalize();
      padico_fatal("PMI2: pmi2 was initialized but negative values for job and/or rank was returned.\n");
    }
  if(getenv("PADICO_BOOT_ID") != NULL)
    {
      /* check consistency between pmi2 numbering and padico-launch numbering */
      if(padico_pmi2.rank != padico_na_rank())
	{
	  PMI2_Finalize();
	  padico_fatal("PMI2: inconsistency detected, PMI_RANK=%d, PMI2_Init rank=%d.\n",
		       padico_na_rank(), padico_pmi2.rank);
	}
    }
  rc = PMI2_Info_GetJobAttr("hasNameServ", buf, PMI2_MAX_VALLEN, &found);
  if(PMI2_SUCCESS != rc)
    {
      PMI2_Finalize();
      padico_fatal("PMI2: PMI2_Info_GetJobAttr failed.");
    }
  if(!found)
    {
      PMI2_Finalize();
      padico_fatal("PMI2: pmi2 environment doesn't support the name service operations.");
    }
  return 0;
}

static void padico_pmi2_finalize(void)
{
  if(PMI2_Initialized())
    PMI2_Finalize();
}

int padico_pmi2_getrank(void)
{
  return padico_pmi2.rank;
}

int padico_pmi2_getsize(void)
{
  return padico_pmi2.size;
}

void padico_pmi2_abort(const char*msg)
{
  PMI2_Abort(1, msg);
}


/* ********************************************************* */

int padico_pmi2_kvs_publish(const char*key, const char*value)
{
  int rc = PMI2_KVS_Put(key, value);
  if(rc != PMI2_SUCCESS)
    {
      padico_fatal("PMI2: PMI2_KVS_Put failed rc=%d.", rc);
    }
  return 0;
}

int padico_pmi2_kvs_lookup(const char*key, char*value, const int value_size)
{
  int len = value_size;
  int rc = PMI2_KVS_Get(/* jobid */ NULL, PMI2_ID_NULL, key, value, value_size, &len);
  if(rc != PMI2_SUCCESS)
    padico_fatal("PMI2: PMI2_KVS_Get() failed key = %s; rc = %d (%s).\n",
		 key, rc, strerror(-rc));
  return 0;
}

int padico_pmi2_kvs_publish_multipart(const char*key, const char*value)
{
  char size_c[PMI2_MAX_VALLEN];
  int size = strlen(value);
  int nb_msg = (size / PMI2_MAX_VALLEN) + (size % PMI2_MAX_VALLEN ? 1 : 0);

  sprintf(size_c , "%d", size);
  padico_pmi2_kvs_publish(key, size_c);

  for(int i = 0; i < nb_msg; i++)
    {
      char part_key[PMI2_MAX_KEYLEN];
      char part_val[PMI2_MAX_VALLEN+1];
      int part_size = (i < nb_msg-1) ? PMI2_MAX_VALLEN : size - (i*PMI2_MAX_VALLEN);
      int offset = i ? (i * PMI2_MAX_VALLEN)  : 0;

      sprintf(part_key, "%s-part-%d", key, i);
      memcpy(part_val, value + offset, part_size);
      part_val[part_size] = 0;
      padico_pmi2_kvs_publish(part_key, part_val);
    }
  return 0;
}

int padico_pmi2_kvs_lookup_multipart(const char*key, char**value)
{
  char size_c[PMI2_MAX_VALLEN];
  int nb_msg;
  int size;

  padico_pmi2_kvs_lookup(key, size_c, PMI2_MAX_VALLEN);
  size = atoi(size_c);
  nb_msg = (size / PMI2_MAX_VALLEN) + (size % PMI2_MAX_VALLEN ? 1 : 0);
  assert(*value == NULL);
  *value = (char*)padico_malloc(size + 1);
  **value = 0;
  for(int i = 0; i < nb_msg; i++)
    {
      char part_key[PMI2_MAX_KEYLEN];
      char part_val[PMI2_MAX_VALLEN+1];
      int part_size = (i < nb_msg-1) ? PMI2_MAX_VALLEN : size - (i*PMI2_MAX_VALLEN);

      sprintf(part_key, "%s-part-%d", key, i);
      padico_pmi2_kvs_lookup(part_key, part_val, part_size+1);
      strcat(*value, part_val);
    }
  return 0;
}

int padico_pmi2_kvs_fence(void)
{
  int rc;
  if(PMI2_SUCCESS != (rc = PMI2_KVS_Fence()))
    {
      padico_fatal("PMI2: PMI2_KVS_Fence failed (rc=%d).", rc);
    }
  return 0;
}
