/** @file
 * @brief ZIP Packet Filter component
 */

#include <Padico/Module.h>
#include <Padico/PadicoTM.h>
#include "PacketFilter.h"
#include "PadicoTM_config.h"

#ifdef HAVE_ZLIB_H
#include <zlib.h>
#endif

PADICO_MODULE_DECLARE(PacketFilter_ZIP);
PADICO_MODULE_INIT(zpf_init);

/* ********************************************************* */
/* *** ZIP Packet Filter *********************************** */
/* ********************************************************* */

/** @defgroup PacketFilter_ZIP component: PacketFilter-ZIP- a packet filter that performs ZIP compression
 * @ingroup PadicoComponent
 */

static void*zpf_instantiate(puk_instance_t assembly, puk_context_t context);
static void zpf_destroy(void*state);
static void zpf_encode(void*, const void*, size_t, void**, size_t*);
static void zpf_decode(void*, const void*, size_t, void*, size_t*);

/** instanciation facet for PacketFilter-ZIP
 * @ingroup PacketFilter_ZIP
 */
const static struct puk_component_driver_s zip_component_driver =
  {
    .instantiate = &zpf_instantiate,
    .destroy     = &zpf_destroy
  };

/** 'PacketFilter' facet for PacketFilter-ZIP
 * @ingroup PacketFilter_ZIP
 */
const static struct packetfilter_driver_s zpf_driver = 
  {
    .pf_encode = &zpf_encode,
    .pf_decode = &zpf_decode
  };
/** instance status for PacketFilter-ZIP
 * @ingroup PacketFiler_ZIP
 */
struct zip_state_s
{
  int level;
  z_stream in_stream;
  z_stream out_stream;
};

static void* zpf_instantiate(puk_instance_t assembly, puk_context_t context)
{
  struct zip_state_s*state = padico_malloc(sizeof(struct zip_state_s));
  const char*s_level = puk_context_getattr(context, "level");
  assert(s_level);
  int level = atoi(s_level);
  state->level = level;
  return state;
}

static void zpf_destroy(void*state)
{
  assert(state != NULL);
  padico_free(state);
}


/* 'out' allocated by encode function
 */
static void zpf_encode(void*w,
		       const void*in, size_t in_size,
		       void**out, size_t* out_size)
{
  struct zip_state_s*state = w;
  unsigned long s = in_size + in_size / 16 + 64;
  void*p = padico_malloc(s);
  int rc = compress2(p, &s, in, in_size, state->level);
  if(rc != Z_OK)
    {
      /* compression failed: don't compress */
      padico_fatal("ZIP: compression failed.\n");
      memcpy(p, in, in_size);
      s = in_size;
    }
  *out = p;
  *out_size = s;
}

/* 'out' allocated by caller
*/
static void zpf_decode(void*state,
		       const void*in, size_t in_size,
		       void*out, size_t* out_size)
{
  unsigned long s = *out_size;
  int rc = uncompress(out, &s, in, in_size);
  if(rc != Z_OK)
    {
      padico_fatal("ZIP: CRC error while uncompressing- in_size=%lu; out_size=%lu; rc=%d\n",
		   (unsigned long)in_size, (unsigned long)*out_size, rc);
    }
  *out_size = s;
}


static int zpf_init(void)
{
  padico_print("PacketFilter-ZIP: using zlib version %s\n", zlibVersion());

  puk_component_declare("PacketFilter-ZIP",
			puk_component_provides("PadicoComponent", "component", &zip_component_driver),
			puk_component_provides("PacketFilter", "pf", &zpf_driver),
			puk_component_attr("level", "3"));
  return 0;
}
