/** @file
 * @brief SysIO: System I/O manager
 */

#ifndef PADICO_SYSIO_H
#define PADICO_SYSIO_H

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/uio.h>

/** @defgroup SysIO API: SysIO interface
 */

/** @addtogroup SysIO
 * @{
 */

/** a NetAccess I/O handle */
typedef struct padico_sysio_s* padico_sysio_t;

/** for backwards compatibility with legacy codes */
#define padico_io_t padico_sysio_t

/** @brief a NetAccess SysIO callback
 * @arg fd the file descriptor on which the event happened
 * @arg key a key given at registration of the callback
 * @return 1 to keep the callback registered, 0 to cancel
 */
typedef int (*padico_sysio_callback_t)(int fd, void*key);

#define PADICO_IO_EVENT_READ   0x01 /**< the fildes is readable */
#define PADICO_IO_EVENT_WRITE  0x02 /**< the fildes is writable */
#define PADICO_IO_EVENT_ERROR  0x04 /**< the fildes has an error */

/** @brief Registers a file descriptor for polling
 * @param fd File descriptor (should already exist)
 * @param what The operation that triggers callback -- several bits may be OR-ed
 * @param callback The callback function that will be called uppon I/O event
 * @param key an argument that will be given to the callback
 * @return an I/O handle for future reference
 * @note after registration, the I/O handle is not activated
 */
padico_io_t padico_io_register(int fd, unsigned int what, 
			       padico_sysio_callback_t callback, void*key);

/** @brief Unregister the I/O handle 
 * @note even if it is not mandatory, it is better to:
 *  -# deactivate the io
 *  -# perform the required synchronisation (check callback completion)
 *  -# finally, unregister.
 * @note Unregister a callback has a high cost. Most of the time, use 
 * deactivation; unregister the callback just before closing the file 
 * descriptor. @see padico_io_deactivate()
 */
void padico_io_unregister(padico_io_t);

/** Activates the I/O handle - when not activated, the callback is not triggered */
void padico_io_activate(padico_io_t);

/** Deactivate the I/O handle - should be used for temporarly deactivate callback trigger */
void padico_io_deactivate(padico_io_t);

/** Waits until at least one polling/callback cycle has 
 * been performed on the I/O handle.
 */
int padico_sysio_refresh(padico_io_t io);

int padico_sysio_getfd(padico_io_t io);

int     padico_sysio_socket(int domain, int type, int protocol);
ssize_t padico_sysio_read(int fildes, void *buf, size_t nbyte);
ssize_t padico_sysio_write(int fildes, const void *buf, size_t nbyte);
ssize_t padico_sysio_writev(int fildes, const struct iovec *iov, int iovcnt);
int     padico_sysio_bound_inet_socket(int*port);;
int     padico_sysio_connected_inet_socket(const struct in_addr*, int port);
int     padico_sysio_connected_inet_socket_sockaddr(const struct sockaddr* addr);
ssize_t padico_sysio_in(int fildes, void *buf, size_t nbyte);
ssize_t padico_sysio_out(int fildes, const void *buf, size_t nbyte);
ssize_t padico_sysio_outv(int fildes, const struct iovec *buf, int iovcnt);
int     padico_sysio_eos(int fd);
int     padico_sysio_shutdown_wr(int s);
int     padico_sysio_close(int fd);
int     padico_sysio_accept(int s, struct sockaddr*addr, socklen_t*addrlen);
int     padico_sysio_errno(void);

/** @} */


#endif /* PADICO_SYSIO_H */
