/** @file slab-based allocator for objects with fixed size
 * uses lock-free stacks
 */

#ifndef PUK_ALLOCATOR_H
#define PUK_ALLOCATOR_H


#include <stdlib.h>
#include <unistd.h>


#define puk_container_of(ptr, type, member)				\
        ((type *)((char *)(__typeof__ (&((type *)0)->member))(ptr)- ((size_t) &((type *)0)->member)))


#define PUK_ALLOCATOR_ALIGN 16

static inline void puk_allocator_cell_clear(void*ptr, size_t size)
{
#ifdef PUK_DEBUG
  memset(ptr, 0, size);
#endif
}
#ifdef PUK_DEBUG
#define PUK_ALLOCATOR_REDZONE_SIZE   8
#define PUK_ALLOCATOR_REDZONE(LABEL) char LABEL[PUK_ALLOCATOR_REDZONE_SIZE];
#else
#define PUK_ALLOCATOR_REDZONE(LABEL)
#define PUK_ALLOCATOR_REDZONE_SIZE   0
#endif

#define PUK_ALLOCATOR_NULL { .blocks = NULL, .slices = NULL, .slab_size = 0 }

/** regular allocator type */
#define PUK_ALLOCATOR_TYPE(ENAME, TYPE)		\
  PUK_ALLOCATOR_TYPE_EXT(ENAME, TYPE, 0)

/** single thread allocator type, to use in single thread context or with external locking */
#define PUK_ALLOCATOR_TYPE_SINGLE(ENAME, TYPE)	\
  PUK_ALLOCATOR_TYPE_EXT(ENAME, TYPE, 1)

#define PUK_ALLOCATOR_TYPE_EXT(ENAME, TYPE, SINGLE)			\
  PUK_LFSTACK_TYPE(ENAME##_block,					\
		   PUK_ALLOCATOR_REDZONE(_z1); 				\
		   TYPE block __attribute__((aligned(PUK_ALLOCATOR_ALIGN))); \
		   PUK_ALLOCATOR_REDZONE(_z2); );			\
  PUK_LFSTACK_TYPE(ENAME##_slice, struct ENAME##_block_lfstack_cell_s block; ); \
  struct ENAME##_allocator_s						\
  {									\
    ENAME##_block_lfstack_t blocks;					\
    ENAME##_slice_lfstack_t slices;					\
    ssize_t slab_size; /**< size of the last allocated slab */		\
    int new_slab; /**< whether a new slab is currently beeing allocated */ \
  };									\
  typedef struct ENAME##_allocator_s*ENAME##_allocator_t;  	   	\
  static inline void ENAME##_allocator_init(struct ENAME##_allocator_s*a, ssize_t slab_size) \
  {									\
    ENAME ## _block_lfstack_init(&a->blocks);				\
    ENAME ## _slice_lfstack_init(&a->slices);				\
    a->slab_size = slab_size;						\
    a->new_slab = 0;							\
    PUK_VG_CREATE_MEMPOOL(a, PUK_ALLOCATOR_REDZONE_SIZE, 0);		\
  }									\
  static inline ENAME##_allocator_t ENAME##_allocator_new(ssize_t slab_size) \
  {									\
    struct ENAME##_allocator_s*a =					\
      (struct ENAME##_allocator_s*)padico_malloc(sizeof(struct ENAME##_allocator_s)); \
    ENAME##_allocator_init(a, slab_size);				\
    return a;								\
  }									\
  static inline void ENAME##_allocator_destroy(struct ENAME##_allocator_s*a) \
  {									\
    struct ENAME##_slice_lfstack_cell_s*slice = NULL;			\
    do {								\
      slice = ENAME##_slice_lfstack_pop_ext(&a->slices, (SINGLE));	\
      padico_free(slice);						\
    } while(slice != NULL);						\
    ENAME ## _block_lfstack_destroy(&a->blocks);			\
    ENAME ## _slice_lfstack_destroy(&a->slices);			\
    a->slab_size = -1;							\
    PUK_VG_DESTROY_MEMPOOL(a);						\
  }									\
  static inline void ENAME##_allocator_delete(ENAME##_allocator_t a)	\
  {									\
    ENAME##_allocator_destroy(a);					\
    padico_free(a);							\
  }									\
  static inline TYPE*ENAME##_malloc(ENAME##_allocator_t a)		\
  {									\
    int r = 0;								\
    struct ENAME##_block_lfstack_cell_s*cell = NULL;			\
  retry:								\
    cell = ENAME##_block_lfstack_pop_ext(&a->blocks, (SINGLE));		\
    if(cell == NULL)							\
      {									\
	__sync_synchronize();						\
	if(!__sync_bool_compare_and_swap(&a->new_slab, 0, 1))		\
	  { puk_lfbackoff(r++); goto retry; }				\
	else								\
	  {								\
	    const ssize_t slab_size = a->slab_size;			\
	    a->slab_size *= 2;						\
	    const ssize_t slice_size = sizeof(struct ENAME##_slice_lfstack_cell_s) + \
	      sizeof(struct ENAME##_block_lfstack_cell_s) * slab_size;	\
	    struct ENAME##_slice_lfstack_cell_s*slice = NULL;		\
	    posix_memalign((void**)&slice, 16, slice_size);		\
	    if(slice == NULL)						\
	      {								\
		fprintf(stderr, "Puk: out of memory in allocator %s for type '%s' (block size = %ld)." \
			" Cannot allocate slice of size %ld bytes (%ld entries).\n", \
			#ENAME, #TYPE, sizeof(TYPE), slice_size, slab_size); \
		abort();						\
	      }								\
	    ENAME##_slice_lfstack_push_ext(&a->slices, slice, (SINGLE)); \
	    struct ENAME##_block_lfstack_cell_s*block = &slice->block;	\
	    cell = &block[0];						\
	    ssize_t i;							\
	    for(i = 1; i < slab_size; i++)				\
	      {								\
		PUK_VG_MAKE_MEM_NOACCESS(&block[i].block, sizeof(TYPE)); \
		ENAME##_block_lfstack_push_ext(&a->blocks, &block[i], (SINGLE)); \
	      }								\
	    a->new_slab = 0;						\
	    __sync_synchronize();					\
	  }								\
      }									\
    PUK_VG_MEMPOOL_ALLOC(a, &cell->block, sizeof(TYPE));		\
    puk_allocator_cell_clear(&cell->block, sizeof(TYPE));		\
    return &cell->block;						\
  }									\
  static inline void ENAME##_free(ENAME##_allocator_t a, TYPE*b)	\
  {									\
    struct ENAME##_block_lfstack_cell_s*cell = puk_container_of(b, struct ENAME##_block_lfstack_cell_s, block); \
    puk_allocator_cell_clear(&cell->block, sizeof(TYPE));		\
    PUK_VG_MEMPOOL_FREE(a, b);						\
    ENAME##_block_lfstack_push_ext(&a->blocks, cell, (SINGLE));		\
  }


#endif /* PUK_ALLOCATOR_H */
