/** @file lock-free FIFO as circular array
 */

#ifndef PUK_LFQUEUE_H
#define PUK_LFQUEUE_H

#ifdef PUK_USE_TRANSACTION
#  include "Puk-transaction.h"
#  if defined(PUK_ENABLE_GNU_TM)
/* by default, use transational-memory based lock-free queues */
#    define PUK_LFQUEUE_TYPE(ENAME, TYPE, LFQUEUE_NULL, LFQUEUE_SIZE)	\
  PUK_LFQUEUE_TM_TYPE(ENAME, TYPE, LFQUEUE_NULL, LFQUEUE_SIZE)
#  elif defined(PUK_ENABLE_RTM)
#    define PUK_LFQUEUE_TYPE(ENAME, TYPE, LFQUEUE_NULL, LFQUEUE_SIZE)	\
  PUK_LFQUEUE_RTM_TYPE(ENAME, TYPE, LFQUEUE_NULL, LFQUEUE_SIZE)
#  else
#    error "no TM method"
#  endif
#elif defined(PUK_USE_SPINLOCKS)
#  include <pthread.h>
#  define PUK_LFQUEUE_TYPE(ENAME, TYPE, LFQUEUE_NULL, LFQUEUE_SIZE)	\
  PUK_LFQUEUE_SPINLOCK_TYPE(ENAME, TYPE, LFQUEUE_NULL, LFQUEUE_SIZE)
#else
/* by default, use atomic-based lock-free queues */
#  define PUK_LFQUEUE_TYPE(ENAME, TYPE, LFQUEUE_NULL, LFQUEUE_SIZE)	\
  PUK_LFQUEUE_ATOMIC_TYPE(ENAME, TYPE, LFQUEUE_NULL, LFQUEUE_SIZE)
#endif /* PUK_USE_TRANSACTION */

extern void puk_lfbackoff(int p);

#define PUK_LFQUEUE_INITIALIZER(LFQUEUE_NULL) { ._head = 0, ._tail = 0, ._queue = { (LFQUEUE_NULL) } }

/** builds a lock-free FIFO type and functions.
 *  ENAME is base name for symbols
 *  TYPE is the type of elements in queue- must be atomic (1, 2, 4, or 8 bytes long)
 *  LFQUEUE_NULL is a nil value to fill and detect empty cells
 *  LFQUEUE_SIZE is the array size
 *  Only 'init', 'enqueue', 'dequeue' operations are available.
 *  Others operations such as 'empty', 'size' are useless for lock-free
 *  Iterators are not possible.
 *  push_front/push_back/pop_front/pop_back is probably possible at a high
 *  cost in code size and performance. For simplicity and performance, only
 *  'enqueue' = push_back; 'dequeue' = pop_front are available (1 atomic op each).
 * Conventions:
 *   _head: next cell to enqueue
 *   _tail: next cell to dequeue
 *   _head == _tail     => queue empty
 *   _head == _tail - 1 => queue full
 */
#define PUK_LFQUEUE_ATOMIC_TYPE(ENAME, TYPE, LFQUEUE_NULL, LFQUEUE_SIZE) \
  typedef TYPE ENAME ## _lfqueue_elem_t;				\
  									\
  struct ENAME ## _lfqueue_s						\
  {									\
    volatile ENAME ## _lfqueue_elem_t _queue[(LFQUEUE_SIZE)];		\
    volatile unsigned _head;						\
    volatile unsigned _tail;						\
  };									\
									\
  static inline void ENAME ## _lfqueue_init(struct ENAME ## _lfqueue_s*queue) \
  {									\
    int i;								\
    for(i = 0; i < (LFQUEUE_SIZE); i++)					\
      {									\
	queue->_queue[i] = (LFQUEUE_NULL);				\
      }									\
    queue->_head = 0;							\
    queue->_tail = 0;							\
  }									\
  									\
  static inline int ENAME ## _lfqueue_enqueue_ext(struct ENAME ## _lfqueue_s*queue, ENAME ## _lfqueue_elem_t e, int single) \
  {									\
    int r = 0; /* allocate a slot for value */				\
  retry:								\
    ;									\
    const unsigned head = queue->_head;					\
    const unsigned tail = queue->_tail;					\
    const unsigned next = (head + 1) % (LFQUEUE_SIZE);			\
    if(((head + (unsigned)(LFQUEUE_SIZE) - tail) % (LFQUEUE_SIZE)) >= (LFQUEUE_SIZE - 2)) \
      {									\
	/* Queue full, abort enqueue					\
	 * Actually, we cannot distinguish whether queue is actually	\
	 * full or we see dequeue before the corresponding enqueue.	\
	 * (no memory fence in case of single reader)			\
	 */								\
	return -1;							\
      }									\
    if(single)								\
      {									\
	queue->_head = next;						\
      }									\
    else								\
      if(!__sync_bool_compare_and_swap(&queue->_head, head, next))	\
	{ puk_lfbackoff(r++); goto retry; }				\
    /* slot is still NULL for concurrent readers, already reserved if concurrent writers */ \
    while(queue->_queue[head] != (LFQUEUE_NULL))			\
      { __sync_synchronize(); }						\
    /* store value in reserved slot */					\
    queue->_queue[head] = e;						\
    return 0;								\
  }									\
  									\
  static inline int ENAME ## _lfqueue_enqueue(struct ENAME ## _lfqueue_s*queue, ENAME ## _lfqueue_elem_t e) \
  { return ENAME ## _lfqueue_enqueue_ext(queue, e, 0); }		\
  									\
  static inline int ENAME ## _lfqueue_enqueue_single_writer(struct ENAME ## _lfqueue_s*queue, ENAME ## _lfqueue_elem_t e) \
  { return ENAME ## _lfqueue_enqueue_ext(queue, e, 1); }		\
  									\
  static inline ENAME ## _lfqueue_elem_t ENAME ## _lfqueue_dequeue_ext(struct ENAME ## _lfqueue_s*queue, int single) \
  {									\
    ENAME ## _lfqueue_elem_t e = (LFQUEUE_NULL);			\
    unsigned tail;							\
    int r = 0; /* try to dequeue */					\
  retry:								\
    ;									\
    tail = queue->_tail;						\
    const unsigned next = (tail + 1) % (LFQUEUE_SIZE);			\
    if(tail == queue->_head)						\
      {									\
	/* queue was empty, abort dequeue */				\
	return e;							\
      }									\
    if(single)								\
      { 								\
	queue->_tail = next;						\
      }									\
    else								\
      if(!__sync_bool_compare_and_swap(&queue->_tail, tail, next))	\
	{ puk_lfbackoff(r++); goto retry; }				\
    e = queue->_queue[tail];						\
    while(e == (LFQUEUE_NULL)) 						\
      { __sync_synchronize(); e = queue->_queue[tail]; }		\
    queue->_queue[tail] = (LFQUEUE_NULL);				\
    assert(e != (LFQUEUE_NULL));					\
    return e;								\
  }									\
  									\
  static inline ENAME ## _lfqueue_elem_t ENAME ## _lfqueue_dequeue(struct ENAME ## _lfqueue_s*queue) \
  { return ENAME ## _lfqueue_dequeue_ext(queue, 0); }			\
									\
  static inline ENAME ## _lfqueue_elem_t ENAME ## _lfqueue_dequeue_single_reader(struct ENAME ## _lfqueue_s*queue) \
  { return ENAME ## _lfqueue_dequeue_ext(queue, 1); }			\
									\
  /** hint whether the queue is empty.					\
   *  Do not rely on it: state may have changed before dequeue		\
   *  Usefull to reduce active polling, though */		       	\
  static inline int ENAME ## _lfqueue_empty(struct ENAME ## _lfqueue_s*queue) \
  {									\
    return(queue->_tail == queue->_head);				\
  }									\
  /** hint about the number of elements in queue .			\
   * Do not rely on it: state may have changed before function return.	\
   */									\
  static inline int ENAME ## _lfqueue_load(struct ENAME ## _lfqueue_s*queue) \
  {									\
    return((LFQUEUE_SIZE + queue->_head - queue->_tail) % LFQUEUE_SIZE); \
  }									\
  /** get the allocated queue size */					\
  static inline int ENAME ## _lfqueue_size(struct ENAME ## _lfqueue_s*queue) \
  {									\
    return (LFQUEUE_SIZE);						\
  }


/** Build a dynamic lock-free queue. Compared to regular lfqueue, this one
 * is dynamically allocated, with a dynamic size.
 */
#define PUK_DLFQ_TYPE(ENAME, TYPE, DLFQ_NULL)				\
  typedef TYPE ENAME ## _dlfq_elem_t;					\
  									\
  struct ENAME ## _dlfq_s						\
  {									\
    ENAME ## _dlfq_elem_t*_queue;					\
    volatile unsigned _head;						\
    volatile unsigned _tail;						\
    unsigned _size;							\
  };									\
									\
  static inline void ENAME ## _dlfq_init(struct ENAME ## _dlfq_s*queue, unsigned size) \
  {									\
    queue->_queue = padico_malloc(size * sizeof(ENAME##_dlfq_elem_t));	\
    int i;								\
    for(i = 0; i < (size); i++)						\
      {									\
	queue->_queue[i] = (DLFQ_NULL);					\
      }									\
    queue->_size = size;						\
    queue->_head = 0;							\
    queue->_tail = 0;							\
  }									\
  									\
  static inline void ENAME ## _dlfq_destroy(struct ENAME ## _dlfq_s*queue) \
  {									\
    padico_free(queue->_queue);						\
    queue->_queue = NULL;						\
  }									\
  									\
  static inline int ENAME ## _dlfq_enqueue_ext(struct ENAME ## _dlfq_s*queue, ENAME ## _dlfq_elem_t e, int single) \
  {									\
    int r = 0; /* allocate a slot for value */				\
  retry:								\
    ;									\
    const unsigned head = queue->_head;					\
    const unsigned tail = queue->_tail;					\
    const unsigned size = queue->_size;					\
    const unsigned next = (head + 1) % (size);				\
    if(((head + size - tail) % size) >= (size - 2))			\
      {									\
	/* Queue full, abort enqueue					\
	 * Actually, we cannot distinguish whether queue is actually	\
	 * full or we see dequeue before the corresponding enqueue.	\
	 * (no memory fence in case of single reader)			\
	 */								\
	return -1;							\
      }									\
    if(single)								\
      {									\
	queue->_head = next;						\
      }									\
    else								\
      if(!__sync_bool_compare_and_swap(&queue->_head, head, next))	\
	{ puk_lfbackoff(r++); goto retry; }				\
    /* slot is still NULL for concurrent readers, already reserved if concurrent writers */ \
    while(queue->_queue[head] != (DLFQ_NULL))				\
      { __sync_synchronize(); }						\
    /* store value in reserved slot */					\
    queue->_queue[head] = e;						\
    return 0;								\
  }									\
  									\
  static inline int ENAME ## _dlfq_enqueue(struct ENAME ## _dlfq_s*queue, ENAME ## _dlfq_elem_t e) \
  { return ENAME ## _dlfq_enqueue_ext(queue, e, 0); }			\
  									\
  static inline int ENAME ## _dlfq_enqueue_single_writer(struct ENAME ## _dlfq_s*queue, ENAME ## _dlfq_elem_t e) \
  { return ENAME ## _dlfq_enqueue_ext(queue, e, 1); }			\
  									\
  static inline ENAME ## _dlfq_elem_t ENAME ## _dlfq_dequeue_ext(struct ENAME ## _dlfq_s*queue, int single) \
  {									\
    ENAME ## _dlfq_elem_t e = (DLFQ_NULL);				\
    unsigned tail;							\
    int r = 0; /* try to dequeue */					\
  retry:								\
    ;									\
    tail = queue->_tail;						\
    const unsigned next = (tail + 1) % queue->_size;			\
    if(tail == queue->_head)						\
      {									\
	/* queue was empty, abort dequeue */				\
	return e;							\
      }									\
    if(single)								\
      { 								\
	queue->_tail = next;						\
      }									\
    else								\
      if(!__sync_bool_compare_and_swap(&queue->_tail, tail, next))	\
	{ puk_lfbackoff(r++); goto retry; }				\
    e = queue->_queue[tail];						\
    while(e == (DLFQ_NULL)) 						\
      { __sync_synchronize(); e = queue->_queue[tail]; }		\
    queue->_queue[tail] = (DLFQ_NULL);					\
    assert(e != (DLFQ_NULL));						\
    return e;								\
  }									\
  									\
  static inline ENAME ## _dlfq_elem_t ENAME ## _dlfq_dequeue(struct ENAME ## _dlfq_s*queue) \
  { return ENAME ## _dlfq_dequeue_ext(queue, 0); }			\
									\
  static inline ENAME ## _dlfq_elem_t ENAME ## _dlfq_dequeue_single_reader(struct ENAME ## _dlfq_s*queue) \
  { return ENAME ## _dlfq_dequeue_ext(queue, 1); }			\
									\
  /** hint whether the queue is empty.					\
   *  Do not rely on it: state may have changed before dequeue		\
   *  Usefull to reduce active polling, though */		       	\
  static inline int ENAME ## _dlfq_empty(struct ENAME ## _dlfq_s*queue) \
  {									\
    return(queue->_tail == queue->_head);				\
  }									\
  /** hint about the number of elements in queue .			\
   * Do not rely on it: state may have changed before function return.	\
   */									\
  static inline int ENAME ## _dlfq_load(struct ENAME ## _dlfq_s*queue)	\
  {									\
    return((queue->_size + queue->_head - queue->_tail) % queue->_size); \
  }									\
  /** get the allocated queue size */					\
  static inline int ENAME ## _dlfq_size(struct ENAME ## _dlfq_s*queue)	\
  {									\
    return queue->_size;						\
  }									\
  /* Resize the DLFQ; needs external locking! */			\
  static inline void ENAME ## _dlfq_resize(struct ENAME ## _dlfq_s*queue, int newsize) \
  {									\
    const unsigned load = ENAME ## _dlfq_load(queue);			\
    if(newsize - 2 < load)						\
      {									\
	padico_fatal("Puk: cannot resize DLFQ smaller than actual content.\n");	\
      }									\
    ENAME##_dlfq_elem_t*newbuf = padico_malloc(newsize * sizeof(ENAME##_dlfq_elem_t)); \
    int i;								\
    int head = 0;							\
    for(i = 0; i < newsize; i++)					\
      {									\
	ENAME##_dlfq_elem_t e = ENAME##_dlfq_dequeue_single_reader(queue); \
	if(e != (DLFQ_NULL))						\
	  {								\
	    newbuf[i] = e;						\
	    head++;							\
	  }								\
	else								\
	  {								\
	    newbuf[i] = (DLFQ_NULL);					\
	  }								\
      }									\
    padico_free(queue->_queue);						\
    queue->_size = newsize;						\
    queue->_queue = newbuf;						\
    queue->_head = head;						\
    queue->_tail = 0;							\
  }


/* ********************************************************* */
/* ** spinlock variants */

#define PUK_LFQUEUE_SPINLOCK_TYPE(ENAME, TYPE, LFQUEUE_NULL, LFQUEUE_SIZE) \
  typedef TYPE ENAME ## _lfqueue_elem_t;				\
  									\
  struct ENAME ## _lfqueue_s						\
  {									\
    pthread_spinlock_t _lock;						\
    ENAME ## _lfqueue_elem_t _queue[(LFQUEUE_SIZE)];			\
    unsigned _head;							\
    unsigned _tail;							\
  };									\
									\
  static inline void ENAME ## _lfqueue_init(struct ENAME ## _lfqueue_s*queue) \
  {									\
    int i;								\
    for(i = 0; i < (LFQUEUE_SIZE); i++)					\
      {									\
	queue->_queue[i] = (LFQUEUE_NULL);				\
      }									\
    queue->_head = 0;							\
    queue->_tail = 0;							\
    pthread_spin_init(&queue->_lock, PTHREAD_PROCESS_PRIVATE);		\
  }									\
  									\
  static inline int ENAME ## _lfqueue_enqueue(struct ENAME ## _lfqueue_s*queue, ENAME ## _lfqueue_elem_t e) \
  {									\
    pthread_spin_lock(&queue->_lock);					\
    const unsigned head = queue->_head;					\
    if(((head + (unsigned)(LFQUEUE_SIZE) - queue->_tail) % (LFQUEUE_SIZE)) >= (LFQUEUE_SIZE - 2)) \
      {									\
	/* Queue full, abort enqueue */					\
	pthread_spin_unlock(&queue->_lock);				\
	return -1;							\
      }									\
    queue->_head = (head + 1) % (LFQUEUE_SIZE);				\
    assert(queue->_queue[head] == (LFQUEUE_NULL));			\
    queue->_queue[head] = e;						\
    pthread_spin_unlock(&queue->_lock);					\
    return 0;								\
  }									\
  									\
  static inline int ENAME ## _lfqueue_enqueue_single_writer(struct ENAME ## _lfqueue_s*queue, ENAME ## _lfqueue_elem_t e) \
  { return ENAME ## _lfqueue_enqueue(queue, e); }			\
  									\
  static inline ENAME ## _lfqueue_elem_t ENAME ## _lfqueue_dequeue(struct ENAME ## _lfqueue_s*queue) \
  {									\
    pthread_spin_lock(&queue->_lock);					\
    unsigned tail = queue->_tail;					\
    if(tail == queue->_head)						\
      {									\
	/* queue was empty, abort dequeue */				\
	pthread_spin_unlock(&queue->_lock);				\
	return (LFQUEUE_NULL);						\
      }									\
    const unsigned next = (tail + 1) % (LFQUEUE_SIZE);			\
    queue->_tail = next;						\
    ENAME ## _lfqueue_elem_t e = queue->_queue[tail];			\
    queue->_queue[tail] = (LFQUEUE_NULL);				\
    assert(e != (LFQUEUE_NULL));					\
    pthread_spin_unlock(&queue->_lock);					\
    return e;								\
  }									\
  									\
  static inline ENAME ## _lfqueue_elem_t ENAME ## _lfqueue_dequeue_single_reader(struct ENAME ## _lfqueue_s*queue) \
  { return ENAME ## _lfqueue_dequeue(queue); }				\
									\
  /** hint whether the queue is empty.					\
   *  Do not rely on it: state may have changed before dequeue		\
   *  Usefull to reduce active polling, though */		       	\
  static inline int ENAME ## _lfqueue_empty(struct ENAME ## _lfqueue_s*queue) \
  {									\
    return(queue->_tail == queue->_head);				\
  }									\
  /** hint about the number of elements in queue .			\
   * Do not rely on it: state may have changed before function return.	\
   */									\
  static inline int ENAME ## _lfqueue_load(struct ENAME ## _lfqueue_s*queue) \
  {									\
    return((LFQUEUE_SIZE + queue->_head - queue->_tail) % LFQUEUE_SIZE); \
  }									\
  /** get the allocated queue size */					\
  static inline int ENAME ## _lfqueue_size(struct ENAME ## _lfqueue_s*queue) \
  {									\
    return (LFQUEUE_SIZE);						\
  }


#endif /* PUK_LFQUEUE_H */
