/* @file
*/

#include "Module.h"
#include "Puk-internals.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <ifaddrs.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netpacket/packet.h>
#include <limits.h>

#ifdef __linux__
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#endif

PADICO_MODULE_HOOK(Puk);

/* ********************************************************* */

/** get local IPv4 address */
struct in_addr puk_inet_getaddr(void)
{
  struct in_addr inaddr;
  inaddr.s_addr = INADDR_ANY;
  if(puk_mckernel())
    {
#ifdef __linux__
      /* create netlink socket */
      int fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
      if(fd < 0)
	{
	  padico_fatal("Topology: cannot open netlink socket (socket: %s).\n", strerror(errno));
	}
      struct sockaddr_nl nladdr = {
	.nl_family = AF_NETLINK,
	.nl_pad    = 0,
	.nl_pid    = getpid(),
	.nl_groups = 0
      };
      int rc = bind(fd, (struct sockaddr*)&nladdr, sizeof(nladdr));
      if(rc < 0)
	{
	  padico_fatal("Topology: cannot bind netlink socket.\n");
	}
      /* netlink request */
      struct 
      {
	struct nlmsghdr nl; /**< netlink header */
	struct ifaddrmsg ifa;
	char            buf[8192];
      } req;
      memset(&req, 0, sizeof(req));
      req.nl.nlmsg_len   = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
      req.nl.nlmsg_type  = RTM_GETADDR;
      req.nl.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
      req.nl.nlmsg_seq   = 0;
      req.nl.nlmsg_pid   = getpid();
      req.ifa.ifa_family = AF_INET;
      struct iovec iov = { .iov_base = (void*)&req.nl, .iov_len = req.nl.nlmsg_len };
      /* send request */
      struct sockaddr_nl peer = {
	.nl_family = AF_NETLINK,
	.nl_pad    = 0,
	.nl_pid    = 0, /* 0 = kernel */
	.nl_groups = 0
      };
      struct msghdr msg = {
	.msg_iov     = &iov,
	.msg_iovlen  = 1,
	.msg_name    = (void*)&peer,
	.msg_namelen = sizeof(peer)
      };
      rc = sendmsg(fd, &msg, 0);
      if(rc < 0)
	padico_fatal("Topology: error while sending netlink request.\n");
      /* receive reply */
      char buf[8192];
      memset(buf, 0, sizeof(buf));
      char*p = buf;
      int nll = 0;
      while(1) 
	{
	  rc = recv(fd, p, sizeof(buf) - nll, 0);
	  struct nlmsghdr*nlp = (struct nlmsghdr*)p;
	  if(nlp->nlmsg_type == NLMSG_DONE)
	    break;
	  p += rc;
	  nll += rc;
	}
      /* parse reply */
      const struct nlmsghdr*nlp = (struct nlmsghdr*)buf;
      while(NLMSG_OK(nlp, nll))
	{
	  const struct ifaddrmsg*ifa = (struct ifaddrmsg*)NLMSG_DATA(nlp);
	  const struct rtattr*rtap = (struct rtattr*)IFA_RTA(ifa);
	  int rtl = IFA_PAYLOAD(nlp);
	  while(RTA_OK(rtap, rtl))
	    {
	      switch(rtap->rta_type)
		{
		case IFA_LOCAL:
		  {
		    if((ifa->ifa_family == AF_INET) && !(ifa->ifa_flags & IFF_LOOPBACK))
		      {
			const struct in_addr*_inaddr = (const struct in_addr*)RTA_DATA(rtap);
			inaddr = *_inaddr;
		      }
		  }
		  break;
		default:
		  break;
		}
	      rtap = RTA_NEXT(rtap,rtl);
	    }
	  nlp = NLMSG_NEXT(nlp, nll);
	}
      close(fd);
#else
      padico_fatal("Puk: internal error- running on mckernel, without linux headers.");
#endif
    }
  else
    {
      /* !MCKERNEL */
      const char*puk_inet_iface = padico_getenv("PUK_INET_IFACE");
      const char*puk_inet_resolve = padico_getenv("PUK_INET_RESOLVE");
      if(puk_inet_resolve && puk_inet_iface)
	{
	  padico_fatal("Puk: cannot give PUK_INET_RESOLVE & PUK_INET_IFACE at the same time.\n");
	}
      if(puk_inet_resolve)
	{
	  /* DNS-based address detection */
	  char hostname[HOST_NAME_MAX];
	  gethostname(hostname, HOST_NAME_MAX);
	  const struct hostent*he = gethostbyname2(hostname, AF_INET);
	  if(he != NULL)
	    {
	      const struct in_addr**addr = (const struct in_addr**)he->h_addr_list;
	      inaddr.s_addr = (*addr)->s_addr;
	    }
	}
      else
	{
	  /* regular method: getifaddrs */
	  struct ifaddrs*ifa_list = NULL;
	  int rc = getifaddrs(&ifa_list);
	  if(rc == 0)
	    {
	      struct ifaddrs*i;
	      for(i = ifa_list; i != NULL; i = i->ifa_next)
		{
		  const int is_loopback = !!(i->ifa_flags & IFF_LOOPBACK);
		  const int is_up = !!(i->ifa_flags & IFF_UP);
		  if( is_up && (!is_loopback) &&
		      (i->ifa_addr != NULL) &&
		      i->ifa_addr->sa_family == AF_INET)
		    {
		      if((puk_inet_iface == NULL) || (strcmp(puk_inet_iface, i->ifa_name) == 0))
			{
			  struct sockaddr_in*saddr_in = (struct sockaddr_in*)i->ifa_addr;
			  inaddr = saddr_in->sin_addr;
			  break;
			}
		    }
		}
	      freeifaddrs(ifa_list);
	      ifa_list = NULL;
	    }
	}
    }
  if(inaddr.s_addr == INADDR_ANY)
    {
      padico_fatal("Puk: unable to detect IP address.\n");
    }
  return inaddr;
}

