/*
 * NewMadeleine
 * Copyright (C) 2006-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/* -*- Mode: C; tab-width: 2; c-basic-offset: 2 -*- */


/* *********************************************************
 * Rationale of the newmad/ibverbs driver
 *
 * This is a multi-method driver. Four methods are available:
 *   -- bycopy: data is copied in a pre-allocated, pre-registered
 *      memory region, then sent through RDMA into a pre-registered
 *      memory region of the peer node.
 *   -- adaptrdma: data is copied in a pre-allocated, pre-registered
 *      memory region, with an adaptive super-pipeline. Memory blocks are
 *      copied as long as the previous RDMA send doesn't complete.
 *      A guard check ensures that block size progression is at least
 *      2-base exponential to prevent artifacts to kill performance.
 *   -- regrdma: memory is registered on the fly on both sides, using
 *      a pipeline with variable block size. For each block, the
 *      receiver sends an ACK with RDMA info (raddr, rkey), then
 *      zero-copy RDMA is performed.
 *   -- rcache: memory is registered on the fly on both sides,
 *      sequencially, using puk_mem_* functions from Puk-ABI that
 *      manage a cache.
 *
 * Method is chosen as follows:
 *   -- tracks for small messages always uses 'bycopy'
 *   -- tracks with rendez-vous use 'adaptrdma' for smaller messages
 *      and 'regrdma' for larger messages. Usually, the threshold
 *      is 224kB (from empirical results about registration/send overlap)
 *      on MT23108, and 2MB on ConnectX.
 *   -- tracks with rendez-vous use 'rcache' when ib_rcache is activated
 *      in the flavor.
 */


#include "nm_ibverbs.h"
#include <Padico/Module.h>
#include <Padico/Puk.h>
#include <nm_data.h>
#include <nm_private.h>
#ifdef NMAD_HWLOC
#include <hwloc.h>
#include <hwloc/openfabrics-verbs.h>
#endif /* NMAD_HWLOC */

#include <unistd.h>
#include <fcntl.h>
#include <poll.h>


static const struct puk_component_driver_s nm_ibverbs_common_component =
  {
    .instantiate = NULL,
    .destroy     = NULL
  };

struct nm_ibverbs_context_s*nm_ibverbs_context_new(puk_context_t p_context, const struct nm_ibverbs_opts_s*p_opts);

static const struct nm_ibverbs_iface_s nm_ibverbs_common_iface =
  {
    .context_new = &nm_ibverbs_context_new
  };

PADICO_MODULE_COMPONENT(NewMad_ibverbs_common,
  puk_component_declare("NewMad_ibverbs_common",
                        puk_component_provides("PadicoComponent", "component", &nm_ibverbs_common_component),
                        puk_component_provides("NewMad_ibverbs", "ibverbs", &nm_ibverbs_common_iface),
                        puk_component_attr("ibv_device", "auto"),
                        puk_component_attr("ibv_port", "auto"),
                        puk_component_attr("subnet_prefix", "auto"),
                        puk_component_attr("network", "Infiniband-default")));

/** checksum algorithm to use with ibverbs */
static puk_checksum_t _nm_ibverbs_checksum = NULL;

/** ptr alignment enforced in every packets (through padding) */
int nm_ibverbs_alignment = 64;
/** ptr alignment enforced for buffer allocation */
int nm_ibverbs_memalign  = 4096;

/* ********************************************************* */

static void nm_ibverbs_cnx_qp_reset(struct nm_ibverbs_cnx_s*p_ibverbs_cnx);
static void nm_ibverbs_cnx_qp_init(struct nm_ibverbs_cnx_s*p_ibverbs_cnx);
static void nm_ibverbs_cnx_qp_rtr(struct nm_ibverbs_cnx_s*p_ibverbs_cnx);
static void nm_ibverbs_cnx_qp_rts(struct nm_ibverbs_cnx_s*p_ibverbs_cnx);

static int nm_ibverbs_sync_send(const void*sbuf, const void*rbuf, int size, struct nm_ibverbs_cnx_s*p_ibverbs_cnx);

/* ********************************************************* */

static void nm_ibverbs_common_init(void)
{
  static int init_done = 0;
  if(!init_done)
    {
      init_done = 1;
      const char*checksum_env = getenv("NMAD_IBVERBS_CHECKSUM");
      if(_nm_ibverbs_checksum == NULL && checksum_env != NULL)
        {
          puk_checksum_t checksum = puk_checksum_get(checksum_env);
          if(checksum == NULL)
            NM_FATAL("ibverbs: checksum algorithm *%s* not available.\n", checksum_env);
          _nm_ibverbs_checksum = checksum;
          NM_DISPF("ibverbs- checksum enabled (%s).\n", checksum->name);
        }
      const char*align_env = getenv("NMAD_IBVERBS_ALIGN");
      if(align_env != NULL)
        {
          nm_ibverbs_alignment = atoi(align_env);
          NM_DISPF("ibverbs- alignment forced to %d\n", nm_ibverbs_alignment);
        }
      const char*memalign_env = getenv("NMAD_IBVERBS_MEMALIGN");
      if(memalign_env != NULL)
        {
          nm_ibverbs_memalign = atoi(memalign_env);
          NM_DISPF("ibverbs- memalign forced to %d\n", nm_ibverbs_memalign);
        }
    }
}

/** get the ibverbs component from its parent component & create context */
struct nm_ibverbs_context_s*nm_ibverbs_context_resolve_opts(puk_context_t p_context, const struct nm_ibverbs_opts_s*p_opts)
{
  if(p_context == NULL)
    {
      NM_FATAL("nm_ibverbs: parent context is NULL; cannot create ibverbs context.\n");
    }
  puk_component_conn_t p_ibverbs_conn = puk_context_conn_lookup(p_context, puk_iface_NewMad_ibverbs(), "ibverbs");
  if(p_ibverbs_conn == NULL)
    {
      NM_FATAL("nm_ibverbs: 'ibverbs' receptacle not found in component '%s'.\n", p_context->component->name);
    }
  if(p_ibverbs_conn->facet == NULL || p_ibverbs_conn->facet->driver == NULL)
    {
      NM_FATAL("nm_ibverbs: 'ibverbs' facet not connected in component '%s', cannot create ibverbs context.\n",
               p_context->component->name);
    }
  const struct nm_ibverbs_iface_s*p_ibverbs_iface = p_ibverbs_conn->facet->driver;
  return (p_ibverbs_iface->context_new)(p_ibverbs_conn->context, p_opts);
}

/** get in ibverbs context from a puk context */
struct nm_ibverbs_context_s*nm_ibverbs_context_new(puk_context_t p_context, const struct nm_ibverbs_opts_s*p_opts)
{
  nm_ibverbs_common_init();
  struct nm_ibverbs_context_s*p_ibverbs_context = malloc(sizeof(struct nm_ibverbs_context_s));
  memset(p_ibverbs_context, 0, sizeof(struct nm_ibverbs_context_s));
  const int use_comp_channel = p_opts->use_comp_channel;
  const int need_srq = p_opts->use_srq;
  const char*device = puk_context_getattr(p_context, "ibv_device");
  const char*s_port = puk_context_getattr(p_context, "ibv_port");
  p_ibverbs_context->ib_opts.mtu = p_opts->mtu;
  /* open HCA */
  if(device == NULL || s_port == NULL)
    NM_FATAL("ibverbs: cannot find ibv_device/ibv_port attributes.\n");
  const int port = (strcmp(s_port, "auto") == 0) ? 1 : atoi(s_port);
  nm_ibverbs_hca_open(&p_ibverbs_context->hca, device, port);
  /* open SRQ & comp channel if needed */
  p_ibverbs_context->srq.p_srq = NULL;
  p_ibverbs_context->srq.srq_cq = NULL;      
  if(need_srq)
    {
      if(p_ibverbs_context->hca.ib_caps.max_srq > 0)
        {
          struct ibv_comp_channel*p_comp_channel = NULL;
          struct ibv_srq_init_attr srq_attr = { .srq_context  = p_ibverbs_context,
                                                .attr.max_wr  = NM_IBVERBS_RX_DEPTH,
                                                .attr.max_sge = NM_IBVERBS_MAX_SG_RQ };
          p_ibverbs_context->srq.p_srq = ibv_create_srq(p_ibverbs_context->hca.pd, &srq_attr);
          if(p_ibverbs_context->srq.p_srq == NULL)
            {
              NM_WARN("ibverbs: cannot allocate SRQ.\n");
              p_ibverbs_context->srq.p_srq = NULL;
            }
          if(use_comp_channel)
            {
              p_comp_channel = ibv_create_comp_channel(p_ibverbs_context->hca.context);
              if(p_comp_channel == NULL)
                {
                  NM_FATAL("ibverbs: cannot create comp_channel (%s).\n", strerror(errno));
                }
              p_ibverbs_context->srq.comp_channel = p_comp_channel;
              p_ibverbs_context->ib_opts.use_comp_channel = 1;
              NM_SYS(pipe)(p_ibverbs_context->srq.cancel_fd);
              p_ibverbs_context->srq.waiting = 0;
            }
          p_ibverbs_context->srq.srq_cq = ibv_create_cq(p_ibverbs_context->hca.context, NM_IBVERBS_SRQ_DEPTH,
                                                    p_ibverbs_context, p_comp_channel, 0);
          if(p_ibverbs_context->srq.srq_cq == NULL)
            {
              NM_FATAL("ibverbs: cannot create CQ (%s).\n", strerror(errno));
            }
          p_ibverbs_context->ib_opts.use_srq = 1;
          if(p_ibverbs_context->ib_opts.use_comp_channel)
            {
              int rc = ibv_req_notify_cq(p_ibverbs_context->srq.srq_cq, 0);
              if(rc)
                {
                  NM_FATAL("ibverbs- couldn't request CQ notification (%s)\n", strerror(errno));
                }
            }
        }
      else
        {
          NM_FATAL("ibverbs-  SRQ requested but not available on hardware.\n");
        }
    }
  NM_DISPF("ibverbs-   SRQ available = %d; enabled = %d; comp_channel = %d; MTU = %d\n",
           (p_ibverbs_context->hca.ib_caps.max_srq > 0),
           p_ibverbs_context->ib_opts.use_srq, p_ibverbs_context->ib_opts.use_comp_channel,
           nm_ibverbs_mtu_values[p_opts->mtu]);
  /* open UD endpoint & pre-post a recv */
  nm_ibverbs_ud_ep_open(&p_ibverbs_context->connect.udep, &p_ibverbs_context->hca);
  nm_ibverbs_cnx_addr_entry_hashtable_init(&p_ibverbs_context->connect.addr_hash);
  memset(&p_ibverbs_context->connect.recv_packets, 0, NM_IBVERBS_RX_DEPTH * sizeof(struct nm_ibverbs_cnx_connect_packet_s));
  nm_ibverbs_ud_recv_pool_post(&p_ibverbs_context->connect.recv_pool, &p_ibverbs_context->connect.udep,
                               &p_ibverbs_context->connect.recv_packets[0], sizeof(struct nm_ibverbs_cnx_addr_msg_s),
                               NM_IBVERBS_RX_DEPTH - 1 );
  return p_ibverbs_context;
}

void nm_ibverbs_context_get_url(struct nm_ibverbs_context_s*p_ibverbs_context, const void**pp_url, size_t*p_urlsize)
{
  *pp_url = &p_ibverbs_context->connect.udep.local_addr;
  *p_urlsize = sizeof(p_ibverbs_context->connect.udep.local_addr);
}

void nm_ibverbs_context_delete(struct nm_ibverbs_context_s*p_ibverbs_context)
{
  if(p_ibverbs_context->srq.p_srq != NULL)
    ibv_destroy_srq(p_ibverbs_context->srq.p_srq);
  nm_ibverbs_hca_close(&p_ibverbs_context->hca);
  nm_ibverbs_cnx_addr_entry_hashtable_destroy(&p_ibverbs_context->connect.addr_hash);
  free(p_ibverbs_context);
}

int nm_ibverbs_context_wait_event(struct nm_ibverbs_context_s*p_ibverbs_context)
{
  p_ibverbs_context->srq.waiting = 1;
  if(p_ibverbs_context->ib_opts.use_comp_channel)
    {
      /* wait on fd set, to allow cancellation */
      struct pollfd fds[2];
      fds[0].fd      = p_ibverbs_context->srq.comp_channel->fd;
      fds[0].events  = POLLIN;
      fds[0].revents = 0;
      fds[1].fd      = p_ibverbs_context->srq.cancel_fd[0];
      fds[1].events  = POLLIN;
      fds[1].revents = 0;
      int rc = NM_SYS(poll)(fds, 2 /* nfds */, 0 /* timeout_ms */);
      if(rc > 0 && fds[1].revents & POLLIN)
        {
          char dummy;
          NM_SYS(read)(p_ibverbs_context->srq.cancel_fd[0], &dummy, 1);
        }
      else if(rc > 0 && fds[0].revents & POLLIN)
        {
          /* Wait for the completion event */
          struct ibv_cq*ev_cq = p_ibverbs_context->srq.srq_cq;
          void*ev_context = NULL;
          rc = ibv_get_cq_event(p_ibverbs_context->srq.comp_channel, &ev_cq, &ev_context);
          if(rc)
            {
              NM_WARN("ibverbs- failed to get cq_event\n");
              return -NM_EBROKEN;
            }
          assert(ev_cq == p_ibverbs_context->srq.srq_cq);
          /* Ack the event */
          ibv_ack_cq_events(p_ibverbs_context->srq.srq_cq, 1);
          
          /* Request notification upon the next completion event */
          rc = ibv_req_notify_cq(p_ibverbs_context->srq.srq_cq, 0);
          if(rc)
            {
              NM_WARN("ibverbs- couldn't request CQ notification\n");
              return -NM_EBROKEN;
            }
        }
      else if(rc < 0)
        {
          NM_WARN("ibverbs- error %d (%s) while polling completion channel.\n", errno, strerror(errno));
          return -NM_EBROKEN;
        }
    }
  else
    {
      NM_FATAL("ibverbs- ibv_comp_channel is disabled; cannot wait on comp_channel.\n");
    }
  p_ibverbs_context->srq.waiting = 0;
  return NM_ESUCCESS;
}

void nm_ibverbs_context_cancel_wait(struct nm_ibverbs_context_s*p_ibverbs_context)
{
  char dummy = 0;
  NM_SYS(write)(p_ibverbs_context->srq.cancel_fd[1], &dummy, 1);
  while(p_ibverbs_context->srq.waiting)
    {
      sched_yield();
    }
}

void nm_ibverbs_hca_open(struct nm_ibverbs_hca_s*p_hca, const char*device, int port)
{
  if(port == -1)
    port = 1;
  const int autodetect = (strcmp(device, "auto") == 0);
  /* find IB device */
  int dev_amount = -1;
  int dev_auto = 0;
  struct ibv_device**dev_list = ibv_get_device_list(&dev_amount);
  if(!dev_list)
    {
      NM_FATAL("ibverbs: no device found.\n");
    }
 retry_autodetect:
  if(autodetect)
    {
      p_hca->ib_dev = dev_list[dev_auto];
    }
  else
    {
      struct ibv_device**dev_ptr = dev_list;
      p_hca->ib_dev = NULL;
      while((p_hca->ib_dev == NULL) && (*dev_ptr != NULL))
	{
	  struct ibv_device*ib_dev = *dev_ptr;
	  const char*ib_devname = ibv_get_device_name(ib_dev);
	  if(strcmp(ib_devname, device) == 0)
	    {
	      p_hca->ib_dev = ib_dev;
	      break;
	    }
	  dev_ptr++;
	}
    }
  if(p_hca->ib_dev == NULL)
    {
      NM_FATAL("ibverbs: cannot find required device %s.\n", device);
    }

  /* open IB context */
  p_hca->context = ibv_open_device(p_hca->ib_dev);
  if(p_hca->context == NULL)
    {
      NM_FATAL("ibverbs: cannot open IB context (%s).\n", strerror(errno));
    }

  /* get IB context attributes */
  struct ibv_device_attr device_attr;
  int rc = ibv_query_device(p_hca->context, &device_attr);
  if(rc != 0)
    {
      NM_FATAL("ibverbs: cannot get device capabilities (%s).\n", strerror(errno));
    }

  /* detect LID */
  struct ibv_port_attr port_attr;
  rc = ibv_query_port(p_hca->context, port, &port_attr);
  if(rc != 0)
    {
      NM_FATAL("ibverbs: dev = %s; port = %d; cannot get local port attributes (%s).\n",
               ibv_get_device_name(p_hca->ib_dev), port, strerror(rc));
    }
  if(port_attr.state != IBV_PORT_ACTIVE)
    {
      NM_WARN("ibverbs- dev = %s; port = %d; port is down.\n",
	      ibv_get_device_name(p_hca->ib_dev), port);
      if(autodetect)
	{
	  if(port < device_attr.phys_port_cnt)
	    port++;
	  else
	    dev_auto++;
	  if(dev_list[dev_auto] == NULL)
	    {
	      NM_FATAL("ibverbs: cannot find active IB device.\n");
	    }
	  goto retry_autodetect;
	}

    }
  p_hca->lid = port_attr.lid;
  if(p_hca->lid == 0)
    {
      NM_WARN("ibverbs: LID is null- subnet manager has probably crashed.\n");
    }

  /* IB capabilities */
  p_hca->ib_caps.max_qp        = device_attr.max_qp;
  p_hca->ib_caps.max_qp_wr     = device_attr.max_qp_wr;
  p_hca->ib_caps.max_cq        = device_attr.max_cq;
  p_hca->ib_caps.max_cqe       = device_attr.max_cqe;
  p_hca->ib_caps.max_mr        = device_attr.max_mr;
  p_hca->ib_caps.max_srq       = device_attr.max_srq;
  p_hca->ib_caps.max_mr_size   = device_attr.max_mr_size;
  p_hca->ib_caps.page_size_cap = device_attr.page_size_cap;
  p_hca->ib_caps.max_msg_size  = port_attr.max_msg_sz;

  int link_width = -1;
  switch(port_attr.active_width)
    {
    case 1: link_width = 1;  break;
    case 2: link_width = 4;  break;
    case 4: link_width = 8;  break;
    case 8: link_width = 12; break;
    default:
      NM_WARN("ibverbs: unknown active_width = %d\n", port_attr.active_width);
      break;
    }
  int link_rate = -1;
  const char*s_link_rate = "unknown";
  switch(port_attr.active_speed)
    {
    case 0x01: link_rate = 2;  s_link_rate = "SDR";   break; /* 2.5 Gbps on the wire + 8/10 encoding */
    case 0x02: link_rate = 4;  s_link_rate = "DDR";   break; /* 5 Gbps on the wire + 8/10 encoding  */
    case 0x04: link_rate = 8;  s_link_rate = "QDR";   break; /* 8 Gbps on the wire + 8/10 encoding  */
    case 0x08: link_rate = 10; s_link_rate = "FDR10"; break; /* 10 Gbps + 64/66 encoding */
    case 0x10: link_rate = 14; s_link_rate = "FDR";   break; /* 14 Gbps + 64/66 encoding */
    case 0x20: link_rate = 25; s_link_rate = "EDR";   break; /* 25 Gbps + 64/66 encoding */
    case 0x40: link_rate = 50; s_link_rate = "HDR";   break; /* 50 Gbps + 64/66 encoding */
    case 0x80: link_rate = 100; s_link_rate = "NDR";  break;
    default:
      NM_WARN("ibverbs: unknown active_speed = %d\n", port_attr.active_speed);
      break;
    }
  p_hca->ib_caps.data_rate = link_width * link_rate;

  NM_DISPF("ibverbs- device '%s'- %dx %s (%d Gb/s); LID = 0x%02X\n",
	   ibv_get_device_name(p_hca->ib_dev), link_width, s_link_rate, p_hca->ib_caps.data_rate, p_hca->lid);
#ifdef DEBUG
  NM_DISPF("ibverbs-   max_qp=%d; max_qp_wr=%d; max_cq=%d; max_cqe=%d; max_srq=%d\n",
	   p_hca->ib_caps.max_qp, p_hca->ib_caps.max_qp_wr,
	   p_hca->ib_caps.max_cq, p_hca->ib_caps.max_cqe,
           p_hca->ib_caps.max_srq);
  NM_DISPF("ibverbs-   max_mr=%d; max_mr_size=%llu; page_size_cap=%llu; max_msg_size=%llu\n",
	   p_hca->ib_caps.max_mr,
	   (unsigned long long) p_hca->ib_caps.max_mr_size,
	   (unsigned long long) p_hca->ib_caps.page_size_cap,
	   (unsigned long long) p_hca->ib_caps.max_msg_size);
#endif

  /* allocate Protection Domain */
  p_hca->pd = ibv_alloc_pd(p_hca->context);
  if(p_hca->pd == NULL)
    {
      NM_FATAL("ibverbs: cannot allocate IB protection domain (%s).\n", strerror(errno));
    }
  p_hca->device = strdup(device);
  p_hca->port = port;
  ibv_free_device_list(dev_list);
}

void nm_ibverbs_hca_close(struct nm_ibverbs_hca_s*p_hca)
{
  ibv_dealloc_pd(p_hca->pd);
  ibv_close_device(p_hca->context);
}

void nm_ibverbs_context_get_profile(struct nm_ibverbs_context_s*p_ibverbs_context, struct nm_drv_profile_s*p_profile)
{
  /* driver profile encoding */
#ifdef NMAD_HWLOC
  static hwloc_topology_t __topology = NULL;
  if(__topology == NULL)
    {
      hwloc_topology_init(&__topology);
      hwloc_topology_load(__topology);
    }
  p_profile->cpuset = hwloc_bitmap_alloc();
  int rc = hwloc_ibv_get_device_cpuset(__topology, p_ibverbs_context->hca.ib_dev, p_profile->cpuset);
  if(rc)
    {
      NM_WARN("ibverbs- error while detecting ibv device location.\n");
      hwloc_bitmap_copy(p_profile->cpuset, hwloc_topology_get_complete_cpuset(__topology));
    }
#endif /* NMAD_HWLOC */
  p_profile->latency = 1200; /* from sampling */
  p_profile->bandwidth = 1024 * (p_ibverbs_context->hca.ib_caps.data_rate / 8) * 0.75; /* empirical estimation of software+protocol overhead */
}

/** create QP and both CQs */
struct nm_ibverbs_cnx_s*nm_ibverbs_cnx_create(struct nm_ibverbs_context_s*p_ibverbs_context)
{
  const int use_srq = p_ibverbs_context->ib_opts.use_srq;
  struct nm_ibverbs_cnx_s*p_ibverbs_cnx = malloc(sizeof(struct nm_ibverbs_cnx_s));
  memset(p_ibverbs_cnx, 0, sizeof(struct nm_ibverbs_cnx_s));
  p_ibverbs_cnx->p_ibverbs_context = p_ibverbs_context;
  p_ibverbs_cnx->p_peer_addr_entry = NULL;
  if(use_srq)
    {
      p_ibverbs_cnx->if_cq = NULL;
    }
  else
    {
      /* init inbound CQ */
      p_ibverbs_cnx->if_cq = ibv_create_cq(p_ibverbs_context->hca.context, NM_IBVERBS_RX_DEPTH, NULL, NULL, 0);
      if(p_ibverbs_cnx->if_cq == NULL)
        {
          NM_FATAL("ibverbs: cannot create in CQ (%s).\n", strerror(errno));
        }
    }
  /* init outbound CQ */
  p_ibverbs_cnx->of_cq = ibv_create_cq(p_ibverbs_context->hca.context, NM_IBVERBS_TX_DEPTH, NULL, NULL, 0);
  if(p_ibverbs_cnx->of_cq == NULL)
    {
      NM_FATAL("ibverbs: cannot create out CQ (%s).\n", strerror(errno));
    }
  /* create QP */
  struct ibv_qp_init_attr qp_init_attr = {
    .send_cq = p_ibverbs_cnx->of_cq,
    .recv_cq = use_srq ? p_ibverbs_context->srq.srq_cq : p_ibverbs_cnx->if_cq,
    .srq     = use_srq ? p_ibverbs_context->srq.p_srq : NULL,
    .cap     = {
      .max_send_wr     = NM_IBVERBS_TX_DEPTH,
      .max_recv_wr     = NM_IBVERBS_RX_DEPTH,
      .max_send_sge    = NM_IBVERBS_MAX_SG_SQ,
      .max_recv_sge    = NM_IBVERBS_MAX_SG_RQ,
      .max_inline_data = NM_IBVERBS_MAX_INLINE
    },
    .qp_type = IBV_QPT_RC
  };
  p_ibverbs_cnx->qp = ibv_create_qp(p_ibverbs_context->hca.pd, &qp_init_attr);
  if(p_ibverbs_cnx->qp == NULL)
    {
      NM_FATAL("ibverbs: couldn't create QP (%s)\n", strerror(errno));
    }
  p_ibverbs_cnx->max_inline = qp_init_attr.cap.max_inline_data;
  p_ibverbs_cnx->local_addr.addr.lid = p_ibverbs_context->hca.lid;
  p_ibverbs_cnx->local_addr.addr.qpn = p_ibverbs_cnx->qp->qp_num;
  p_ibverbs_cnx->local_addr.addr.psn = lrand48() & 0xffffff;
  return p_ibverbs_cnx;
}


static inline struct nm_ibverbs_cnx_addr_entry_s*
nm_ibverbs_cnx_addr_lookup(struct nm_ibverbs_context_s*p_ibverbs_context, const struct nm_ibverbs_addr_s*p_remote_addr)
{
  struct nm_ibverbs_cnx_addr_entry_s key;
  memcpy(&key.msg.addr, p_remote_addr, sizeof(struct nm_ibverbs_addr_s));
  struct nm_ibverbs_cnx_addr_entry_s*p_remote_entry = 
        nm_ibverbs_cnx_addr_entry_hashtable_lookup(&p_ibverbs_context->connect.addr_hash, &key);
  return p_remote_entry;
}
                           
static void nm_ibverbs_cnx_addr_send(struct nm_ibverbs_cnx_s*p_ibverbs_cnx, const struct nm_ibverbs_addr_s*p_remote_addr)
{
  struct nm_ibverbs_context_s*p_ibverbs_context = p_ibverbs_cnx->p_ibverbs_context;
  struct nm_ibverbs_cnx_connect_packet_s send_packet;
  send_packet.addr_msg.addr     = p_ibverbs_context->connect.udep.local_addr;
  send_packet.addr_msg.cnx_addr = p_ibverbs_cnx->local_addr;
  send_packet.addr_msg.state    = p_ibverbs_cnx->addr_state;
  nm_ibverbs_ud_send(&p_ibverbs_context->connect.udep, &send_packet, sizeof(struct nm_ibverbs_cnx_addr_msg_s), p_remote_addr);
}

static void nm_ibverbs_cnx_addr_poll(struct nm_ibverbs_context_s*p_ibverbs_context)
{
  int rc = 0;
  do
    {
      rc = nm_ibverbs_ud_recv_pool_poll(&p_ibverbs_context->connect.recv_pool);
      if(rc >= 0)
        {
          assert(rc < NM_IBVERBS_RX_DEPTH);
          struct nm_ibverbs_cnx_addr_msg_s*p_remote_msg = &p_ibverbs_context->connect.recv_packets[rc].addr_msg;
          /* insert received packet into addr hashtable */
          struct nm_ibverbs_cnx_addr_entry_s*p_addr_entry =
            nm_ibverbs_cnx_addr_lookup(p_ibverbs_context, &p_remote_msg->addr);
          if(p_addr_entry == NULL)
            {
              p_addr_entry = malloc(sizeof(struct nm_ibverbs_cnx_addr_entry_s));
              p_addr_entry->p_cnx = NULL;
              p_addr_entry->msg.state = NM_IBVERBS_CNX_ADDR_NONE;
              nm_ibverbs_cnx_addr_entry_hashtable_insert(&p_ibverbs_context->connect.addr_hash, p_addr_entry, p_addr_entry);
            }
          p_addr_entry->msg = *p_remote_msg;
          if(p_remote_msg->state == NM_IBVERBS_CNX_ADDR_TIMEOUT)
            {
              /* send again in case of timeout */
              if((p_addr_entry != NULL) && (p_addr_entry->p_cnx != NULL))
                {
                  nm_ibverbs_cnx_addr_send(p_addr_entry->p_cnx, &p_addr_entry->msg.addr);
                }
            }
          /* rearm receive */
          nm_ibverbs_ud_recv_pool_reload(&p_ibverbs_context->connect.recv_pool, rc);
        }
    }
  while(rc >= 0);
}

static struct nm_ibverbs_cnx_addr_entry_s*
nm_ibverbs_cnx_addr_wait(struct nm_ibverbs_cnx_s*p_ibverbs_cnx, const struct nm_ibverbs_addr_s*p_remote_addr,
                         enum nm_ibverbs_cnx_addr_state_e min_state)
{
  struct nm_ibverbs_context_s*p_ibverbs_context = p_ibverbs_cnx->p_ibverbs_context;
  struct nm_ibverbs_cnx_addr_entry_s*p_remote_entry = NULL;
  puk_tick_t t1, t2;
  PUK_GET_TICK(t1);
  puk_tick_t t_warn_origin = t1;
  double cur_timeout = NM_IBVERBS_TIMEOUT_CONNECT;
  do
    {
      nm_ibverbs_cnx_addr_poll(p_ibverbs_context);
      PUK_GET_TICK(t2);
      const double usec = PUK_TIMING_DELAY(t1, t2);
      if(usec > cur_timeout)
        {
          p_ibverbs_cnx->addr_state = NM_IBVERBS_CNX_ADDR_TIMEOUT;
          nm_ibverbs_cnx_addr_send(p_ibverbs_cnx, p_remote_addr);
          if(PUK_TIMING_DELAY(t_warn_origin, t2) > NM_IBVERBS_TIMEOUT_WARN)
            {
              NM_WARN("ibverbs: timeout for address receive. Retrying...");
              t_warn_origin = t2;
            }
          PUK_GET_TICK(t1);
          if(cur_timeout < NM_IBVERBS_TIMEOUT_WARN)
            cur_timeout *= 2;
        }
      p_remote_entry = nm_ibverbs_cnx_addr_lookup(p_ibverbs_context, p_remote_addr);
    }
  while((!p_remote_entry) || (p_remote_entry->msg.state < min_state));
  return p_remote_entry;
}

static void nm_ibverbs_cnx_ready_wait(struct nm_ibverbs_cnx_s*p_ibverbs_cnx, const struct nm_ibverbs_addr_s*p_remote_addr)
{
  struct nm_ibverbs_context_s*p_ibverbs_context = p_ibverbs_cnx->p_ibverbs_context;
  struct nm_ibverbs_cnx_addr_entry_s*p_remote_entry = nm_ibverbs_cnx_addr_lookup(p_ibverbs_context, p_remote_addr);
  int wait_ready = 1;
  if(p_ibverbs_context->ib_opts.use_srq)
    {
      wait_ready = 0;
    }
  /* post recv HELLO */
  if(wait_ready)
    {
      struct ibv_recv_wr wr;
      wr.wr_id    = NM_IBVERBS_WRID_RECV_HELLO;
      wr.sg_list  = NULL;
      wr.num_sge  = 0;
      wr.next     = NULL;
      struct ibv_recv_wr*bad_wr_recv;
      int rc = ibv_post_recv(p_ibverbs_cnx->qp, &wr, &bad_wr_recv);
      if(rc)
        {
          NM_FATAL("ibverbs: post recv failed (rc = %d; %s).\n", rc, strerror(rc));
        }
    }
  /* switch to state DONE */
  p_ibverbs_cnx->addr_state = NM_IBVERBS_CNX_ADDR_DONE;
  nm_ibverbs_cnx_addr_send(p_ibverbs_cnx, p_remote_addr);

  /* wait for remote DONE */
  int ready = 0;
  puk_tick_t t0, t1, t2;
  PUK_GET_TICK(t0);
  PUK_GET_TICK(t1);
  do
    {
      nm_ibverbs_cnx_addr_poll(p_ibverbs_context);
      if(wait_ready)
        {
          struct ibv_wc wc;
          int ne = ibv_poll_cq(p_ibverbs_cnx->if_cq, 1, &wc);
          if(ne < 0)
            {
              NM_FATAL("ibverbs: poll in CQ failed.\n");
            }
          else if(ne == 1)
            {
              ready = 1;
              break;
            }
        }
      PUK_GET_TICK(t2);
      const double usec = PUK_TIMING_DELAY(t1, t2);
      if(usec > NM_IBVERBS_TIMEOUT_CONNECT)
        {
          /* send again in case DONE msg was lost */
          nm_ibverbs_cnx_addr_send(p_ibverbs_cnx, p_remote_addr);
          NM_WARN("ibverbs: timeout for ready notification. Retrying...");
          PUK_GET_TICK(t1);
        }
      if(!wait_ready)
        {
          const double usec0 = PUK_TIMING_DELAY(t0, t2);
          if(usec0 > NM_IBVERBS_TIMEOUT_READY)
            {
              NM_WARN("ibverbs: timeout for ready notification. Giving up.");
              break;
            }
        }
    }
  while((p_remote_entry->msg.state != NM_IBVERBS_CNX_ADDR_DONE) && (!ready));

  /* post send HELLO */
  if(wait_ready)
    {
      struct ibv_send_wr wr =
        {
          .wr_id      = NM_IBVERBS_WRID_SEND_HELLO,
          .sg_list    = NULL,
          .num_sge    = 0,
          .opcode     = IBV_WR_SEND,
          .send_flags = 0 /* IBV_SEND_SIGNALED */,
          .next       = NULL,
        };
      struct ibv_send_wr*bad_wr;
      int rc = ibv_post_send(p_ibverbs_cnx->qp, &wr, &bad_wr);
      if(rc)
        {
          NM_FATAL("ibverbs: post send failed\n");
        }  
    }
  /* wait for remote HELLO */
  if(wait_ready && !ready)
    {
      do
        {
          nm_ibverbs_cnx_addr_poll(p_ibverbs_context);
          if(wait_ready)
            {
              struct ibv_wc wc;
              int ne = ibv_poll_cq(p_ibverbs_cnx->if_cq, 1, &wc);
              if(ne < 0)
                {
                  NM_FATAL("ibverbs: poll in CQ failed.\n");
                }
              else if(ne == 1)
                {
                  ready = 1;
                  break;
                }
            }
          PUK_GET_TICK(t2);
          const double usec = PUK_TIMING_DELAY(t1, t2);
          if(usec > NM_IBVERBS_TIMEOUT_CONNECT)
            {
              /* send again in case DONE msg was lost */
              nm_ibverbs_cnx_addr_send(p_ibverbs_cnx, p_remote_addr);
              NM_WARN("ibverbs: connection didn't get ready before timeout. Retrying...");
              PUK_GET_TICK(t1);
            }      
        }
      while(!ready);
    }
}

void nm_ibverbs_cnx_connect2(struct nm_ibverbs_cnx_s*p_ibverbs_cnx, const void*p_url, size_t urlsize)
{
  nm_ibverbs_cnx_connect_async(p_ibverbs_cnx, p_url, urlsize);
  nm_ibverbs_cnx_connect_wait(p_ibverbs_cnx);                              
}

void nm_ibverbs_cnx_connect_async(struct nm_ibverbs_cnx_s*p_ibverbs_cnx, const void*p_url, size_t urlsize)
{
  struct nm_ibverbs_context_s*p_ibverbs_context = p_ibverbs_cnx->p_ibverbs_context;
  const struct nm_ibverbs_addr_s*p_remote_addr = p_url;
  assert(urlsize == sizeof(struct nm_ibverbs_addr_s));

  /* register new cnx and ud remote addr in addr hashtable */
  struct nm_ibverbs_cnx_addr_entry_s*p_new_entry = nm_ibverbs_cnx_addr_lookup(p_ibverbs_context, p_remote_addr);
  if(p_new_entry == NULL)
    {
      p_new_entry = malloc(sizeof(struct nm_ibverbs_cnx_addr_entry_s));
      p_new_entry->msg.addr = *p_remote_addr;
      memset(&p_new_entry->msg.cnx_addr, 0, sizeof(struct nm_ibverbs_cnx_addr_s));
      p_new_entry->msg.state = NM_IBVERBS_CNX_ADDR_NONE;
      nm_ibverbs_cnx_addr_entry_hashtable_insert(&p_ibverbs_context->connect.addr_hash, p_new_entry, p_new_entry);
    }
  p_new_entry->p_cnx = p_ibverbs_cnx;
  p_ibverbs_cnx->p_peer_addr_entry = p_new_entry;

  /* send local address */
  p_ibverbs_cnx->addr_state = NM_IBVERBS_CNX_ADDR_OK;
  nm_ibverbs_cnx_addr_send(p_ibverbs_cnx, p_remote_addr);
}

void nm_ibverbs_cnx_connect_wait(struct nm_ibverbs_cnx_s*p_ibverbs_cnx)
{
  const struct nm_ibverbs_addr_s*p_remote_addr = &p_ibverbs_cnx->p_peer_addr_entry->msg.addr;
  /* wait peer address */
  struct nm_ibverbs_cnx_addr_entry_s*p_remote_entry = nm_ibverbs_cnx_addr_wait(p_ibverbs_cnx, p_remote_addr, NM_IBVERBS_CNX_ADDR_OK);
  p_ibverbs_cnx->remote_addr = p_remote_entry->msg.cnx_addr;
  
  /* QP connect sequence */
  nm_ibverbs_cnx_qp_reset(p_ibverbs_cnx);
  nm_ibverbs_cnx_qp_init(p_ibverbs_cnx);
  nm_ibverbs_cnx_qp_rtr(p_ibverbs_cnx);
  nm_ibverbs_cnx_qp_rts(p_ibverbs_cnx);

  /* wait for the connection to be up and ready */
  nm_ibverbs_cnx_ready_wait(p_ibverbs_cnx, p_remote_addr);
}

void nm_ibverbs_cnx_connect(struct nm_ibverbs_cnx_s*p_ibverbs_cnx)
{
  nm_ibverbs_cnx_qp_reset(p_ibverbs_cnx);
  nm_ibverbs_cnx_qp_init(p_ibverbs_cnx);
  nm_ibverbs_cnx_qp_rtr(p_ibverbs_cnx);
  nm_ibverbs_cnx_qp_rts(p_ibverbs_cnx);
}

void nm_ibverbs_cnx_close(struct nm_ibverbs_cnx_s*p_ibverbs_cnx)
{
  ibv_destroy_qp(p_ibverbs_cnx->qp);
  if(p_ibverbs_cnx->if_cq != NULL)
    {
      ibv_destroy_cq(p_ibverbs_cnx->if_cq);
    }
  ibv_destroy_cq(p_ibverbs_cnx->of_cq);
  free(p_ibverbs_cnx);
}

struct ibv_mr*nm_ibverbs_reg_mr(const struct nm_ibverbs_context_s*p_ibverbs_context, void*buf, size_t size,
                                struct nm_ibverbs_segment_s*p_segment)
{
  struct ibv_mr*p_mr = ibv_reg_mr(p_ibverbs_context->hca.pd, buf, size,
                                  IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ | IBV_ACCESS_LOCAL_WRITE);
  if(p_mr == NULL)
    {
      const int err = errno;
      NM_WARN("ibverbs: cannot register MR (errno = %d; %s).\n", err, strerror(err));
      return NULL;
    }
  if(p_segment)
    {
      p_segment->raddr = (uintptr_t)buf;
      p_segment->rkey = p_mr->rkey;
    }
  return p_mr;
}

void nm_ibverbs_cnx_sync(struct nm_ibverbs_cnx_s*p_ibverbs_cnx)
{
  volatile int*buffer = (void*)p_ibverbs_cnx->local_addr.segment.raddr;
  buffer[0] = 0; /* recv buffer */
  buffer[1] = 1; /* send buffer */

  for(;;)
    {
      /* check connection */
      int rc = nm_ibverbs_sync_send((void*)&buffer[1], (void*)&buffer[0], sizeof(int), p_ibverbs_cnx);
      if(rc)
	{
	  NM_WARN("ibverbs: connection failed to come up before RNR timeout; reset QP and try again.\n");
	  /* nm_ibverbs_connect_send(p_ibverbs_drv, remote_url, trk_id, &p_ibverbs_cnx->local_addr.addr, 1); */
	  nm_ibverbs_cnx_qp_reset(p_ibverbs_cnx);
	  nm_ibverbs_cnx_qp_init(p_ibverbs_cnx);
	  nm_ibverbs_cnx_qp_rtr(p_ibverbs_cnx);
	  nm_ibverbs_cnx_qp_rts(p_ibverbs_cnx);
	}
      else if(buffer[0] == 0)
	{
	  puk_tick_t t1, t2;
	  PUK_GET_TICK(t1);
	  do
	    {
	      PUK_GET_TICK(t2);
	    }
	  while(buffer[0] == 0 && PUK_TIMING_DELAY(t1, t2) < NM_IBVERBS_TIMEOUT_CHECK);
	  if(buffer[0] == 0)
	    {
	      /*  nm_ibverbs_connect_send(p_ibverbs_drv, remote_url, trk_id, &p_ibverbs_cnx->local_addr.addr, 1); */
	    }
	}
      else
	{
	  /* connection ok- escape from loop */
	  break;
	}
    }
}


/* ********************************************************* */
/* ** state transitions for QP finite-state automaton */

/** modify QP to state RESET */
static void nm_ibverbs_cnx_qp_reset(struct nm_ibverbs_cnx_s*p_ibverbs_cnx)
{
  /* modify QP- step: RTS */
  struct ibv_qp_attr attr =
    {
      .qp_state = IBV_QPS_RESET
    };
  int rc = ibv_modify_qp(p_ibverbs_cnx->qp, &attr, IBV_QP_STATE);
  if(rc != 0)
    {
      NM_FATAL("ibverbs: failed to modify QP to RESET (rc=%d; %s).\n", rc, strerror(rc));
    }
}

/** modify QP to state INIT */
static void nm_ibverbs_cnx_qp_init(struct nm_ibverbs_cnx_s*p_ibverbs_cnx)
{
  struct ibv_qp_attr attr =
    {
      .qp_state        = IBV_QPS_INIT,
      .pkey_index      = 0,
      .port_num        = p_ibverbs_cnx->p_ibverbs_context->hca.port,
      .qp_access_flags = IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ
    };
  int rc = ibv_modify_qp(p_ibverbs_cnx->qp, &attr,
			 IBV_QP_STATE              |
			 IBV_QP_PKEY_INDEX         |
			 IBV_QP_PORT               |
			 IBV_QP_ACCESS_FLAGS);
  if(rc != 0)
    {
      NM_FATAL("ibverbs: failed to modify QP to INIT (rc=%d; %s).\n", rc, strerror(rc));
    }
}

/** modify QP to state RTR */
static void nm_ibverbs_cnx_qp_rtr(struct nm_ibverbs_cnx_s*p_ibverbs_cnx)
{
  struct ibv_qp_attr attr =
    {
      .qp_state           = IBV_QPS_RTR,
      .path_mtu           = p_ibverbs_cnx->p_ibverbs_context->ib_opts.mtu,
      .dest_qp_num        = p_ibverbs_cnx->remote_addr.addr.qpn,
      .rq_psn             = p_ibverbs_cnx->remote_addr.addr.psn,
      .max_dest_rd_atomic = NM_IBVERBS_RDMA_DEPTH,
      .min_rnr_timer      = 12, /* 12 */
      .ah_attr            = {
	.is_global        = 0,
	.dlid             = p_ibverbs_cnx->remote_addr.addr.lid,
	.sl               = 0,
	.src_path_bits    = 0,
	.port_num         =  p_ibverbs_cnx->p_ibverbs_context->hca.port
      }
    };
  int rc = ibv_modify_qp(p_ibverbs_cnx->qp, &attr,
			 IBV_QP_STATE              |
			 IBV_QP_AV                 |
			 IBV_QP_PATH_MTU           |
			 IBV_QP_DEST_QPN           |
			 IBV_QP_RQ_PSN             |
			 IBV_QP_MAX_DEST_RD_ATOMIC |
			 IBV_QP_MIN_RNR_TIMER);
  if(rc != 0)
    {
      NM_FATAL("ibverbs: dev = %s; port = %d; failed to modify QP to RTR (%s)\n",
               ibv_get_device_name(p_ibverbs_cnx->p_ibverbs_context->hca.ib_dev),
               p_ibverbs_cnx->p_ibverbs_context->hca.port, strerror(rc));
    }
}

/** modify QP to state RTS */
static void nm_ibverbs_cnx_qp_rts(struct nm_ibverbs_cnx_s*p_ibverbs_cnx)
{
  struct ibv_qp_attr attr =
    {
      .qp_state      = IBV_QPS_RTS,
      .timeout       = 2, /* 14 */
      .retry_cnt     = 7,  /* 7 */
      .rnr_retry     = 7,  /* 7 = unlimited */
      .sq_psn        = p_ibverbs_cnx->local_addr.addr.psn,
      .max_rd_atomic = NM_IBVERBS_RDMA_DEPTH /* 1 */
    };
  int rc = ibv_modify_qp(p_ibverbs_cnx->qp, &attr,
			 IBV_QP_STATE              |
			 IBV_QP_TIMEOUT            |
			 IBV_QP_RETRY_CNT          |
			 IBV_QP_RNR_RETRY          |
			 IBV_QP_SQ_PSN             |
			 IBV_QP_MAX_QP_RD_ATOMIC);
  if(rc != 0)
    {
      NM_FATAL("ibverbs: failed to modify QP to RTS (rc=%d; %s)\n", rc, strerror(rc));
    }
}

/** RDMA used to synchronize connection establishment
 */
static int nm_ibverbs_sync_send(const void*sbuf, const void*rbuf, int size, struct nm_ibverbs_cnx_s*p_ibverbs_cnx)
{
  const int roffset = ((uintptr_t)rbuf) - p_ibverbs_cnx->local_addr.segment.raddr;
  assert(rbuf >= (void*)p_ibverbs_cnx->local_addr.segment.raddr);
  struct ibv_sge list = {
    .addr   = (uintptr_t)sbuf,
    .length = size,
    .lkey   = p_ibverbs_cnx->local_addr.segment.rkey
  };
  struct ibv_send_wr wr = {
    .wr_id      = NM_IBVERBS_WRID_SYNC,
    .sg_list    = &list,
    .num_sge    = 1,
    .opcode     = IBV_WR_RDMA_WRITE,
    .send_flags = IBV_SEND_SIGNALED | IBV_SEND_INLINE,
    .next       = NULL,
    .imm_data   = 0,
    .wr.rdma =
    {
      .remote_addr = p_ibverbs_cnx->remote_addr.segment.raddr + roffset,
      .rkey        = p_ibverbs_cnx->remote_addr.segment.rkey
    }
  };
  struct ibv_send_wr*bad_wr = NULL;
  int rc = ibv_post_send(p_ibverbs_cnx->qp, &wr, &bad_wr);
  if(rc)
    {
      NM_FATAL("ibverbs: post sync send failed.\n");
    }
  struct ibv_wc wc;
  int ne = 0;
  do
    {
      ne = ibv_poll_cq(p_ibverbs_cnx->of_cq, 1, &wc);
      if(ne < 0)
	{
	  NM_FATAL("ibverbs: poll out CQ failed.\n");
	  abort();
	}
    }
  while(ne == 0);
  if(ne != 1 || wc.status != IBV_WC_SUCCESS)
    {
      NM_WARN("ibverbs: WC send failed (status=%d; %s)\n",
	      wc.status, nm_ibverbs_status_strings[wc.status]);
      return 1;
    }
  return 0;
}



/* ** checksum ********************************************* */


/** checksum algorithm. Set NMAD_IBVERBS_CHECKSUM to non-null to enable checksums.
 */
uint32_t nm_ibverbs_checksum(const char*data, nm_len_t len)
{
  if(_nm_ibverbs_checksum)
    return (*_nm_ibverbs_checksum->func)(data, len);
  else
    return 0;
}

int nm_ibverbs_checksum_enabled(void)
{
  return _nm_ibverbs_checksum != NULL;
}

uint32_t nm_ibverbs_memcpy_and_checksum(void*_dest, const void*_src, nm_len_t len)
{
  if(_nm_ibverbs_checksum)
    {
      if(_nm_ibverbs_checksum->checksum_and_copy)
	{
	  return (*_nm_ibverbs_checksum->checksum_and_copy)(_dest, _src, len);
	}
      else
	{
	  memcpy(_dest, _src, len);
	  return nm_ibverbs_checksum(_dest, len);
	}
    }
  else
    {
      memcpy(_dest, _src, len);
    }
  return 1;
}

uint32_t nm_ibverbs_copy_from_and_checksum(void*dest, nm_data_slicer_t*p_slicer, const void*src, nm_len_t offset, nm_len_t len)
{
  assert(!(nm_data_slicer_isnull(p_slicer) && (src == NULL)));
  assert(!(!nm_data_slicer_isnull(p_slicer) && (src != NULL)));
  if(_nm_ibverbs_checksum && !nm_data_slicer_isnull(p_slicer))
    {
      NM_FATAL("ibverbs: checksums not supported yet with nm_data.\n");
    }
  if(!nm_data_slicer_isnull(p_slicer))
    {
      nm_data_slicer_copy_from(p_slicer, dest, len);
      return 1;
    }
  else
    {
      return nm_ibverbs_memcpy_and_checksum(dest, src + offset, len);
    }
}

uint32_t nm_ibverbs_copy_to_and_checksum(const void*src, nm_data_slicer_t*p_slicer, void*dest, nm_len_t offset, nm_len_t len)
{
  assert(!(nm_data_slicer_isnull(p_slicer) && (dest == NULL)));
  assert(!((!nm_data_slicer_isnull(p_slicer)) && (dest != NULL)));
  if(_nm_ibverbs_checksum && (p_slicer->p_data != NULL))
    {
      NM_FATAL("ibverbs: checksums not supported yet with nm_data.\n");
    }
  if(!nm_data_slicer_isnull(p_slicer))
    {
      nm_data_slicer_copy_to(p_slicer, src, len);
      return 1;
    }
  else
    {
      return nm_ibverbs_memcpy_and_checksum(dest + offset, src, len);
    }
}


/* ** IB UD ************************************************ */

void nm_ibverbs_ud_ep_open(struct nm_ibverbs_ud_ep_s*p_ud_ep, struct nm_ibverbs_hca_s*p_hca)
{
  p_ud_ep->p_hca = p_hca;

  /* init inbound CQ */
  p_ud_ep->if_cq = ibv_create_cq(p_ud_ep->p_hca->context, NM_IBVERBS_RX_DEPTH, NULL, NULL, 0);
  if(p_ud_ep->if_cq == NULL)
    {
      NM_FATAL("ibverbs: cannot create in CQ (%s).\n", strerror(errno));
    }
  /* init outbound CQ */
  p_ud_ep->of_cq = ibv_create_cq(p_ud_ep->p_hca->context, NM_IBVERBS_TX_DEPTH, NULL, NULL, 0);
  if(p_ud_ep->of_cq == NULL)
    {
      NM_FATAL("ibverbs: cannot create out CQ (%s).\n", strerror(errno));
    }
  /* create QP */
  struct ibv_qp_init_attr qp_init_attr =
    {
      .send_cq = p_ud_ep->of_cq,
      .recv_cq = p_ud_ep->if_cq,
      .srq     = NULL,
      .cap     = {
        .max_send_wr     = NM_IBVERBS_TX_DEPTH,
        .max_recv_wr     = NM_IBVERBS_RX_DEPTH,
        .max_send_sge    = NM_IBVERBS_MAX_SG_SQ,
        .max_recv_sge    = NM_IBVERBS_MAX_SG_RQ,
        .max_inline_data = NM_IBVERBS_MAX_INLINE
      },
      .qp_type = IBV_QPT_UD
    };
  p_ud_ep->qp = ibv_create_qp(p_ud_ep->p_hca->pd, &qp_init_attr);
  if(p_ud_ep->qp == NULL)
    {
      NM_FATAL("ibverbs: couldn't create UD QP (%s)\n", strerror(errno));
    }
  /* get local address */
  memset(&p_ud_ep->local_addr, 0, sizeof(struct nm_ibverbs_addr_s)); /* clear struct padding, if any */
  p_ud_ep->local_addr.lid = p_ud_ep->p_hca->lid;
  p_ud_ep->local_addr.qpn = p_ud_ep->qp->qp_num;
  p_ud_ep->local_addr.psn = lrand48() & 0xffffff;

  NM_DISPF("ibverbs: IB UD create- lid = %d; qpn = %d; psn = %d (%d)\n",
           p_ud_ep->local_addr.lid, p_ud_ep->local_addr.qpn, p_ud_ep->local_addr.psn,
           (int)sizeof(struct nm_ibverbs_addr_s));
    
  /* modify QP to state RESET */
  struct ibv_qp_attr attr_qp_reset =
    {
      .qp_state      = IBV_QPS_RESET
    };
  int rc = ibv_modify_qp(p_ud_ep->qp, &attr_qp_reset,
                         IBV_QP_STATE);
  if(rc != 0)
    {
      NM_FATAL("ibverbs: failed to modify QP to RESET (rc=%d; %s).\n", rc, strerror(rc));
    }
  /* modify QP to state INIT */
  struct ibv_qp_attr attr_qp_init =
    {
      .qp_state      = IBV_QPS_INIT,
      .pkey_index    = 0,
      .port_num      = p_ud_ep->p_hca->port,
      .qkey          = 0x11111111
    };
  rc = ibv_modify_qp(p_ud_ep->qp, &attr_qp_init,
                     IBV_QP_STATE              |
                     IBV_QP_PKEY_INDEX         |
                     IBV_QP_PORT               |
                     IBV_QP_QKEY);
  if(rc != 0)
    {
      NM_FATAL("ibverbs: failed to modify QP to INIT (rc=%d; %s).\n", rc, strerror(rc));
    }
  /* modify QP to state RTR */
  struct ibv_qp_attr attr_qp_rtr =
    {
      .qp_state      = IBV_QPS_RTR
    };
  rc = ibv_modify_qp(p_ud_ep->qp, &attr_qp_rtr,
                     IBV_QP_STATE);
  if(rc != 0)
    {
      NM_FATAL("ibverbs: dev = %s; port = %d; failed to modify QP to RTR (%s)\n",
               ibv_get_device_name(p_ud_ep->p_hca->ib_dev),
               p_ud_ep->p_hca->port, strerror(rc));
    }
  /* modify QP to state RTS */
  struct ibv_qp_attr attr_qp_rts =
    {
      .qp_state      = IBV_QPS_RTS,
      .sq_psn        = p_ud_ep->local_addr.psn
    };
  rc = ibv_modify_qp(p_ud_ep->qp, &attr_qp_rts,
                     IBV_QP_STATE              |
                     IBV_QP_SQ_PSN);
  if(rc != 0)
    {
      NM_FATAL("ibverbs: failed to modify QP to RTS (rc=%d; %s)\n", rc, strerror(rc));
    }       
}

void nm_ibverbs_ud_ep_close(struct nm_ibverbs_ud_ep_s*p_ud_ep)
{
  /* TODO- close QP & CQs */
}

void nm_ibverbs_ud_send_post(struct nm_ibverbs_ud_send_s*p_send,
                             struct nm_ibverbs_ud_ep_s*p_ud_ep, void*buf, size_t size,
                             const struct nm_ibverbs_addr_s*p_remote_addr)
{
  p_send->p_ud_ep = p_ud_ep;
  /* create AH */
  struct ibv_ah_attr ah_attr =
    {
     .is_global      = 0,
     .dlid           = p_remote_addr->lid,
     .sl             = 0,
     .src_path_bits  = 0,
     .port_num       = p_ud_ep->p_hca->port
    };
  p_send->ah = ibv_create_ah(p_ud_ep->p_hca->pd, &ah_attr);
  if(!p_send->ah)
    {
      NM_FATAL("ibverbs: failed to create AH for UD send (%s)\n", strerror(errno));
    }
  p_send->mr = ibv_reg_mr(p_ud_ep->p_hca->pd, buf, size + GRH_OFFSET,
                          IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE);
  if(!p_send->mr)
    {
      NM_FATAL("ibverbs: couldn't allocate MR (%s)\n", strerror(errno));
    }
  /* create WR & post */
  struct ibv_sge list =
    {
      .addr    = (uintptr_t) buf + GRH_OFFSET,
      .length  = size,
      .lkey    = p_send->mr->lkey
    };
  struct ibv_send_wr wr =
    {
      .wr_id      = NM_IBVERBS_WRID_UD_SEND,
      .sg_list    = &list,
      .num_sge    = 1,
      .opcode     = IBV_WR_SEND,
      .next       = NULL,
      .send_flags = IBV_SEND_SIGNALED,
      .wr.ud.ah          = p_send->ah,
      .wr.ud.remote_qpn  = p_remote_addr->qpn,
      .wr.ud.remote_qkey = 0x11111111
    };
  struct ibv_send_wr*bad_wr;
  int rc = ibv_post_send(p_ud_ep->qp, &wr, &bad_wr);
  if(rc)
    {
      NM_FATAL("ibverbs: post send failed (%d; %s)\n", rc, strerror(rc));
    }  
}

int nm_ibverbs_ud_send_poll(struct nm_ibverbs_ud_send_s*p_send)
{
  /* poll on CQ */
  struct ibv_wc wc;
  int ne = ibv_poll_cq(p_send->p_ud_ep->of_cq, 1, &wc);
  if(ne < 0)
    {
      NM_FATAL("ibverbs: poll out CQ failed.\n");
    }
  if(ne == 0)
    {
      return -1;
    }
  if(wc.status != IBV_WC_SUCCESS)
    {
      NM_FATAL("ibverbs: WC send failed (status=%d; %s)\n",
	      wc.status, nm_ibverbs_status_strings[wc.status]);
    }
  ibv_destroy_ah(p_send->ah);
  ibv_dereg_mr(p_send->mr);
  return 0;
}

void nm_ibverbs_ud_send(struct nm_ibverbs_ud_ep_s*p_ud_ep, void*buf, size_t size,
                        const struct nm_ibverbs_addr_s*p_remote_addr)
{
  struct nm_ibverbs_ud_send_s send;
  nm_ibverbs_ud_send_post(&send, p_ud_ep, buf, size, p_remote_addr);
  int rc = -1;
  do
    {
      rc = nm_ibverbs_ud_send_poll(&send);
    }
  while(rc != 0);  
}

void nm_ibverbs_ud_recv_post(struct nm_ibverbs_ud_recv_s*p_recv,
                             struct nm_ibverbs_ud_ep_s*p_ud_ep, void*buf, size_t size)
{
  p_recv->p_ud_ep = p_ud_ep;
  p_recv->mr = ibv_reg_mr(p_ud_ep->p_hca->pd, buf, size + GRH_OFFSET,
                          IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE);
  if(!p_recv->mr)
    {
      NM_FATAL("ibverbs: couldn't allocate MR\n");
    }
  /* create WR & post */
  struct ibv_sge list =
    {
      .addr   = (uintptr_t) buf,
      .length = size + GRH_OFFSET,
      .lkey   = p_recv->mr->lkey
    };
  struct ibv_recv_wr wr =
    {
      .wr_id    = NM_IBVERBS_WRID_UD_RECV,
      .sg_list  = &list,
      .num_sge  = 1,
      .next     = NULL
    };
  struct ibv_recv_wr*bad_wr_recv;
  int rc = ibv_post_recv(p_recv->p_ud_ep->qp, &wr, &bad_wr_recv);
  if(rc)
    {
      NM_FATAL("ibverbs: post recv failed.\n");
    }
}

int nm_ibverbs_ud_recv_poll(struct nm_ibverbs_ud_recv_s*p_recv)
{
  /* poll on CQ */
  struct ibv_wc wc;
  int ne = ibv_poll_cq(p_recv->p_ud_ep->if_cq, 1, &wc);
  if(ne < 0)
    {
      NM_FATAL("ibverbs: poll in CQ failed.\n");
    }
  if(ne == 0)
    {
      return -1;
    }
  if(ne != 1 || wc.status != IBV_WC_SUCCESS)
    {
      NM_FATAL("ibverbs: WC recv failed (status=%d; %s)\n",
	      wc.status, nm_ibverbs_status_strings[wc.status]);
    }
  ibv_dereg_mr(p_recv->mr);
  return 0;
}

void nm_ibverbs_ud_recv(struct nm_ibverbs_ud_ep_s*p_ud_ep, void*buf, size_t size)
{
  struct nm_ibverbs_ud_recv_s recv;
  nm_ibverbs_ud_recv_post(&recv, p_ud_ep, buf, size);
  int rc = -1;
  do
    {
      rc = nm_ibverbs_ud_recv_poll(&recv);
    }
  while(rc != 0);  
}

void nm_ibverbs_ud_recv_pool_post(struct nm_ibverbs_ud_recv_pool_s*p_recv,
                                  struct nm_ibverbs_ud_ep_s*p_ud_ep, void*buf, size_t size, int depth)
{
  p_recv->p_ud_ep = p_ud_ep;
  p_recv->buf = buf;
  p_recv->size = size;
  p_recv->depth = depth;
  p_recv->mr = ibv_reg_mr(p_ud_ep->p_hca->pd, p_recv->buf, depth * (size + GRH_OFFSET),
                          IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE);
  if(!p_recv->mr)
    {
      NM_FATAL("ibverbs: couldn't allocate MR\n");
    }
  /* create WR & post */
  assert(depth < NM_IBVERBS_WRID_UD_RECV_POOL_LAST - NM_IBVERBS_WRID_UD_RECV_POOL_FIRST);
  int i;
  for(i = 0; i < depth; i++)
    {
      nm_ibverbs_ud_recv_pool_reload(p_recv, i);
    }
}

int nm_ibverbs_ud_recv_pool_poll(struct nm_ibverbs_ud_recv_pool_s*p_recv)
{
  /* poll on CQ */
  struct ibv_wc wc = { 0 };
  int ne = ibv_poll_cq(p_recv->p_ud_ep->if_cq, 1, &wc);
  if(ne < 0)
    {
      NM_FATAL("ibverbs: poll in CQ failed.\n");
    }
  if(ne == 0)
    {
      return -1;
    }
  if(ne != 1 || wc.status != IBV_WC_SUCCESS)
    {
      NM_FATAL("ibverbs: WC recv failed (status=%d; %s)\n",
	      wc.status, nm_ibverbs_status_strings[wc.status]);
    }
  const int i = wc.wr_id - NM_IBVERBS_WRID_UD_RECV_POOL_FIRST;
  return i;
}

void nm_ibverbs_ud_recv_pool_reload(struct nm_ibverbs_ud_recv_pool_s*p_recv, int i)
{
  struct ibv_sge list =
    {
      .addr   = (uintptr_t)(p_recv->buf + i * (p_recv->size + GRH_OFFSET)),
      .length = p_recv->size + GRH_OFFSET,
      .lkey   = p_recv->mr->lkey
    };
  struct ibv_recv_wr wr =
    {
      .wr_id    = NM_IBVERBS_WRID_UD_RECV_POOL_FIRST + i,
      .sg_list  = &list,
      .num_sge  = 1,
      .next     = NULL
    };
  struct ibv_recv_wr*bad_wr_recv;
  int rc = ibv_post_recv(p_recv->p_ud_ep->qp, &wr, &bad_wr_recv);
  if(rc)
    {
      NM_FATAL("ibverbs: post recv failed.\n");
    }
}

void nm_ibverbs_ud_recv_pool_delete(struct nm_ibverbs_ud_recv_pool_s*p_recv)
{
  /* TODO: send to self to cancel all pre-posted requests before we can dereg memory */
  if(p_recv->mr)
    {
      ibv_dereg_mr(p_recv->mr);
    }
  p_recv->mr = NULL;
}
