/*
 * NewMadeleine
 * Copyright (C) 2006-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/* -*- Mode: C; tab-width: 2; c-basic-offset: 2 -*- */

#include "nm_ibverbs.h"
#include "nm_ibverbs_rcache.h"

#include <Padico/Module.h>


/* *** method: 'rcache' ********************************* */

/** Header sent by the receiver to ask for RDMA data (rcache rdv) */
struct nm_ibverbs_rcache_rdvhdr_s
{
  uint64_t raddr;
  uint32_t rkey;
  uint32_t seq;
  volatile int busy;
};

/** Header sent to signal presence of rcache data */
struct nm_ibverbs_rcache_sighdr_s
{
  uint32_t head_offset;
#ifdef DEBUG
  uint32_t crc;
#endif /* DEBUG */
  uint32_t seq;
  volatile int busy;
};

#define NM_IBVERBS_RCACHE_ALIGN 64

#define NM_IBVERBS_RCACHE_SIGPAYLOAD 64

struct nm_ibverbs_rcache_sigpkt_s
{
  char head[NM_IBVERBS_RCACHE_SIGPAYLOAD];
  struct nm_ibverbs_rcache_sighdr_s sighdr;
};

/** context for ibverbs 'rcache' */
struct nm_ibverbs_rcache_context_s
{
  struct nm_ibverbs_context_s*p_ibverbs_context;
  const struct nm_ibverbs_reg_driver_s*p_reg_driver;
  enum
    {
     NM_IBVERBS_RCACHE_WRITE = 1, 
     NM_IBVERBS_RCACHE_READ
    } mode;
#ifdef DEBUG
  int do_crc;
  puk_checksum_func_t p_crc_func;
#endif
};

/** Connection state for 'rcache' tracks */
struct nm_ibverbs_rcache_s
{
  struct ibv_mr*mr;                /**< global MR (used for headers) */
  struct nm_ibverbs_segment_s seg; /**< remote segment */
  struct nm_ibverbs_cnx_s*p_cnx;
  struct nm_ibverbs_rcache_context_s*p_rcache_context;
  puk_context_t context;
  struct
  {
    char*message;
    int size;
    void*p_reg_handle;
    struct ibv_mr*mr;
    int seq;
  } recv;
  struct
  {
    const char*message;
    int size;
    void*p_reg_handle;
    struct ibv_mr*mr;
#ifdef DEBUG
    uint32_t crc;
#endif /* DEBUG */
    int seq;
  } send;
  struct
  {
    struct nm_ibverbs_rcache_rdvhdr_s shdr, rhdr;
    struct nm_ibverbs_rcache_sigpkt_s ssigpkt, rsigpkt;
  } headers;
};

static void nm_ibverbs_rcache_getprops(puk_context_t context, struct nm_minidriver_properties_s*p_props);
static void nm_ibverbs_rcache_init(puk_context_t context, const void**p_url, size_t*p_url_size);
static void nm_ibverbs_rcache_close(puk_context_t context);
static void nm_ibverbs_rcache_connect(void*_status, const void*remote_url, size_t url_size);
static void nm_ibverbs_rcache_connect_async(void*_status, const void*remote_url, size_t url_size);
static void nm_ibverbs_rcache_connect_wait(void*_status);
static int  nm_ibverbs_rcache_send_iov_post(void*_status, const struct iovec*v, int n);
static int  nm_ibverbs_rcache_send_poll(void*_status);
static int  nm_ibverbs_rcache_recv_init(void*_status, struct iovec*v, int n);
static int  nm_ibverbs_rcache_poll_one(void*_status);

static const struct nm_minidriver_iface_s nm_ibverbs_rcache_minidriver =
  {
    .getprops        = &nm_ibverbs_rcache_getprops,
    .init            = &nm_ibverbs_rcache_init,
    .close           = &nm_ibverbs_rcache_close,
    .connect         = &nm_ibverbs_rcache_connect,
    .connect_async   = &nm_ibverbs_rcache_connect_async,
    .connect_wait    = &nm_ibverbs_rcache_connect_wait,
    .disconnect      = NULL,
    .send_data_post  = NULL,
    .send_iov_post   = &nm_ibverbs_rcache_send_iov_post,
    .send_poll       = &nm_ibverbs_rcache_send_poll,
    .recv_iov_post   = &nm_ibverbs_rcache_recv_init,
    .recv_data_post  = NULL,
    .recv_poll_one   = &nm_ibverbs_rcache_poll_one,
    .recv_cancel     = NULL
  };

static void*nm_ibverbs_rcache_instantiate(puk_instance_t instance, puk_context_t context);
static void nm_ibverbs_rcache_destroy(void*);

static const struct puk_component_driver_s nm_ibverbs_rcache_component =
  {
    .instantiate = &nm_ibverbs_rcache_instantiate,
    .destroy = &nm_ibverbs_rcache_destroy
  };


PADICO_MODULE_COMPONENT(NewMad_ibverbs_rcache,
  puk_component_declare("NewMad_ibverbs_rcache",
			puk_component_provides("PadicoComponent", "component", &nm_ibverbs_rcache_component),
			puk_component_provides("NewMad_minidriver", "minidriver", &nm_ibverbs_rcache_minidriver),
                        puk_component_uses("NewMad_ibverbs", "ibverbs")));

static void* nm_ibverbs_rcache_instantiate(puk_instance_t instance, puk_context_t context)
{
  struct nm_ibverbs_rcache_s*p_rcache = malloc(sizeof(struct nm_ibverbs_rcache_s));
  memset(&p_rcache->headers, 0, sizeof(p_rcache->headers));
  memset(&p_rcache->send, 0, sizeof(p_rcache->send));
  memset(&p_rcache->recv, 0, sizeof(p_rcache->recv));
  p_rcache->mr  = NULL;
  p_rcache->p_cnx = NULL;
  p_rcache->p_rcache_context = NULL;
  p_rcache->context = context;
  return p_rcache;
}

static void nm_ibverbs_rcache_destroy(void*_status)
{
  struct nm_ibverbs_rcache_s*p_rcache = _status;
  free(p_rcache);
}

/* *** rcache connection *********************************** */

static void nm_ibverbs_rcache_getprops(puk_context_t p_context, struct nm_minidriver_properties_s*p_props)
{
  struct nm_ibverbs_rcache_context_s*p_rcache_context = malloc(sizeof(struct nm_ibverbs_rcache_context_s));
  puk_context_set_status(p_context, p_rcache_context);
  p_rcache_context->mode = NM_IBVERBS_RCACHE_WRITE;
#ifdef DEBUG
  p_rcache_context->do_crc = puk_opt_parse_bool(getenv("NMAD_RCACHE_CHECKSUM"));
  if(p_rcache_context->do_crc)
    {
      puk_checksum_t p_checksum = puk_checksum_get("default");
      p_rcache_context->p_crc_func = p_checksum->func;
      NM_WARN("ibrcache: checksum enabled (%s).\n", p_checksum->name);
    }
#else
  if(puk_opt_parse_bool(getenv("NMAD_RCACHE_CHECKSUM")))
    {
      NM_WARN("ibrcache: cannot use option NMAD_RCACHE_CHECKSUM in non-debug build.\n");
    }
#endif /* DEBUG */
  struct nm_ibverbs_opts_s ib_opts;
  nm_ibverbs_context_default_opts(&ib_opts); /* get default options, then add srq & comp_channel */
  ib_opts.mtu = IBV_MTU_4096; /* enforce large MTU for large messages */
  p_rcache_context->p_ibverbs_context = nm_ibverbs_context_resolve_opts(p_context, &ib_opts);
  nm_ibverbs_context_get_profile(p_rcache_context->p_ibverbs_context, &p_props->profile);
  p_props->capabilities.supports_data = 0;
  p_props->capabilities.supports_iovec = 0;
  p_props->capabilities.trk_rdv = 1;
  p_props->capabilities.max_msg_size = p_rcache_context->p_ibverbs_context->hca.ib_caps.max_msg_size;
}

static void nm_ibverbs_rcache_init(puk_context_t context, const void**p_url, size_t*p_url_size)
{
  struct nm_ibverbs_rcache_context_s*p_rcache_context = puk_context_get_status(context);
  nm_ibverbs_context_get_url(p_rcache_context->p_ibverbs_context, p_url, p_url_size);

  if(puk_opt_parse_bool(getenv("NMAD_RCACHE_NOCACHE")))
    {
      p_rcache_context->p_reg_driver = &nm_ibverbs_reg_driver_nocache;
    }
  else if(puk_opt_parse_bool(getenv("NMAD_RCACHE_ODP")))
    {
#if defined(NMAD_IBVERBS_ODP)
      p_rcache_context->p_reg_driver = &nm_ibverbs_reg_driver_odp;
#else
      NM_FATAL("ibverbs: odp not supported.\n");
#endif
    }
  else if(puk_opt_parse_bool(getenv("NMAD_RCACHE_IODP")))
    {
#if defined(NMAD_IBVERBS_IODP)
      p_rcache_context->p_reg_driver = &nm_ibverbs_reg_driver_iodp;
#else
      NM_FATAL("ibverbs: iodp not supported.\n");
#endif
    }
  else
    {
      p_rcache_context->p_reg_driver = &nm_ibverbs_reg_driver_mini;
#if 0
#if defined(PUKABI) || defined(MINI_PUKABI)
      p_rcache_context->p_reg_driver = &nm_ibverbs_reg_driver_puk;
#endif
#endif
    }

  if(p_rcache_context->p_reg_driver->init)
    {
      (*p_rcache_context->p_reg_driver->init)(p_rcache_context->p_ibverbs_context);
    }
}

static void nm_ibverbs_rcache_close(puk_context_t context)
{
  struct nm_ibverbs_rcache_context_s*p_rcache_context = puk_context_get_status(context);
  if(p_rcache_context->p_reg_driver->close)
    {
      (*p_rcache_context->p_reg_driver->close)(p_rcache_context->p_ibverbs_context);
    }
}

static void nm_ibverbs_rcache_connect(void*_status, const void*p_remote_url, size_t url_size)
{
  nm_ibverbs_rcache_connect_async(_status, p_remote_url, url_size);
  nm_ibverbs_rcache_connect_wait(_status);
}

static void nm_ibverbs_rcache_connect_async(void*_status, const void*p_remote_url, size_t url_size)
{
  struct nm_ibverbs_rcache_s*p_rcache = _status; 
  struct nm_ibverbs_rcache_context_s*p_rcache_context = puk_context_get_status(p_rcache->context);
  struct nm_ibverbs_cnx_s*p_ibverbs_cnx = nm_ibverbs_cnx_create(p_rcache_context->p_ibverbs_context);
  p_rcache->p_cnx = p_ibverbs_cnx;
  p_rcache->p_rcache_context = p_rcache_context;
  /* register Memory Region */
  p_rcache->mr = nm_ibverbs_reg_mr(p_rcache_context->p_ibverbs_context,
                                   &p_rcache->headers, sizeof(p_rcache->headers),
                                   &p_ibverbs_cnx->local_addr.segment);
  nm_ibverbs_cnx_connect_async(p_ibverbs_cnx, p_remote_url, url_size);
}

static void nm_ibverbs_rcache_connect_wait(void*_status)
{
  struct nm_ibverbs_rcache_s*p_rcache = _status; 
  nm_ibverbs_cnx_connect_wait(p_rcache->p_cnx);
  p_rcache->seg = p_rcache->p_cnx->remote_addr.segment;
}


/* ** reg cache I/O **************************************** */

static int nm_ibverbs_rcache_send_iov_post(void*_status, const struct iovec*v, int n)
{
  struct nm_ibverbs_rcache_s*p_rcache = _status;
  char*message = v[0].iov_base;
  const int size = v[0].iov_len;
  assert(n == 1);
  assert(size > 2 * NM_IBVERBS_RCACHE_ALIGN);
  if(p_rcache->send.message != NULL)
    {
      NM_FATAL("ibverbs: rendez-vous already posted on sender side.\n");
    }
  p_rcache->send.message = message;
  p_rcache->send.size = size;
#ifdef DEBUG
  if(p_rcache->p_rcache_context->do_crc)
    {
      p_rcache->send.crc = (*p_rcache->p_rcache_context->p_crc_func)(message, size);
    }
#endif /* DEBUG */
  p_rcache->send.seq++;
  NM_RCACHE_LOG("rcache = %p; seq = %d\n", p_rcache, p_rcache->send.seq);

  if(p_rcache->p_rcache_context->mode == NM_IBVERBS_RCACHE_WRITE)
    {
      /* prefetch head */
      uintptr_t base = (uintptr_t)message;
      nm_len_t head_offset = NM_IBVERBS_RCACHE_ALIGN - 1 - ((base - 1) % NM_IBVERBS_RCACHE_ALIGN);
      assert((base + head_offset) % NM_IBVERBS_RCACHE_ALIGN == 0);
      assert(head_offset >= 0 && head_offset < NM_IBVERBS_RCACHE_ALIGN);
      memcpy(&p_rcache->headers.ssigpkt.head[0], message, head_offset);
      p_rcache->headers.ssigpkt.sighdr.head_offset = head_offset;

      (*p_rcache->p_rcache_context->p_reg_driver->mem_reg)(p_rcache->p_rcache_context->p_ibverbs_context,
                                                           message + head_offset, size - head_offset,
                                                           &p_rcache->send.mr, &p_rcache->send.p_reg_handle);

      return NM_ESUCCESS;
    }
  else if(p_rcache->p_rcache_context->mode == NM_IBVERBS_RCACHE_READ)
    {
      (*p_rcache->p_rcache_context->p_reg_driver->mem_reg)(p_rcache->p_rcache_context->p_ibverbs_context, message, size,
                                                           &p_rcache->send.mr, &p_rcache->send.p_reg_handle);
      
      /* send rdvhdr */
      struct nm_ibverbs_rcache_rdvhdr_s*const p_rdvhdr = &p_rcache->headers.shdr;
      p_rdvhdr->raddr = (uintptr_t)p_rcache->send.message;
      p_rdvhdr->seq = p_rcache->send.seq;
      p_rdvhdr->busy  = 1;
      p_rdvhdr->rkey  = p_rcache->send.mr->rkey;
      NM_RCACHE_LOG("rdmaread- sent rdvhdr seq = %d; rdma raddr = %p; rkey = %d\n",
                    p_rcache->send.seq, (void*)p_rdvhdr->raddr, (int)p_rdvhdr->rkey);
      int rc = nm_ibverbs_rdma_send(p_rcache->p_cnx, sizeof(struct nm_ibverbs_rcache_rdvhdr_s),
                                    &p_rcache->headers.shdr, 
                                    &p_rcache->headers.rhdr,
                                    &p_rcache->headers,
                                    &p_rcache->seg,
                                    p_rcache->mr,
                                    NM_IBVERBS_WRID_RDV);
      if(rc != NM_ESUCCESS)
        return rc;
      /* flush pending requests */
      rc = nm_ibverbs_send_flush(p_rcache->p_cnx, NM_IBVERBS_WRID_RDV);
      if(rc != NM_ESUCCESS)
        return rc;
      return NM_ESUCCESS;
    }
  else
    {
      return -NM_ENOTIMPL;
    }
}

static int nm_ibverbs_rcache_send_poll(void*_status)
{
  struct nm_ibverbs_rcache_s*p_rcache = _status;
  if(p_rcache->p_rcache_context->mode == NM_IBVERBS_RCACHE_WRITE)
    {
      struct nm_ibverbs_rcache_rdvhdr_s*const p_rdvhdr = &p_rcache->headers.rhdr;
      if(p_rdvhdr->busy)
        {
          /* recv rdv header */
          const uint64_t raddr = p_rdvhdr->raddr;
          const uint32_t rkey  = p_rdvhdr->rkey;
          assert(p_rdvhdr->seq == p_rcache->send.seq);
          struct ibv_mr*mr = p_rcache->send.mr;
          p_rdvhdr->raddr = 0;
          p_rdvhdr->rkey  = 0;
          p_rdvhdr->busy  = 0;
          /* send data */
          const nm_len_t head_offset = p_rcache->headers.ssigpkt.sighdr.head_offset;
          int rc = nm_ibverbs_do_rdma(p_rcache->p_cnx, 
                                      p_rcache->send.message + head_offset,
                                      p_rcache->send.size - head_offset,
                                      raddr + head_offset, IBV_WR_RDMA_WRITE, IBV_SEND_SIGNALED,
                                      mr->lkey, rkey, NM_IBVERBS_WRID_DATA);
          if(rc != NM_ESUCCESS)
            return rc;
          NM_RCACHE_LOG("got rdv seq = %d; rdma send raddr = %p; rkey = %d; lkey = %d; rc = %d\n",
                        p_rcache->send.seq, (void*)raddr, (int)rkey, mr->lkey,  rc);
          /* send signal header */
          p_rcache->headers.ssigpkt.sighdr.seq = p_rcache->send.seq;
          p_rcache->headers.ssigpkt.sighdr.busy = 1;
#ifdef DEBUG
          if(p_rcache->p_rcache_context->do_crc)
            {
              const uint32_t crc = (*p_rcache->p_rcache_context->p_crc_func)(p_rcache->send.message, p_rcache->send.size);
              if(crc != p_rcache->send.crc)
                {
                  NM_FATAL("ibrcache: inconsistency detected in data between send_iov_post() and send_poll(); crc_post = 0x%x; crc_poll = 0x%x\n", p_rcache->send.crc, crc);
                }
              p_rcache->headers.ssigpkt.sighdr.crc = crc;
            }
#endif /* DEBUG */      
          rc = nm_ibverbs_rdma_send(p_rcache->p_cnx, sizeof(struct nm_ibverbs_rcache_sigpkt_s),
                                    &p_rcache->headers.ssigpkt,
                                    &p_rcache->headers.rsigpkt,
                                    &p_rcache->headers,
                                    &p_rcache->seg,
                                    p_rcache->mr,
                                    NM_IBVERBS_WRID_HEADER);
          if(rc != NM_ESUCCESS)
            return rc;
          /* flush pending requests */
          rc = nm_ibverbs_send_flush(p_rcache->p_cnx, NM_IBVERBS_WRID_DATA);
          if(rc != NM_ESUCCESS)
            return rc;
          rc = nm_ibverbs_send_flush(p_rcache->p_cnx, NM_IBVERBS_WRID_HEADER);
          if(rc != NM_ESUCCESS)
            return rc;
          (*p_rcache->p_rcache_context->p_reg_driver->mem_unreg)(p_rcache->p_rcache_context->p_ibverbs_context,
                                                                 p_rcache->send.message + head_offset,
                                                                 p_rcache->send.size - head_offset,
                                                                 p_rcache->send.mr, p_rcache->send.p_reg_handle);
        }
      else
        {
          return -NM_EAGAIN;
        }
    }
  else if(p_rcache->p_rcache_context->mode == NM_IBVERBS_RCACHE_READ)
    {
      /* wait for sighdr */
      struct nm_ibverbs_rcache_sighdr_s*p_rsig = &p_rcache->headers.rsigpkt.sighdr;
      if(p_rsig->busy)
        {
          NM_RCACHE_LOG("got sighdr\n");
          assert(p_rsig->seq == p_rcache->send.seq);
          p_rsig->busy = 0;
          (*p_rcache->p_rcache_context->p_reg_driver->mem_unreg)(p_rcache->p_rcache_context->p_ibverbs_context, p_rcache->send.message, p_rcache->send.size, p_rcache->send.mr, p_rcache->send.p_reg_handle);
        }
      else
        {
          return -NM_EAGAIN;
        }
    }
  else
    {
      return -NM_ENOTIMPL;
    }
  p_rcache->send.mr = NULL;
  p_rcache->send.p_reg_handle = NULL;
  p_rcache->send.message = NULL;
  p_rcache->send.size    = -1;
  return NM_ESUCCESS;
}

static int nm_ibverbs_rcache_recv_init(void*_status, struct iovec*v, int n)
{
  struct nm_ibverbs_rcache_s*p_rcache = _status;
  if(p_rcache->recv.message != NULL)
    {
      NM_FATAL("ibverbs: rendez-vous already posted on receiver side.\n");
    }
  assert(p_rcache->headers.rsigpkt.sighdr.busy == 0);

  p_rcache->recv.seq++;
  p_rcache->recv.message = v->iov_base;
  p_rcache->recv.size = v->iov_len;

  if(p_rcache->p_rcache_context->mode == NM_IBVERBS_RCACHE_WRITE)
    {
      (*p_rcache->p_rcache_context->p_reg_driver->mem_reg)(p_rcache->p_rcache_context->p_ibverbs_context,
                                                           p_rcache->recv.message,
                                                           p_rcache->recv.size,
                                                           &p_rcache->recv.mr, &p_rcache->recv.p_reg_handle);

      /* send rdvhdr */
      struct nm_ibverbs_rcache_rdvhdr_s*const p_rdvhdr = &p_rcache->headers.shdr;
      p_rdvhdr->raddr = (uintptr_t)p_rcache->recv.message;
      p_rdvhdr->seq = p_rcache->recv.seq;
      p_rdvhdr->busy  = 1;
      p_rdvhdr->rkey  = p_rcache->recv.mr->rkey;
      NM_RCACHE_LOG("rcache = %p; seq = %d; addr = %p; key = %d\n",
                    p_rcache, p_rcache->recv.seq, (void*)p_rdvhdr->raddr, (int)p_rdvhdr->rkey);
      
      int rc = nm_ibverbs_rdma_send(p_rcache->p_cnx, sizeof(struct nm_ibverbs_rcache_rdvhdr_s),
                                    &p_rcache->headers.shdr, 
                                    &p_rcache->headers.rhdr,
                                    &p_rcache->headers,
                                    &p_rcache->seg,
                                    p_rcache->mr,
                                    NM_IBVERBS_WRID_RDV);
      if(rc != NM_ESUCCESS)
        return rc;
      rc = nm_ibverbs_send_flush(p_rcache->p_cnx, NM_IBVERBS_WRID_RDV);
      if(rc != NM_ESUCCESS)
        return rc;
      return NM_ESUCCESS;
    }
  else if(p_rcache->p_rcache_context->mode == NM_IBVERBS_RCACHE_READ)
    {
      (*p_rcache->p_rcache_context->p_reg_driver->mem_reg)(p_rcache->p_rcache_context->p_ibverbs_context, p_rcache->recv.message, p_rcache->recv.size, &p_rcache->recv.mr, &p_rcache->recv.p_reg_handle);
      /* we're done */
      return NM_ESUCCESS;
    }
  else
    {
      return -NM_ENOTIMPL;
    }
}

static int nm_ibverbs_rcache_poll_one(void*_status)
{
  struct nm_ibverbs_rcache_s*p_rcache = _status;
  if(p_rcache->p_rcache_context->mode == NM_IBVERBS_RCACHE_WRITE)
    {
      /* wait for sighdr */
      struct nm_ibverbs_rcache_sighdr_s*p_rsig = &p_rcache->headers.rsigpkt.sighdr;
      if(p_rsig->busy)
        {
          const nm_len_t head_offset = p_rsig->head_offset;
          memcpy(p_rcache->recv.message, &p_rcache->headers.rsigpkt.head[0], head_offset);
#ifdef DEBUG
          if(p_rcache->p_rcache_context->do_crc)
            {
              const uint32_t crc = (*p_rcache->p_rcache_context->p_crc_func)(p_rcache->recv.message, p_rcache->recv.size);
              if(crc != p_rsig->crc)
                {
                  NM_FATAL("ibrcache: inconsistency detected in data between sender and receiver; crc_send = 0x%x; recv_crc = 0x%x\n", p_rsig->crc, crc);
                }
            }
#endif /* DEBUG */
          assert(p_rsig->seq == p_rcache->recv.seq);
          p_rsig->busy = 0;
          /* unregister */
          (*p_rcache->p_rcache_context->p_reg_driver->mem_unreg)(p_rcache->p_rcache_context->p_ibverbs_context,
                                                                 p_rcache->recv.message + head_offset,
                                                                 p_rcache->recv.size - head_offset,
                                                                 p_rcache->recv.mr, p_rcache->recv.p_reg_handle);
        }
      else
        {
          return -NM_EAGAIN;
        }
    }
  else if(p_rcache->p_rcache_context->mode == NM_IBVERBS_RCACHE_READ)
    {
      struct nm_ibverbs_rcache_rdvhdr_s*const p_rdvhdr = &p_rcache->headers.rhdr;
      if(p_rdvhdr->busy)
        {
          /* recv rdvhdr */
          const uint64_t raddr = p_rdvhdr->raddr;
          const uint32_t rkey  = p_rdvhdr->rkey;
          assert(p_rdvhdr->seq == p_rcache->recv.seq);
          struct ibv_mr*mr = p_rcache->recv.mr;
          p_rdvhdr->raddr = 0;
          p_rdvhdr->rkey  = 0;
          p_rdvhdr->busy  = 0;
          NM_RCACHE_LOG("rdmaread- got rdvhdr seq = %d; rdma raddr = %p; rkey = %d; lkey = %d\n",
                        p_rcache->recv.seq, (void*)raddr, (int)rkey, mr->lkey);
          /* post RDMA read */
          int rc = nm_ibverbs_do_rdma(p_rcache->p_cnx, 
                                      p_rcache->recv.message, p_rcache->recv.size,
                                      raddr, IBV_WR_RDMA_READ, IBV_SEND_SIGNALED,
                                      mr->lkey, rkey, NM_IBVERBS_WRID_DATA);
          if(rc != NM_ESUCCESS)
            return rc;
          /* flush RDMA requests */
          rc = nm_ibverbs_send_flush(p_rcache->p_cnx, NM_IBVERBS_WRID_DATA);
          if(rc != NM_ESUCCESS)
            return rc;
          /* send sighdr */
          p_rcache->headers.ssigpkt.sighdr.seq = p_rcache->recv.seq;
          p_rcache->headers.ssigpkt.sighdr.busy = 1;
          rc = nm_ibverbs_rdma_send(p_rcache->p_cnx, sizeof(struct nm_ibverbs_rcache_sighdr_s),
                                    &p_rcache->headers.ssigpkt.sighdr,
                                    &p_rcache->headers.rsigpkt.sighdr,
                                    &p_rcache->headers,
                                    &p_rcache->seg,
                                    p_rcache->mr,
                                    NM_IBVERBS_WRID_HEADER);
          if(rc != NM_ESUCCESS)
            return rc;
          /* flush sighdr */
          rc = nm_ibverbs_send_flush(p_rcache->p_cnx, NM_IBVERBS_WRID_HEADER);
          if(rc != NM_ESUCCESS)
            return rc;
          /* unregister */
          (*p_rcache->p_rcache_context->p_reg_driver->mem_unreg)(p_rcache->p_rcache_context->p_ibverbs_context, p_rcache->recv.message, p_rcache->recv.size, p_rcache->recv.mr, p_rcache->recv.p_reg_handle);
        }
      else
        {
          return -NM_EAGAIN;
        }
    }
  else
    {
      return -NM_ENOTIMPL;
    }
  /* cleanup */
  p_rcache->recv.mr = NULL;
  p_rcache->recv.p_reg_handle = NULL;
  p_rcache->recv.message = NULL;
  p_rcache->recv.size = -1;
  return NM_ESUCCESS;
}
