/*
 * NewMadeleine
 * Copyright (C) 2006-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "nm_ibverbs.h"
#include "nm_ibverbs_rcache.h"
#include <malloc.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>

#ifdef PUKABI
#include <Padico/Puk-ABI.h>
#endif

#ifdef PIOMAN
#include <pioman.h>
#define nm_ibverbs_mini_spinlock_t   piom_spinlock_t
#define nm_ibverbs_mini_spin_init    piom_spin_init
#define nm_ibverbs_mini_spin_destroy piom_spin_destroy
#define nm_ibverbs_mini_spin_lock    piom_spin_lock
#define nm_ibverbs_mini_spin_unlock  piom_spin_unlock
#define nm_ibverbs_mini_spin_assert_locked piom_spin_assert_locked
#define nm_ibverbs_mini_spin_assert_notlocked piom_spin_assert_notlocked
#else /* PIOMAN */
typedef int nm_ibverbs_mini_spinlock_t;
static inline void nm_ibverbs_mini_spin_ignore(nm_ibverbs_mini_spinlock_t*p_spin)
{ }
#define nm_ibverbs_mini_spin_init    nm_ibverbs_mini_spin_ignore
#define nm_ibverbs_mini_spin_destroy nm_ibverbs_mini_spin_ignore
#define nm_ibverbs_mini_spin_lock    nm_ibverbs_mini_spin_ignore
#define nm_ibverbs_mini_spin_unlock  nm_ibverbs_mini_spin_ignore
#define nm_ibverbs_mini_spin_assert_locked    nm_ibverbs_mini_spin_ignore
#define nm_ibverbs_mini_spin_assert_notlocked nm_ibverbs_mini_spin_ignore
#endif /* PIOMAN */

static void nm_ibverbs_mini_mem_init(struct nm_ibverbs_context_s*p_ibverbs_context);
static void nm_ibverbs_mini_mem_close(struct nm_ibverbs_context_s*p_ibverbs_context);
static void nm_ibverbs_mini_mem_reg(struct nm_ibverbs_context_s*p_ibverbs_context, const void*ptr, size_t len,
                                    struct ibv_mr**pp_mr, void**pp_handle);
static void nm_ibverbs_mini_mem_unreg(struct nm_ibverbs_context_s*p_ibverbs_context, const void*ptr, size_t len,
                                      struct ibv_mr*p_mr, void*p_handle);

const struct nm_ibverbs_reg_driver_s nm_ibverbs_reg_driver_mini =
  {
    .init      = &nm_ibverbs_mini_mem_init,
    .close     = &nm_ibverbs_mini_mem_close,
    .mem_reg   = &nm_ibverbs_mini_mem_reg,
    .mem_unreg = &nm_ibverbs_mini_mem_unreg
  };


/* ********************************************************* */

struct nm_ibverbs_mini_entry_s
{
  const void*ptr;
  size_t len;
  int refcount;
  uint64_t last_use;
  struct ibv_mr*p_mr;
};

#define NM_IBVERBS_MINI_NUM_ENTRIES 128

static struct
{
  struct nm_ibverbs_mini_entry_s entries[NM_IBVERBS_MINI_NUM_ENTRIES];
  nm_ibverbs_mini_spinlock_t lock;
  size_t filling;
  uint64_t seq;
} nm_ibverbs_mini;

/* ********************************************************* */

static void nm_ibverbs_mini_invalidate_hook(void*ptr, size_t size);

static void nm_ibverbs_mini_mem_init(struct nm_ibverbs_context_s*p_ibverbs_context)
{
  int i;
  for(i = 0; i < NM_IBVERBS_MINI_NUM_ENTRIES; i++)
    {
      nm_ibverbs_mini.entries[i].ptr = NULL;
      nm_ibverbs_mini.entries[i].len = 0;
      nm_ibverbs_mini.entries[i].refcount = 0;
    }
  nm_ibverbs_mini.filling = 0;
  nm_ibverbs_mini.seq = 0;
  nm_ibverbs_mini_spin_init(&nm_ibverbs_mini.lock);
#ifdef PUKABI
  /* invalidate mem reg upon free/realloc/munmap/mremap */
  puk_abi_mem_set_hook(&nm_ibverbs_mini_invalidate_hook);
#else
  NM_WARN("ibverbs: using rcache without PukABI.");
  mallopt(M_TRIM_THRESHOLD, -1);
  mallopt(M_MMAP_MAX, 0); /* XXX */
  mallopt(M_ARENA_MAX, 1);
  /*  malloc_stats(); */
#endif
}

static void nm_ibverbs_mini_mem_close(struct nm_ibverbs_context_s*p_ibverbs_context)
{
#ifdef PUKABI
  /* unregister memory hooks */
  puk_abi_mem_set_hook(NULL);
#endif /* PUKABI */
}

/** invalidate the given entry */
static void nm_ibverbs_mini_invalidate(struct nm_ibverbs_mini_entry_s*p_entry)
{
  nm_ibverbs_mini_spin_assert_locked(&nm_ibverbs_mini.lock);
  NM_RCACHE_LOG("evict cache slot #%d; ptr = %p; len = %d; last_use = %lu\n",
                (int)(p_entry - &nm_ibverbs_mini.entries[0]), p_entry->ptr, (int)p_entry->len, p_entry->last_use);
  assert(p_entry->refcount == 0);
  assert(p_entry->ptr != NULL);
  assert(p_entry->p_mr != NULL);
  struct ibv_mr*p_mr = p_entry->p_mr;
  nm_ibverbs_mini.filling -= p_entry->len;
  p_entry->p_mr = NULL;
  p_entry->ptr = NULL;
  p_entry->refcount = 0;
  nm_ibverbs_mini_spin_unlock(&nm_ibverbs_mini.lock);
  int rc = ibv_dereg_mr(p_mr);
  if(rc != 0)
    {
      NM_WARN("ibverbs: error %d (%s) while deregistering memory.\n", rc, strerror(rc));
    }
  nm_ibverbs_mini_spin_lock(&nm_ibverbs_mini.lock);
}

static void nm_ibverbs_mini_invalidate_hook(void*ptr, size_t len)
{
  if(ptr == NULL)
    return;
  int i;
  nm_ibverbs_mini_spin_lock(&nm_ibverbs_mini.lock);
  for(i = 0; i < NM_IBVERBS_MINI_NUM_ENTRIES; i++)
    {
      struct nm_ibverbs_mini_entry_s*p_entry = &nm_ibverbs_mini.entries[i];
      if( (p_entry->ptr != NULL) && (ptr != NULL) &&
          ( ((p_entry->ptr >= ptr) && (p_entry->ptr < ptr + len)) ) )
        {
          if(p_entry->refcount > 0)
            {
              NM_FATAL("ibverbs: cannot invalidate entry with refcount = %d\n", p_entry->refcount)
            }
          else
            {
              nm_ibverbs_mini_invalidate(p_entry);
            }
        }
    }
  nm_ibverbs_mini_spin_unlock(&nm_ibverbs_mini.lock);
}

static struct nm_ibverbs_mini_entry_s*nm_ibverbs_mini_lookup(const void*ptr, size_t len)
{
  nm_ibverbs_mini_spin_assert_locked(&nm_ibverbs_mini.lock);
  int i;
  for(i = 0; i < NM_IBVERBS_MINI_NUM_ENTRIES; i++)
    {
      struct nm_ibverbs_mini_entry_s*p_entry = &nm_ibverbs_mini.entries[i];
      if( (p_entry->ptr != NULL) &&
          (ptr == p_entry->ptr) &&
          (ptr + len == p_entry->ptr + p_entry->len) )
        {
          /* match */
          NM_RCACHE_LOG("ptr = %p; len = %d; matching entry #%d; ptr = %p; len = %d; refcount = %d\n",
                        ptr, (int)len, i, p_entry->ptr, (int)p_entry->len, p_entry->refcount);
          return p_entry;
        }
    }
  return NULL;
}

/** evict one cache entry */
static void nm_ibverbs_mini_evict(void)
{
  nm_ibverbs_mini_spin_assert_locked(&nm_ibverbs_mini.lock);
  uint64_t lru = nm_ibverbs_mini.seq;
  int lru_index = -1;
  struct nm_ibverbs_mini_entry_s*p_entry = NULL;
  int i;
  for(i = 0; i < NM_IBVERBS_MINI_NUM_ENTRIES; i++)
    {
      p_entry = &nm_ibverbs_mini.entries[i];
      if((p_entry->refcount == 0) && (p_entry->last_use < lru))
        {
          lru = p_entry->last_use;
          lru_index = i;
        }
    }
  if(lru_index == -1)
    {
      NM_WARN("ibverbs: all rcache entries in use; cannot evict any entry.\n");
      return;
    }
  /* evict cache entry */
  nm_ibverbs_mini_invalidate(&nm_ibverbs_mini.entries[lru_index]);
  nm_ibverbs_mini_spin_assert_locked(&nm_ibverbs_mini.lock);
}

/** find an empty slot */
static struct nm_ibverbs_mini_entry_s*nm_ibverbs_mini_find_slot(void)
{
  nm_ibverbs_mini_spin_assert_locked(&nm_ibverbs_mini.lock);
  int i;
 retry:
  for(i = 0; i < NM_IBVERBS_MINI_NUM_ENTRIES; i++)
    {
      if(nm_ibverbs_mini.entries[i].ptr == NULL)
        {
          NM_RCACHE_LOG("found empty slot #%d\n", i);
          return &nm_ibverbs_mini.entries[i];
        }
    }
  nm_ibverbs_mini_evict();
  nm_ibverbs_mini_spin_assert_locked(&nm_ibverbs_mini.lock);
  goto retry;
  return NULL;
}

static void nm_ibverbs_mini_mem_reg(struct nm_ibverbs_context_s*p_ibverbs_context, const void*ptr, size_t len,
                                    struct ibv_mr**pp_mr, void**pp_handle)
{
  nm_ibverbs_mini_spin_lock(&nm_ibverbs_mini.lock);
  nm_ibverbs_mini.seq++;
  /* lookup */
  struct nm_ibverbs_mini_entry_s*p_entry = nm_ibverbs_mini_lookup(ptr, len);
  if(p_entry)
    {
      __sync_fetch_and_add(&p_entry->refcount, 1);
      p_entry->last_use = nm_ibverbs_mini.seq;
      nm_ibverbs_mini_spin_unlock(&nm_ibverbs_mini.lock);
    }
  else
    {
      /* not found */
      p_entry = nm_ibverbs_mini_find_slot();
      p_entry->ptr = ptr;
      p_entry->len = len;
      p_entry->refcount = 1;
      p_entry->last_use = nm_ibverbs_mini.seq;
      nm_ibverbs_mini.filling += len;
      nm_ibverbs_mini_spin_unlock(&nm_ibverbs_mini.lock);
    retry:
      p_entry->p_mr = nm_ibverbs_reg_mr(p_ibverbs_context, (void*)ptr, len, NULL);
      if(p_entry->p_mr == NULL)
        {
          if(nm_ibverbs_mini.filling == 0)
            {
              NM_WARN("ibverbs: memory registration failed.\n");
              struct rlimit rlim;
              int rc = getrlimit(RLIMIT_MEMLOCK, &rlim);
              if(rc == 0)
                {
                  NM_DISPF("ibverbs: RLIMIT_MEMLOCK = %lu\n", rlim.rlim_cur);
                }
              abort();
            }
          else
            {
              NM_RCACHE_LOG("reg_mr failed; evict entry and retry.\n");
              nm_ibverbs_mini_spin_lock(&nm_ibverbs_mini.lock);
              nm_ibverbs_mini_evict();
              nm_ibverbs_mini_spin_assert_locked(&nm_ibverbs_mini.lock);
              nm_ibverbs_mini_spin_unlock(&nm_ibverbs_mini.lock);
              goto retry;
            }
        }
    }
  assert(p_entry->refcount >= 1);
  NM_RCACHE_LOG("ptr = %p; len = %d; refcount = %d; filling = %lld; brk = %p\n",
                ptr, (int)len, p_entry->refcount, (long long)nm_ibverbs_mini.filling, sbrk(0));
  *pp_mr = p_entry->p_mr;
  *pp_handle = p_entry;
}

static void nm_ibverbs_mini_mem_unreg(struct nm_ibverbs_context_s*p_ibverbs_context, const void*ptr, size_t len,
                                      struct ibv_mr*p_mr, void*p_handle)
{
  struct nm_ibverbs_mini_entry_s*p_entry = p_handle;
  assert(p_entry->refcount > 0);
  assert(p_entry->p_mr == p_mr);
  __sync_sub_and_fetch(&p_entry->refcount, 1);
  assert(p_entry->refcount >= 0);
}
