/*
 * NewMadeleine
 * Copyright (C) 2006-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdint.h>
#include <string.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/un.h>
#include <unistd.h>

#include <Padico/Module.h>
#include <nm_minidriver.h>
#include <nm_private.h>

#include <rdma/fabric.h>
#include <rdma/fi_atomic.h>
#include <rdma/fi_cm.h>
#include <rdma/fi_domain.h>
#include <rdma/fi_endpoint.h>
#include <rdma/fi_eq.h>
#include <rdma/fi_errno.h>
#include <rdma/fi_rma.h>
#include <rdma/fi_tagged.h>
#include <rdma/fi_trigger.h>

/** buffer size for buffer-based send/recv */
#define NM_OFI_BUF_SIZE (24 * 1024)

#define NM_FI_VERSION FI_VERSION(1, 4)

#if HAVE_DECL_FI_CONTEXT2
#define NM_OFI_USE_CONTEXT2 1
#endif

static void*nm_ofi_instantiate(puk_instance_t instance, puk_context_t context);
static void nm_ofi_destroy(void*status);

static const struct puk_component_driver_s nm_ofi_component =
  {
    .instantiate = &nm_ofi_instantiate,
    .destroy     = &nm_ofi_destroy
  };

static void nm_ofi_getprops(puk_context_t context, struct nm_minidriver_properties_s*props);
static void nm_ofi_init(puk_context_t context, const void**drv_url, size_t*url_size);
static void nm_ofi_close(puk_context_t context);
static void nm_ofi_connect(void* _status, const void*remote_url, size_t url_size);
static int  nm_ofi_send_iov_post(void*_status, const struct iovec*v, int n);
static int  nm_ofi_send_buf_get(void* _status, void**p_buffer, nm_len_t*p_len);
static int  nm_ofi_send_buf_post(void*_status, nm_len_t len);
static int  nm_ofi_send_poll(void*_status);
static int  nm_ofi_recv_iov_post(void*_status, struct iovec*v, int n);
static int  nm_ofi_recv_poll_one(void*_status);
static int  nm_ofi_recv_buf_poll(void*_status, void**p_buffer, nm_len_t*p_len);
static int  nm_ofi_recv_buf_release(void*_status);
static int  nm_ofi_recv_probe_any(puk_context_t p_context, void**_status);
static int  nm_ofi_recv_wait_any(puk_context_t p_context, void**_status);
static int  nm_ofi_recv_cancel(void*_status);
static int  nm_ofi_recv_cancel_any(puk_context_t p_context);

static const struct nm_minidriver_iface_s nm_ofi_minidriver =
  {
    .getprops         = &nm_ofi_getprops,
    .init             = &nm_ofi_init,
    .close            = &nm_ofi_close,
    .connect          = &nm_ofi_connect,
    .send_iov_post    = &nm_ofi_send_iov_post,
    .send_buf_get     = &nm_ofi_send_buf_get,
    .send_buf_post    = &nm_ofi_send_buf_post,
    .send_poll        = &nm_ofi_send_poll,
    .recv_iov_post    = &nm_ofi_recv_iov_post,
    .recv_poll_one    = &nm_ofi_recv_poll_one,
    .recv_buf_poll    = &nm_ofi_recv_buf_poll,
    .recv_buf_release = &nm_ofi_recv_buf_release,
    .recv_probe_any   = &nm_ofi_recv_probe_any,
    .recv_wait_any    = &nm_ofi_recv_wait_any,
    .recv_cancel      = &nm_ofi_recv_cancel,
    .recv_cancel_any  = &nm_ofi_recv_cancel_any
  };

/* ********************************************************* */

PADICO_MODULE_COMPONENT(Minidriver_ofi,
    puk_component_declare("Minidriver_ofi",
                          puk_component_attr("provider", "auto"),
                          puk_component_attr("fabric", "auto"),
                          puk_component_attr("domain", "auto"),
                          puk_component_provides("PadicoComponent", "component", &nm_ofi_component),
                          puk_component_provides("NewMad_minidriver", "minidriver", &nm_ofi_minidriver)));

/* ********************************************************* */

/** 'ofi' driver per-context data. */
struct nm_ofi_context_s
{
  struct fi_info*fi;
  struct fid_fabric*fabric;
  struct fid_domain*domain;
  struct fid_poll*poll;
  struct fid_pep*pep;   /**< passive endpoint for connection */
  struct fid_eq*eq;     /**< event queue, for connection */
  int next_key;         /**< next requested registration key */
  int need_mr;
  nm_trk_kind_t kind;
  void*url;
  size_t url_size;
};

/** 'ofi' per-instance status. */
struct nm_ofi_s
{
  struct nm_ofi_context_s*p_ofi_context;
  struct fid_ep*ep;  /**< active endpoint, used for data exchange */
  struct
  {
    char buffer[NM_OFI_BUF_SIZE];
    nm_len_t len;
    struct fid_mr*mr;
    struct fid_cq*cq;  /**< completion queue for send */
#ifdef NM_OFI_USE_CONTEXT2
    struct fi_context2 context;
#else
    struct fi_context context;
#endif
  } send;
  struct
  {
    char buffer[NM_OFI_BUF_SIZE];
    int posted;
    struct fid_mr*mr;
    struct fid_cq*cq;  /**< completion queue for recv */
#ifdef NM_OFI_USE_CONTEXT2
    struct fi_context2 context;
#else
    struct fi_context context;
#endif
  } recv;
};

#define NM_OFI_CHECK_RC(RC, FUNC)                                       \
  {                                                                     \
    if((RC) != 0)                                                       \
      {                                                                 \
        NM_FATAL("# nm_ofi: %s()- error %d (%s).\n", #FUNC, (RC), fi_strerror(-(RC))); \
      }                                                                 \
  }

/*#define NM_OFI_TRACE(FMT, ...) fprintf(stderr, FMT, ## __VA_ARGS__) */
#define NM_OFI_TRACE(FMT, ...) NM_TRACEF(FMT, ## __VA_ARGS__)

/* ********************************************************* */

static void* nm_ofi_instantiate(puk_instance_t instance, puk_context_t context)
{
  struct nm_ofi_s*p_status = malloc(sizeof(struct nm_ofi_s));
  struct nm_ofi_context_s*p_ofi_context = puk_context_get_status(context);
  assert(p_ofi_context != NULL);
  p_status->p_ofi_context = p_ofi_context;
  p_status->recv.posted = 0;
  p_status->ep = NULL;
  /* register global MR */
  int rc = fi_mr_reg(p_ofi_context->domain, p_status->send.buffer, NM_OFI_BUF_SIZE, FI_MSG, 0,
                     p_ofi_context->next_key++, 0, &p_status->send.mr, NULL);
  NM_OFI_CHECK_RC(rc, "fi_mr_reg");
  rc = fi_mr_reg(p_ofi_context->domain, p_status->recv.buffer, NM_OFI_BUF_SIZE, FI_MSG, 0,
                 p_ofi_context->next_key++, 0, &p_status->recv.mr, NULL);
  NM_OFI_CHECK_RC(rc, "fi_mr_reg");
  return p_status;
}

static void nm_ofi_destroy(void*_status)
{
  struct nm_ofi_s*p_status = _status;
  NM_OFI_TRACE("# nm_ofi: destroy()\n");
  fi_close(&p_status->send.mr->fid);
  fi_close(&p_status->recv.mr->fid);
  if(p_status->send.cq)
    {
      fi_close(&p_status->send.cq->fid);
    }
  if(p_status->send.cq)
    {
      fi_close(&p_status->recv.cq->fid);
    }
  fi_close(&p_status->ep->fid);
  free(p_status);
}

/* ********************************************************* */

static void nm_ofi_endpoint_open(struct nm_ofi_context_s*p_ofi_context, struct nm_ofi_s*p_status, struct fi_info*p_info)
{
  struct fi_cq_attr cq_attr;
  memset(&cq_attr, 0, sizeof(cq_attr));
  cq_attr.size = p_ofi_context->fi->tx_attr->size;
  cq_attr.flags = 0;
  cq_attr.format = FI_CQ_FORMAT_MSG;
  cq_attr.wait_obj = FI_WAIT_NONE;

  int rc = fi_endpoint(p_ofi_context->domain, p_info, &p_status->ep, NULL);
  NM_OFI_CHECK_RC(rc, "fi_endpoint");
  /* bind endpoint to global event queue */
  rc = fi_ep_bind(p_status->ep, &p_ofi_context->eq->fid, 0);
  NM_OFI_CHECK_RC(rc, "fi_ep_bind");
  /* open & bind CQs */
  rc = fi_cq_open(p_ofi_context->domain, &cq_attr, &p_status->send.cq, NULL);
  NM_OFI_CHECK_RC(rc, "fi_cq_open");
  rc = fi_ep_bind(p_status->ep, &p_status->send.cq->fid, FI_TRANSMIT);
  NM_OFI_CHECK_RC(rc, "fi_ep_bind");
  rc = fi_cq_open(p_ofi_context->domain, &cq_attr, &p_status->recv.cq, NULL);
  NM_OFI_CHECK_RC(rc, "fi_cq_open");
  rc = fi_ep_bind(p_status->ep, &p_status->recv.cq->fid, FI_RECV);
  NM_OFI_CHECK_RC(rc, "fi_ep_bind");
}

static void nm_ofi_endpoint_wait_connected(struct nm_ofi_context_s*p_ofi_context, struct nm_ofi_s*p_status)
{
  struct fi_eq_cm_entry entry;
  uint32_t event;
  NM_OFI_TRACE("# nm_ofi: waiting for event...\n");
  int rc = fi_eq_sread(p_ofi_context->eq, &event, &entry, sizeof entry, -1, 0);
  NM_OFI_TRACE("# nm_ofi: got event on event queue.\n");
  if(rc == -FI_EAVAIL)
    {
      struct fi_eq_err_entry error_entry;
      fi_eq_readerr(p_ofi_context->eq, &error_entry, 0);
      char buf[1024];
      NM_FATAL("# nm_ofi: error event: %s\n",
              fi_eq_strerror(p_ofi_context->eq, error_entry.prov_errno, error_entry.err_data, buf, 1024));
    }
  else if(rc < 0)
    {
      NM_FATAL("# nm_ofi: fi_eq_sread failed: %s (%d)\n", fi_strerror(-rc), -rc);
    }
  if(rc != sizeof(entry))
    {
      NM_FATAL("# nm_ofi: Reading from event queue failed (after connect): %s (%d)\n", fi_strerror(-rc), -rc);
    }
  if(event != FI_CONNECTED || entry.fid != &p_status->ep->fid)
    {
      NM_FATAL("# nm_ofi: Unexpected CM (event %d fid %p) (pep %p)", event, entry.fid, p_status->ep);
    }
}

static void nm_ofi_getprops(puk_context_t context, struct nm_minidriver_properties_s*props)
{
  struct nm_ofi_context_s*p_ofi_context = malloc(sizeof(struct nm_ofi_context_s));
  puk_context_set_status(context, p_ofi_context);
  props->profile.latency = 500;
  props->profile.bandwidth = 8000;
  props->capabilities.supports_data = 0;
  props->capabilities.prefers_wait_any = 0;
  props->capabilities.has_recv_any = 0;
  switch(props->hints.kind)
    {
    case nm_trk_small:
      props->capabilities.max_msg_size = NM_OFI_BUF_SIZE;
      props->capabilities.supports_buf_send = 1;
      props->capabilities.supports_buf_recv = 1;
      props->capabilities.supports_iovec = 0;
      break;
    case nm_trk_large:
      props->capabilities.supports_buf_send = 0;
      props->capabilities.supports_buf_recv = 0;
      props->capabilities.supports_iovec = 1;
      break;
    default:
      NM_FATAL("nm_ofi: bad value %d for trk kind hint.\n", props->hints.kind);
    }
  p_ofi_context->kind = props->hints.kind;
}

static void nm_ofi_init(puk_context_t context, const void**drv_url, size_t*url_size)
{
  struct nm_ofi_context_s*p_ofi_context = puk_context_get_status(context);
  struct fi_info*hints = fi_allocinfo();
  hints->caps = FI_MSG;
  hints->ep_attr->type = FI_EP_MSG; /* reliable, connected */
  /* TODO- FI_EP_RDM (reliable datagram) */
  /*  hints->domain_attr->mr_mode = FI_MR_LOCAL; */
  hints->mode = FI_LOCAL_MR | FI_CONTEXT;
#ifdef NM_OFI_USE_CONTEXT2
  hints->mode |= FI_CONTEXT2;
#endif
  /* detect provider */
  const char*p_provider = puk_context_getattr(context, "provider");
  if(strcmp(p_provider, "auto") != 0)
    {
      fprintf(stderr, "# nm_ofi: provider forced by attribute = %s\n", p_provider);
      hints->fabric_attr->prov_name = strdup(p_provider);
    }
  /* detect fabric */
  const char*p_fabric = puk_context_getattr(context, "fabric");
  if(strcmp(p_fabric, "auto") != 0)
    {
      fprintf(stderr, "# nm_ofi: fabric forced by attribute = %s\n", p_fabric);
      hints->fabric_attr->name = strdup(p_fabric);
    }
  /* detect domain */
  const char*p_domain = puk_context_getattr(context, "domain");
  if(strcmp(p_domain, "auto") != 0)
    {
      fprintf(stderr, "# nm_ofi: domain forced by attribute = %s\n", p_domain);
      hints->domain_attr->name = strdup(p_domain);
    }
  int rc = fi_getinfo(NM_FI_VERSION, NULL, NULL, 0, hints, &p_ofi_context->fi);
  NM_OFI_CHECK_RC(rc, "fi_getinfo");
  assert(p_ofi_context->fi != NULL);
  p_ofi_context->need_mr = !!(p_ofi_context->fi->mode & FI_LOCAL_MR);

  const struct fi_info*fi = p_ofi_context->fi;
  int i = 0;
  while(fi)
    {
      fprintf(stderr, "# nm_ofi: info[%d] provider = %s; fabric = %s; domain = %s; addr_format = %d; local_mr = %d; max_msg_size = %lu; next = %p\n",
              i,
              fi->fabric_attr->prov_name,
              fi->fabric_attr->name,
              fi->domain_attr->name,
              fi->addr_format,
              !!(fi->mode & FI_LOCAL_MR),
              fi->ep_attr->max_msg_size,
              fi->next);
      fi = fi->next;
      i++;
    }
  /* open fabric */
  rc = fi_fabric(p_ofi_context->fi->fabric_attr, &p_ofi_context->fabric, NULL);
  NM_OFI_CHECK_RC(rc, "fi_fabric");
  /* open event queue */
  struct fi_eq_attr eq_attr;
  memset(&eq_attr, 0, sizeof(eq_attr));
  eq_attr.size = p_ofi_context->fi->tx_attr->size;
  eq_attr.flags = 0;
  eq_attr.wait_obj = FI_WAIT_UNSPEC;
  rc = fi_eq_open(p_ofi_context->fabric, &eq_attr, &p_ofi_context->eq, NULL);
  NM_OFI_CHECK_RC(rc, "fi_eq_open");
  /* open domain */
  rc = fi_domain(p_ofi_context->fabric, p_ofi_context->fi, &p_ofi_context->domain, NULL);
  NM_OFI_CHECK_RC(rc, "fi_domain");
  /* open passive endpoint */
  rc = fi_passive_ep(p_ofi_context->fabric, p_ofi_context->fi, &p_ofi_context->pep, NULL);
  NM_OFI_CHECK_RC(rc, "fi_passive_ep");
  rc = fi_pep_bind(p_ofi_context->pep, &p_ofi_context->eq->fid, 0);
  NM_OFI_CHECK_RC(rc, "fi_pep_bind");
  rc = fi_listen(p_ofi_context->pep);
  NM_OFI_CHECK_RC(rc, "fi_listen");
  /* get URL */
  p_ofi_context->url_size = 0;
  fi_getname(&p_ofi_context->pep->fid, NULL, &p_ofi_context->url_size);
  p_ofi_context->url = malloc(p_ofi_context->url_size);
  rc = fi_getname(&p_ofi_context->pep->fid, p_ofi_context->url, &p_ofi_context->url_size);
  NM_OFI_CHECK_RC(rc, "fi_getname");
  p_ofi_context->next_key = 0;
  *drv_url = p_ofi_context->url;
  *url_size = p_ofi_context->url_size;
  NM_OFI_TRACE("# nm_ofi- init done; url_size = %d.\n", p_ofi_context->url_size);
}

static void nm_ofi_close(puk_context_t context)
{
  struct nm_ofi_context_s*p_ofi_context = puk_context_get_status(context);
  NM_OFI_TRACE("# nm_ofi- closing...\n");
  puk_context_set_status(context, NULL);
  free(p_ofi_context->url);
  fi_close(&p_ofi_context->pep->fid);
  fi_close(&p_ofi_context->eq->fid);
  fi_close(&p_ofi_context->domain->fid);
  fi_close(&p_ofi_context->fabric->fid);
  fi_freeinfo(p_ofi_context->fi);
  free(p_ofi_context);
  usleep(100000); /* HACK HERE- allow libfabric internal threads to exit before we unload the driver;
                  * fi_close() on domain & fabric is _not_ synchronous */
}

static void nm_ofi_connect(void*_status, const void*remote_url, size_t url_size)
{
  struct nm_ofi_s*p_status = _status;
  struct nm_ofi_context_s*p_ofi_context = p_status->p_ofi_context;
  const int is_server = (p_ofi_context->url_size > url_size) ||
    ( (p_ofi_context->url_size == url_size) && (memcmp(p_ofi_context->url, remote_url, url_size) > 0) );
  NM_OFI_TRACE("# nm_ofi- connecting...\n");

  if(!is_server)
    {
      /* ** active connect */
      NM_OFI_TRACE("# nm_ofi: active connect...\n");
      /* open connection endpoint */
      nm_ofi_endpoint_open(p_ofi_context, p_status, p_ofi_context->fi);

      /*
        struct fi_poll_attr poll_attr;
        poll_attr.flags = 0;
        rc = fi_poll_open(p_ofi_context->domain, &poll_attr, &p_ofi_context->poll);
        if(rc != 0)
        {
        NM_FATAL("# nm_ofi: fi_poll_open()- error %d (%s).\n", rc, fi_strerror(-rc));
        }
        rc = fi_poll_add(p_ofi_context->poll, &p_status->cq->fid, 0);
        if(rc != 0)
        {
        NM_FATAL("# nm_ofi: fi_poll_add()- error %d (%s).\n", rc, fi_strerror(-rc));
        }
      */

      /* issue conneciton request */
      int rc = fi_connect(p_status->ep, remote_url, NULL, 0);
      NM_OFI_CHECK_RC(rc, "fi_connect");
      /* wait for the connection to be established */
      nm_ofi_endpoint_wait_connected(p_ofi_context, p_status);
    }
  else
    {
      /* ** passive connect */
      NM_OFI_TRACE("# nm_ofi: passive connect...\n");
      /* Wait for a connection request */
      struct fi_eq_cm_entry entry;
      uint32_t event;
      int rc = fi_eq_sread(p_ofi_context->eq, &event, &entry, sizeof entry, -1, 0);
      if(rc < 0)
        {
          NM_FATAL("# nm_ofi: fi_eq_sread()- error %d (%s)\n", -rc, fi_strerror(-rc));
        }
      if(rc != sizeof entry)
        {
          NM_FATAL("# nm_ofi: Reading from event queue failed (after listen): %s (%d)\n", fi_strerror(-rc), -rc);
        }
      NM_OFI_TRACE("# nm_ofi: got event %d on event queue.\n", event);
      if(event != FI_CONNREQ)
        {
          NM_FATAL("# nm_ofi: Unexpected event: %d", event);
        }
      /* open active endpoint */
      nm_ofi_endpoint_open(p_ofi_context, p_status, entry.info);
      /* accept connection */
      rc = fi_accept(p_status->ep, NULL, 0);
      NM_OFI_CHECK_RC(rc, "fi_accept");
      /* wait for the connection to be etablished */
      nm_ofi_endpoint_wait_connected(p_ofi_context, p_status);
    }
  NM_OFI_TRACE("# nm_ofi: connection success!\n");
  int rc1 = fi_enable(p_status->ep);
  NM_OFI_CHECK_RC(rc1, "fi_enable");
}

static int nm_ofi_send_iov_post(void*_status, const struct iovec*v, int n)
{
  struct nm_ofi_s*p_status = _status;
  int rc = fi_sendv(p_status->ep, v, fi_mr_desc(p_status->send.mr), n, 0, &p_status->send.context);
  NM_OFI_CHECK_RC(rc, "fi_send");
  p_status->send.len = 0;
  int i;
  for(i = 0; i < n; i++)
    {
      p_status->send.len += v[i].iov_len;
    }
  return NM_ESUCCESS;
}

static int nm_ofi_send_buf_get(void*_status, void**p_buffer, nm_len_t*p_len)
{
  struct nm_ofi_s*p_status = _status;
  assert(p_status->p_ofi_context->kind == nm_trk_small);
  *p_buffer = &p_status->send.buffer;
  *p_len = NM_OFI_BUF_SIZE;
  return NM_ESUCCESS;
}

static int nm_ofi_send_buf_post(void*_status, nm_len_t len)
{
  struct nm_ofi_s*p_status = _status;
  assert(len <= NM_OFI_BUF_SIZE);
  assert(p_status->p_ofi_context->kind == nm_trk_small);
  p_status->send.len = len;
  int rc = fi_send(p_status->ep, p_status->send.buffer, p_status->send.len, fi_mr_desc(p_status->send.mr), 0, &p_status->send.context);
  NM_OFI_CHECK_RC(rc, "fi_send");
  return NM_ESUCCESS;
}

static int nm_ofi_send_poll(void*_status)
{
  struct nm_ofi_s*p_status = _status;
  struct fi_cq_msg_entry entry;
  int rc = fi_cq_read(p_status->send.cq, &entry, 1);
  if(rc == 1)
    {
      NM_OFI_TRACE("# nm_ofi- send SUCCESS; len = %d.\n", (int)entry.len);
      assert(entry.len == p_status->send.len);
      return NM_ESUCCESS;
    }
  else if(rc == -FI_EAGAIN)
    {
      return -NM_EAGAIN;
    }
  else if(rc == -FI_EAVAIL)
    {
      /* TODO- fi_cq_readerr() */
      NM_FATAL("nm_ofi: fi_cq_read()- asynchronous operation failed.");
    }
  else
    {
      NM_FATAL("nm_ofi: fi_cq_read()- error %d (%s)\n", rc, fi_strerror(-rc));
    }
}

static int nm_ofi_recv_iov_post(void*_status, struct iovec*v, int n)
{
  struct nm_ofi_s*p_status = _status;
  assert(!p_status->p_ofi_context->need_mr);
  assert(p_status->recv.posted == 0);
  int rc = fi_recvv(p_status->ep, v, NULL /*fi_mr_desc(p_status->send.mr) */, n, 0, &p_status->recv.context);
  NM_OFI_CHECK_RC(rc, "fi_send");
  p_status->recv.posted = 1;
  return NM_ESUCCESS;
}

static int nm_ofi_recv_poll_one(void*_status)
{
  struct nm_ofi_s*p_status = _status;
  struct fi_cq_msg_entry entry;
  int rc = fi_cq_read(p_status->recv.cq, &entry, 1);
  if(rc == 1)
    {
      NM_OFI_TRACE("# nm_ofi- recv SUCCESS - len = %d.\n", (int)entry.len);
      p_status->recv.posted = 0;
      return NM_ESUCCESS;
    }
  else if(rc == -FI_EAGAIN)
    {
      return -NM_EAGAIN;
    }
  else if(rc == -FI_EAVAIL)
    {
      /* TODO- fi_cq_readerr() */
      NM_FATAL("nm_ofi: fi_cq_read()- asynchronous operation failed.");
    }
  else
    {
      NM_FATAL("nm_ofi: fi_cq_read()- error %d (%s)\n", rc, fi_strerror(-rc));
    }
}

static int nm_ofi_recv_buf_poll(void*_status, void**p_buffer, nm_len_t*p_len)
{
  struct nm_ofi_s*p_status = _status;
  assert(p_status->p_ofi_context->kind == nm_trk_small);
  if(!p_status->recv.posted)
    {
      int rc = fi_recv(p_status->ep, &p_status->recv.buffer[0], NM_OFI_BUF_SIZE, fi_mr_desc(p_status->recv.mr), 0, &p_status->recv.context);
      NM_OFI_CHECK_RC(rc, "fi_recv");
      p_status->recv.posted = 1;
    }
  struct fi_cq_msg_entry entry;
  int rc = fi_cq_read(p_status->recv.cq, &entry, 1);
  if(rc == 1)
    {
      NM_OFI_TRACE("# nm_ofi- recv SUCCESS - len = %d.\n", (int)entry.len);
      *p_buffer = &p_status->recv.buffer[0];
      *p_len = entry.len;
      return NM_ESUCCESS;
    }
  else if(rc == -FI_EAGAIN)
    {
      return -NM_EAGAIN;
    }
  else if(rc == -FI_EAVAIL)
    {
      struct fi_cq_err_entry err_entry;
      rc = fi_cq_readerr(p_status->recv.cq, &err_entry, 0);
      NM_FATAL("nm_ofi: fi_cq_read()- asynchronous operation failed- err = %d (%s); prov_errno = %d.",
               err_entry.err, fi_strerror(err_entry.err), err_entry.prov_errno);
    }
  else
    {
      NM_FATAL("nm_ofi: fi_cq_read()- error %d (%s)\n", rc, fi_strerror(-rc));
    }
}

static int nm_ofi_recv_buf_release(void*_status)
{
  struct nm_ofi_s*p_status = _status;
  assert(p_status->p_ofi_context->kind == nm_trk_small);
  p_status->recv.posted = 0;
  return NM_ESUCCESS;
}

static int nm_ofi_recv_probe_any(puk_context_t p_context, void**_status)
{
  /* TODO */
  return -NM_ENOTIMPL;
}

static int nm_ofi_recv_wait_any(puk_context_t p_context, void**_status)
{
  /* TODO */
  return -NM_ENOTIMPL;
}

static int nm_ofi_recv_cancel(void*_status)
{
  /* TODO */
  return -NM_ENOTIMPL;
}

static int nm_ofi_recv_cancel_any(puk_context_t p_context)
{
  /* TODO */
  return -NM_ENOTIMPL;
}
