/*
 * NewMadeleine
 * Copyright (C) 2011-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/** @file
 * frontend to directly use Minidriver components.
 */

#ifndef NM_MINIDRIVER_INTERFACE_H
#define NM_MINIDRIVER_INTERFACE_H

#include <nm_minidriver.h>
#include <nm_log.h>
#include <nm_errno.h>

#include <sys/uio.h>

#define NM_MINIDRIVER_CHECK_RC(OP, RC)                                  \
  {                                                                     \
    if(((RC) != NM_ESUCCESS) && ((RC) != -NM_EAGAIN))                   \
      NM_FATAL("nm_minidriver: op %s returned rc = %d (%s)", (OP), rc, nm_strerror(rc)); \
  }


/* ** send ************************************************* */

static inline void nm_minidriver_isend(struct puk_receptacle_NewMad_minidriver_s*r,
                                       const struct nm_minidriver_properties_s*p_props,
                                       void*buf, size_t len)
{
  /* static data descriptor with a lifespan at least up to send_poll;
   * only _one_ send may be active at the same time */
  static struct iovec v = { .iov_base = NULL, .iov_len = -1 };
  static struct nm_data_s data;
  
  if(r->driver->send_buf_get)
    {
      void*p_driver_buf = NULL;
      nm_len_t driver_len = NM_LEN_UNDEFINED;
      int rc = (*r->driver->send_buf_get)(r->_status, &p_driver_buf, &driver_len);
      NM_MINIDRIVER_CHECK_RC("send_buf_get", rc);
      memcpy(p_driver_buf, buf, len);
      rc = (*r->driver->send_buf_post)(r->_status, len);
      NM_MINIDRIVER_CHECK_RC("send_buf_post", rc);
    }
  else if(r->driver->send_iov_post)
    {
      v = (struct iovec){ .iov_base = buf, .iov_len = len };
      int rc = (*r->driver->send_iov_post)(r->_status, &v, 1);
      NM_MINIDRIVER_CHECK_RC("send_iov_post", rc);
    }
  else
    {
      nm_data_contiguous_build(&data, buf, len);
      int rc = (*r->driver->send_data_post)(r->_status, &data, 0, len);
      NM_MINIDRIVER_CHECK_RC("send_data_post", rc);
    }
}

static inline int nm_minidriver_spoll(struct puk_receptacle_NewMad_minidriver_s*r,
                                      const struct nm_minidriver_properties_s*p_props,
                                      void*buf, size_t len)
{
  int rc = (*r->driver->send_poll)(r->_status);
  NM_MINIDRIVER_CHECK_RC("send_poll", rc);
  return rc;
}

static inline void nm_minidriver_send(struct puk_receptacle_NewMad_minidriver_s*r,
                                      const struct nm_minidriver_properties_s*p_props,
                                      void*buf, size_t len)
{
  nm_minidriver_isend(r, p_props, buf, len);
  int rc = -1;
  do
    {
      rc = nm_minidriver_spoll(r, p_props, buf, len);
    }
  while(rc != NM_ESUCCESS);
}


/* ** receive ********************************************** */

static inline void nm_minidriver_irecv(struct puk_receptacle_NewMad_minidriver_s*r,
                                       const struct nm_minidriver_properties_s*p_props,
                                       void*buf, size_t len)
{
  /* static data descriptor with a lifespan at least up to recv_poll;
   * only _one_ send may be active at the same time */
  static struct iovec v = { .iov_base = NULL, .iov_len = -1 };
  static struct nm_data_s data;

  if(r->driver->recv_buf_poll)
    {
      /* nothing to do */
    }
  else
    {
      if(r->driver->recv_iov_post)
        {
          v = (struct iovec){ .iov_base = buf, .iov_len = len };
          int rc = (*r->driver->recv_iov_post)(r->_status, &v, 1);
          NM_MINIDRIVER_CHECK_RC("recv_iov_post", rc);
        }
      else
        {
          nm_data_contiguous_build(&data, buf, len);
          int rc = (*r->driver->recv_data_post)(r->_status, &data, 0, len);
          NM_MINIDRIVER_CHECK_RC("recv_data_post", rc);
        }
    }
}

static inline int nm_minidriver_rpoll(struct puk_receptacle_NewMad_minidriver_s*r,
                                      const struct nm_minidriver_properties_s*p_props,
                                      void*buf, size_t len)
{
  if(r->driver->recv_buf_poll)
    {
      void*p_driver_buf = NULL;
      nm_len_t driver_len = NM_LEN_UNDEFINED;
      int rc = (*r->driver->recv_buf_poll)(r->_status, &p_driver_buf, &driver_len);
      NM_MINIDRIVER_CHECK_RC("recv_buf_poll", rc);
      if(rc == 0)
        {
          memcpy(buf, p_driver_buf, len);
          rc = (*r->driver->recv_buf_release)(r->_status);
          NM_MINIDRIVER_CHECK_RC("recv_buf_release", rc);
        }
      return rc;
    }
  else
    {
      int rc = (*r->driver->recv_poll_one)(r->_status);
      NM_MINIDRIVER_CHECK_RC("recv_poll_one", rc);
      return rc;
    }
}

static inline void nm_minidriver_recv(struct puk_receptacle_NewMad_minidriver_s*r,
                                      const struct nm_minidriver_properties_s*p_props,
                                      void*buf, size_t len)
{
  nm_minidriver_irecv(r, p_props, buf, len);
  int rc = -1;
  do
    {
      rc = nm_minidriver_rpoll(r, p_props, buf, len);
    }
  while(rc != NM_ESUCCESS);
}

/* ** receive any ****************************************** */

static inline void nm_minidriver_recv_probe_any(const struct nm_minidriver_iface_s*p_minidriver,
                                                puk_context_t p_context, void**_status)
{
  int rc = (*p_minidriver->recv_probe_any)(p_context, _status);
  NM_MINIDRIVER_CHECK_RC("recv_probe_any", rc);
}

static inline int nm_minidriver_recv_cancel(struct puk_receptacle_NewMad_minidriver_s*r,
                                            const struct nm_minidriver_properties_s*p_props)
{
  return (*r->driver->recv_cancel)(r->_status);
}

#endif /* NM_MINIDRIVER_INTERFACE_H */
