/*
 * NewMadeleine
 * Copyright (C) 2006-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#ifndef NM_CORE_H
#define NM_CORE_H

#ifndef NM_PRIVATE_H
#  error "Cannot include this file directly. Please include <nm_private.h>"
#endif

/** asynchronous tasks for nmad core.
 * These tasks are sheduled asynchronously and not immediately
 * in 2 cases:
 *   1. they need the core lock but are scheduled from a code that
 *      doesn't hold the lock (we avoid to acquire/release the lock
 *      too often).
 *   2. they trigger the processing of one more request or packet after
 *      a completion, which could lead to unbounded recursion.
 * Tasks are comprised of a kind and a parameter. They are combined in
 * a tagged pointer to fit a size suitable for atomic operations, so as
 * to avoid memory allocation.
 */
struct nm_core_task_s
{
  union
  {
    void*generic;
    struct
    {
      struct nm_unpack_next_s*p_next;
    } unpack_next;
    struct
    {
      struct nm_pkt_wrap_s*p_pw;
    } completed_pw;
    struct
    {
      struct nm_req_chunk_s*p_req_chunk;
    } pack_submission;
    struct
    {
      struct nm_gate_s*p_gate;
    } strat_schedule;
    struct
    {
      void (*p_handler)(void);
    } handler;
  };
};

enum nm_core_task_kind_e
  {
   NM_CORE_TASK_NONE            = 0,
   NM_CORE_TASK_UNPACK_NEXT     = 1, /**< try to match the next unpack on the given gate/tag/gtag */
   NM_CORE_TASK_COMPLETED_PW    = 2, /**< process a completed pw */
   NM_CORE_TASK_PACK_SUBMISSION = 3, /**< process a submitted pack request */
   NM_CORE_TASK_STRAT_SCHEDULE  = 4, /**< manually trigger nm_strat_schedule() on the given gate */
   NM_CORE_TASK_HANDLER         = 5  /**< call a user handler, mainly for testing/benchmarking */
  };
typedef enum nm_core_task_kind_e nm_core_task_kind_t;

/** lfqueue for pending tasks */
PUK_LFQUEUE_TYPE(nm_core_task, puk_tagged_ptr_t, NULL, 1024);

/** Core NewMadeleine structure.
 */
struct nm_core
{
#ifdef PIOMAN
  piom_spinlock_t lock;                         /**< lock to protect req lists in core */
  struct piom_ltask ltask;                      /**< task used for main core progress */
#endif /* PIOMAN */
#ifdef NMAD_HWLOC
  hwloc_topology_t topology;
#endif /* NMAD_HWLOC */

  puk_component_t strategy_component;              /**< component for the selected strategy */
  puk_context_t strategy_context;                  /**< global context of the strategy */
  const struct nm_strategy_iface_s*strategy_iface; /**< interface of the selected strategy */

  struct nm_gate_list_s gate_list;              /**< list of gates. */
  struct nm_drv_list_s driver_list;             /**< list of drivers. */
  int nb_drivers;                               /**< number of drivers in drivers list */

#ifndef PIOMAN
  /* if PIOMan is enabled, it already manages a list of pw to poll */
  struct nm_pkt_wrap_list_s pending_send_list;  /**< active pw for send to poll */
  struct nm_pkt_wrap_list_s pending_recv_list;  /**< active pw for recv to poll */
#endif /* !PIOMAN */

  struct nm_core_task_lfqueue_s pending_tasks;     /**< core tasks to execute asynchronously; always accessed from locked sections */
  struct nm_core_task_lfqueue_s submission_tasks;  /**< submission queue from non-locked sections, for the above list */

  struct nm_req_chunk_allocator_s req_chunk_allocator;   /**< allocator for req_chunk elements */
  struct nm_ctrl_chunk_allocator_s ctrl_chunk_allocator; /**< allocator for control chunks */
  struct nm_pw_nohd_allocator_s pw_nohd_allocator; /**< allocator for header-less pw*/
  struct nm_pw_buf_allocator_s pw_buf_allocator;   /**< allocator for pw with contiguous buffer */
  uint64_t unpack_seq;                             /**< next sequence number for unpacks */
  int n_packs, n_unpacks;                          /**< number of pending packs & unpacks */
  struct nm_matching_tag_table_s tag_table;        /**< matching infor for requests with tag (and any gate) */
  struct nm_matching_wildcard_table_s wildcard_table; /**< matching info for wildcard requests */
  struct nm_core_monitor_vect_s monitors;          /**< monitors for upper layers to track events in nmad core */
  struct nm_active_gate_list_s active_gates;       /**< list of gates with active requests */

  struct nm_trk_hashtable_s trk_table;             /**< trk table hashed by component status; used for reverse lookup */

  struct nm_core_dispatching_event_lfqueue_s dispatching_events; /**< queue for events to dispatch */
  struct nm_core_dispatching_event_allocator_s dispatching_event_allocator; /**< allocator for elements of above list */
  nm_spinlock_t dispatching_lock;                  /**< lock for dispatching of events */

  int enable_schedopt;                          /**< whether schedopt is enabled atop drivers */
  int enable_auto_flush;                        /**< automatic flush after each pack_submit */
  int enable_isend_csum;                        /**< check message integrity between isend submission & completion */

#ifdef NMAD_PROFILE
  struct
  {
    long long n_locks;
    long long n_schedule;
    long long n_packs;
    long long n_unpacks;
    long long n_unexpected;
    long long n_rdvs;
    long long n_pw_out;
    long long n_pw_in;
    long long n_try_and_commit;
    long long n_strat_apply;
    long long n_outoforder_event;
    long long n_event_queue_full;
    long max_unpacks;
    long max_packs;
  } profiling;
#endif /* NMAD_PROFILE */
};

#endif /* NM_CORE_H */
