/*
 * NewMadeleine
 * Copyright (C) 2013-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#ifndef NM_MINIDRIVER_H
#define NM_MINIDRIVER_H

#include <nm_public.h>
#include <Padico/Puk.h>
#ifdef NMAD_HWLOC
#include <hwloc.h>
#endif
#include <sys/uio.h>

/** Performance information for driver. This information is determined
 *  at init time and may depend on hardware and board number.
 */
struct nm_drv_profile_s
{
#ifdef NMAD_HWLOC
  hwloc_cpuset_t cpuset;  /**< cpu set close to the card (allocated by the driver, or NULL) */
#endif /* NMAD_HWLOC */
  /** Approximative performance of the board
   */
  int latency;   /**< in nanoseconds (10^-9 s) */
  int bandwidth; /**< in MB/s */
};

/** Static driver capabilities.
 */
struct nm_minidriver_capabilities_s
{
  nm_len_t max_msg_size; /**< maximum message size for the track */
  int supports_data;     /**< driver can send/recv direct nm_data_s */
  int supports_buf_send; /**< driver supported buffer-based send */
  int supports_buf_recv; /**< driver supported buffer-based recv */
  int supports_wait_any; /**< drivers supports passive wait_any */
  int prefers_wait_any;  /**< wait_any must be prefered for this driver */
  int supports_iovec;    /**< driver can send iovecs; if set to 0 and send_iov_post is non-NULL, then driver supports only iov count = 1*/
  int has_recv_any;      /**< driver accepts receive from any source (recv_probe_any) */
  int min_period;        /**< minimum delay between poll (in microseconds) */
  int is_exportable;     /**< blocking calls may be exported by PIOMan */
  int trk_rdv;           /**< trk needs a rdv for matched send/recv size; if 0, trk preserves boundaries */
};

/** some hints to help drivers tune themselves */
struct nm_minidriver_hints_s
{
  nm_trk_kind_t kind;
  int num_nodes;       /**< hint about the expected number of nodes; just a hint, not reliable */
};

struct nm_minidriver_properties_s
{
  struct nm_drv_profile_s profile;
  struct nm_minidriver_capabilities_s capabilities;
  struct nm_minidriver_hints_s hints;
};

/** Interface driver for the 'NewMad_minidriver' component interface.
 * A driver interface simpler than NewMad_Driver, that do not rely
 * on nmad core structures, making driver fully encapsulated.
 */
struct nm_minidriver_iface_s
{
  /* ** driver init; applied to context before/after instantiation */

  /** get driver properties (profiling and capabilities) */
  void (*getprops)(puk_context_t context, struct nm_minidriver_properties_s*props);
  /** init the driver and get its url */
  void (*init)(puk_context_t context, const void**drv_url, size_t*url_size);
  /** close the driver */
  void (*close)(puk_context_t context);

  /* ** connection establishment */

  /** connect this instance to the remote url  */
  void (*connect)(void*_status, const void*remote_url, size_t url_size);
  /** asynchronously start the connect process */
  void (*connect_async)(void*_status, const void*remote_url, size_t url_size);
  /** wait for a previously posted connect_async to complete */
  void (*connect_wait)(void*_status);
  /** disconnect this instance */
  void (*disconnect)(void*_status);
  /* TODO- register error handler for fault-tolerance */

  /* ** sending primitives */

  /** post a send for data described by iovec; v is expected to remain valid up to send completion */
  int (*send_iov_post)(void*_status, const struct iovec*v, int n) __attribute__ ((warn_unused_result));
  /** post a send for data described by nm_data iterator; p_data is expected to remain valid up to send completion */
  int (*send_data_post)(void*_status, const struct nm_data_s*p_data, nm_len_t chunk_offset, nm_len_t chunk_len) __attribute__ ((warn_unused_result));
  /** get a buffer to prepare data for send_buf_post */
  int (*send_buf_get)(void*_status, void**p_buffer, nm_len_t*p_len) __attribute__ ((warn_unused_result));
  /** post a send after upper layers have filled-in the buffer */
  int (*send_buf_post)(void*_status, nm_len_t len) __attribute__ ((warn_unused_result));
  /** poll a send operation already posted through send_iov_post, send_data_post, or send_buf_post */
  int (*send_poll)(void*_status) __attribute__ ((warn_unused_result));
  /** prefetch data in driver buffers before send_iov_post
   * @note currently broken; obsoleted by buffer-based send (see above) */
  void (*send_prefetch)(void*_status, const void*ptr, nm_len_t size);

  /* ** receiving primitives */

  /** post a recv for data described by iovec */
  int (*recv_iov_post)(void*_status, struct iovec*v, int n) __attribute__ ((warn_unused_result));
  /** post a recv for data described by nm_data iterator */
  int (*recv_data_post)(void*_status, const struct nm_data_s*p_data, nm_len_t chunk_offset, nm_len_t chunk_len) __attribute__ ((warn_unused_result));
  /** poll a recv operation already posted */
  int (*recv_poll_one)(void*_status) __attribute__ ((warn_unused_result));
  /** poll the driver for a pending recv; returns buffer */
  int (*recv_buf_poll)(void*_status, void**p_buffer, nm_len_t*p_len) __attribute__ ((warn_unused_result));
  /** release a buffer returned by buf_recv_poll() or buf_recv_wait_any() */
  int (*recv_buf_release)(void*_status) __attribute__ ((warn_unused_result));
  /** poll the driver for a pending recv from any source; returns status of instance */
  int (*recv_probe_any)(puk_context_t p_context, void**_status) __attribute__ ((warn_unused_result));
  /** passively wait for a pending recv from any source; returns status of instance */
  int (*recv_wait_any)(puk_context_t p_context, void**_status) __attribute__ ((warn_unused_result));
  /** cancel a posted recv */
  int (*recv_cancel)(void*_status) __attribute__ ((warn_unused_result));
  /** cancel a recv_wait_any */
  int (*recv_cancel_any)(puk_context_t p_context) __attribute__ ((warn_unused_result));
};
PUK_IFACE_TYPE(NewMad_minidriver, struct nm_minidriver_iface_s);


#endif /* NM_MINIDRIVER_H */
