/*
 * NewMadeleine
 * Copyright (C) 2006-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#ifndef NM_TACTICS_H
#define NM_TACTICS_H

/** remaining space in pw buffer */
static inline nm_len_t nm_pw_remaining_buf(struct nm_pkt_wrap_s*p_pw)
{
  assert(p_pw->max_len != NM_LEN_UNDEFINED);
  assert(p_pw->length <= p_pw->max_len);
  return p_pw->max_len - p_pw->length;
}

/** checks whether req_chunk is a SHORT chunk for optimized protocol */
static inline int nm_tactic_req_is_short(const struct nm_req_chunk_s*p_req_chunk)
{
  return ( (p_req_chunk->chunk_offset == 0) &&
	   (p_req_chunk->chunk_len < 255) &&
	   (p_req_chunk->chunk_len == p_req_chunk->p_req->pack.len) &&
	   (p_req_chunk->proto_flags == NM_PROTO_FLAG_LASTCHUNK) );
}

/** returns size in pw for the given short req_chunk */
static inline nm_len_t nm_tactic_req_short_size(const struct nm_req_chunk_s*p_req_chunk)
{
  assert(nm_tactic_req_is_short(p_req_chunk));
  return (NM_HEADER_SHORT_DATA_SIZE + p_req_chunk->chunk_len);
}

/** returns max size in pw for given data req_chunk-
 * value is a upper bound, actual size may be lower */
static inline nm_len_t nm_tactic_req_data_max_size(const struct nm_req_chunk_s*p_req_chunk)
{
  const nm_len_t chunk_len = p_req_chunk->chunk_len;
  const struct nm_req_s*p_pack = p_req_chunk->p_req;
  const struct nm_data_properties_s*p_props = nm_data_properties_get(&p_pack->data);
  const nm_len_t max_blocks = (p_props->blocks > chunk_len) ? chunk_len : p_props->blocks;
  const nm_len_t max_header_len = NM_HEADER_DATA_SIZE + max_blocks * sizeof(struct nm_header_pkt_data_chunk_s) + NM_ALIGN_FRONTIER;
  return (max_header_len + chunk_len);
}

/** compute average block size */
static inline nm_len_t nm_tactic_req_data_density(const struct nm_req_chunk_s*p_req_chunk)
{
  const struct nm_req_s*p_pack = p_req_chunk->p_req;
  const struct nm_data_properties_s*p_props = nm_data_properties_get(&p_pack->data);
  const nm_len_t density = (p_props->blocks > 0) ? p_props->size / p_props->blocks : 0;
  return density;
}

/** Try to pack a control chunk in a pw.
 * @return NM_ESUCCESS for success, -NM_EBUSY if pw full
 * @note function takes ownership of p_ctrl_chunk
 */
static inline int nm_tactic_pack_ctrl(nm_gate_t p_gate, nm_drv_t p_drv,
				      struct nm_ctrl_chunk_s*p_ctrl_chunk,
				      struct nm_pkt_wrap_s*p_pw)
{
  if(NM_HEADER_CTRL_SIZE < nm_pw_remaining_buf(p_pw))
    {
      nm_pw_add_control(p_pw, &p_ctrl_chunk->ctrl);
      nm_ctrl_chunk_list_remove(&p_gate->ctrl_chunk_list, p_ctrl_chunk);
      nm_ctrl_chunk_free(&p_gate->p_core->ctrl_chunk_allocator, p_ctrl_chunk);
      return NM_ESUCCESS;
    }
  else
    {
      return -NM_EBUSY;
    }
}

/** state for nm_tactic_pack_rdv_repack_chunk */
struct nm_tactic_rdv_repack_s
{
  struct nm_req_chunk_s*p_req_chunk_orig;
  nm_len_t offset;
};

/** repack each chunk of a large message separately */
static void nm_tactic_pack_rdv_repack_chunk(void*ptr, nm_len_t len, void*_context)
{
  struct nm_tactic_rdv_repack_s*p_repack = _context;
  struct nm_req_chunk_s*p_req_chunk_orig = p_repack->p_req_chunk_orig;
  struct nm_req_s*p_pack = p_req_chunk_orig->p_req;
  struct nm_core*p_core = p_pack->p_gate->p_core;
  struct nm_req_chunk_s*p_req_chunk = nm_req_chunk_alloc(p_core);
  nm_req_chunk_init(p_req_chunk, p_pack, p_repack->offset, len);
  nm_strat_submit_req_chunk(p_core, p_pack->p_gate, p_req_chunk, 1);
  p_repack->offset += len;
}

static inline int nm_tactic_pack_rdv(nm_gate_t p_gate, nm_drv_t p_drv,
				     struct nm_req_chunk_list_s*p_req_chunk_list,
				     struct nm_req_chunk_s*p_req_chunk,
				     struct nm_pkt_wrap_s*p_pw)
{
  if(NM_HEADER_CTRL_SIZE < nm_pw_remaining_buf(p_pw))
    {
      if(p_req_chunk_list)
	nm_req_chunk_list_remove(p_req_chunk_list, p_req_chunk);
      struct nm_req_s*p_pack = p_req_chunk->p_req;
      struct nm_data_properties_s props;
      nm_data_chunk_properties_compute(&p_pack->data, p_req_chunk->chunk_offset, p_req_chunk->chunk_len, &props);
      const nm_len_t density = props.size / props.blocks;
      const int is_lastchunk = (p_req_chunk->chunk_offset + p_req_chunk->chunk_len == p_pack->pack.len);
      nm_pw_flag_t flags = NM_PW_FLAG_NONE;
      if( (!props.is_contig) &&
          ((!p_drv->props.capabilities.supports_data) && (!p_drv->props.capabilities.supports_iovec)) )
	{
          /* chunk is not contiguous & driver does not support data iterators nor iovecs */
          if(density < NM_LARGE_MIN_DENSITY)
            {
              /* low density -> copy */
              flags |= NM_REQ_CHUNK_FLAG_USE_COPY;
            }
          else
            {
              /* high density (= large chunks) -> repack as multiple chunks */
              struct nm_tactic_rdv_repack_s repack =
                {
                  .p_req_chunk_orig = p_req_chunk,
                  .offset = p_req_chunk->chunk_offset
                };
              nm_data_chunk_extractor_traversal(&p_pack->data, p_req_chunk->chunk_offset, p_req_chunk->chunk_len,
                                                &nm_tactic_pack_rdv_repack_chunk, &repack);
              return -NM_EAGAIN;
            }
	}
      struct nm_pkt_wrap_s*p_large_pw = nm_pw_alloc_noheader(p_gate->p_core);
      nm_pw_add_req_chunk(p_large_pw, p_req_chunk, flags);
      nm_pkt_wrap_list_push_back(&p_pack->p_gate->pending_large_send, p_large_pw);
      union nm_header_ctrl_generic_s rdv;
      nm_header_init_rdv(&rdv, p_pack, p_req_chunk->chunk_len, p_req_chunk->chunk_offset, is_lastchunk ? NM_PROTO_FLAG_LASTCHUNK : 0);
      nm_pw_add_control(p_pw, &rdv);
      return NM_ESUCCESS;
    }
  else
    {
      return -NM_EBUSY;
    }
}

/** a chunk of rdv data.
 * Use one for regular rendezvous; use nb_driver chunks for multi-rail.
 */
struct nm_rdv_chunk_s
{
  nm_trk_id_t trk_id; /**< trk_id of track to send data on */
  nm_len_t len;       /**< length of data chunk */
};

/** Packs a series of RTR for multiple chunks of a pw, and post corresponding recv
*/
static inline void nm_tactic_rtr_pack(struct nm_core*p_core, struct nm_pkt_wrap_s*p_pw, int nb_chunks, const struct nm_rdv_chunk_s*p_chunks)
{
  int i;
  nm_len_t chunk_offset = p_pw->chunk_offset;
  const nm_seq_t seq = p_pw->p_unpack->seq;
  const nm_core_tag_t tag = p_pw->p_unpack->tag;
  nm_gate_t p_gate = p_pw->p_gate;
  for(i = 0; i < nb_chunks; i++)
    {
      struct nm_pkt_wrap_s*p_pw2 = NULL;
      if(p_chunks[i].len < p_pw->length)
	{
	  /* create a new pw with the remaining data */
          nm_pw_split(p_core, p_pw, &p_pw2, p_chunks[i].len);
	  assert(p_pw->length == p_chunks[i].len);
	}
      if(p_gate->trks[p_chunks[i].trk_id].kind == nm_trk_large)
	{
	  nm_core_post_recv(p_pw, p_pw->p_gate, p_chunks[i].trk_id, NULL);
	}
      else
	{
	  nm_pw_free(p_core, p_pw);
	}
      nm_core_post_rtr(p_gate, tag, seq, p_chunks[i].trk_id, chunk_offset, p_chunks[i].len);
      chunk_offset += p_chunks[i].len;
      p_pw = p_pw2;
      p_pw2 = NULL;
    }
}

/** Accept a rdv for the given large pw: split if needed, pack RTR, post recv, and store remaining part
 */
static inline void nm_tactic_rdv_accept(struct nm_core*p_core, struct nm_pkt_wrap_s*p_pw)
{
  nm_gate_t p_gate = p_pw->p_gate;
  struct nm_trk_s*p_trk_large = &p_gate->trks[NM_TRK_LARGE];
  const nm_len_t max_msg_size = p_trk_large->p_drv->props.capabilities.max_msg_size;
  if(p_pw->length > max_msg_size)
    {
      struct nm_pkt_wrap_s*p_pw2 = NULL;
      nm_pw_split(p_core, p_pw, &p_pw2, max_msg_size);
      nm_pkt_wrap_list_push_front(&p_gate->pending_large_recv, p_pw2);
      assert(p_pw->length <= max_msg_size);
      nm_strat_rdv_accept(p_core, p_gate);
    }
  assert(p_trk_large->p_pw_recv == NULL);
  struct nm_rdv_chunk_s chunk =
    { .len = p_pw->length, .trk_id = NM_TRK_LARGE };
  nm_tactic_rtr_pack(p_gate->p_core, p_pw, 1, &chunk);
}


#endif /* NM_TACTICS_H */
