/*
 * NewMadeleine
 * Copyright (C) 2014-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/* **********************************************************/
/* @file
 * trees description for collective operations
 */

#include <nm_log.h>
#include <nm_core_interface.h>

/** @ingroup coll_interface
 * @defgroup coll_tree Trees for collectives
 *
 * Common primitives to manipulate trees (binary, binomial, flat, chain) used in collectives
 */

/** @ingroup coll_tree
 * @{
 */

/** kind of tree */
enum nm_coll_tree_kind_e
  {
    NM_COLL_TREE_NONE = 0,
    NM_COLL_TREE_FLAT,
    NM_COLL_TREE_CHAIN,
    NM_COLL_TREE_BINOMIAL,
    NM_COLL_TREE_4NOMIAL,
    NM_COLL_TREE_BINARY,
    NM_COLL_TREE_4ARY,
    NM_COLL_TREE_DEFAULT,  /**< for public interfaces, will fallback to one of the previous tree kinds */
    _NM_COLL_TREE_MAX
  };
typedef enum nm_coll_tree_kind_e nm_coll_tree_kind_t;

/** description of an instanciated tree */
struct nm_coll_tree_info_s
{
  nm_coll_tree_kind_t kind;
  int max_arity;  /**< maximum number of children in tree */
  int n;          /**< number of nodes involved in collective */
  int self;       /**< rank of local node */
  int root;       /**< root node of the collective */
  int height;     /**< tree height: number of communications steps in the tree (0 for a single node) */
};

/** floor(log2(x)) */
static inline int nm_coll_log2_floor(int x)
{
  assert(x > 0);
  return ((unsigned) (8*sizeof (unsigned long long) - __builtin_clzll((x)) - 1));
}

/** ceil(log2(x)) */
static inline int nm_coll_log2_ceil(int x)
{
  assert(x > 0);
  return nm_coll_log2_floor(x) + (!!(x & (x - 1)));
}

/** floor(log4(x)) */
static inline int nm_coll_log4_floor(int x)
{
  return nm_coll_log2_floor(x) / 2;
}

/** fast integer power() = base ^ exp */
static inline int nm_coll_ipow(int base, int exp)
{
  assert(base >= 0);
  assert(exp >= 0);
  int result = 1;
  for(;;)
    {
      if(exp & 1)
        result *= base;
      exp >>= 1;
      if(!exp)
        break;
      base *= base;
    }
  return result;
}

/** ceil(log_n(x)) */
static inline int nm_coll_log_n_ceil(int x, const int n)
{
  assert(x > 0);
  assert(n >= 2);
  int l = 0;
  int exp = n - 1;
  x--;
  while(x > 0)
    {
      x -= exp;
      exp *= n;
      l++;
    }
  return l;
}

/** translate real ranks to virtual ranks (with root=0) */
static inline int nm_coll_r2v(int i, int n, int root)
{
  return ((n + i - root) % n);
}

/** translate virtual ranks (with root=0) to real ranks */
static inline int nm_coll_v2r(int i, int n, int root)
{
  return ((root + i) % n);
}

/* ** binomial tree **************************************** */

static inline void nm_coll_binomial_init(struct nm_coll_tree_info_s*p_tree)
{
  p_tree->max_arity = 1;
  p_tree->height = nm_coll_log2_ceil(p_tree->n);
}

/** walk to the leafs of the tree
 * @param i rank of local node
 * @param n number of leafs
 * @param level level in tree, numbered from the leafs
 */
static inline int nm_coll_binomial_child(int i, int n, int level)
{
  assert(level >= 0);
  const int skip = (1 << level);
  assert(skip >= 0);
  if((i % (2*skip) == 0) && (i + skip < n))
    return (i + skip);
  else
    return -1;
}

/** walk to the root of the tree, level numbered from the leafs */
static inline int nm_coll_binomial_parent(int i, int n, int level)
{
  const int skip = (1 << level);
  if(i % (2*skip) == skip)
    return (i - skip);
  else
    return -1;
}

/** number of leafs below the vertex */
static inline int nm_coll_binomial_weight(int i, int n, int level)
{
  const int skip = (1 << (level));
  const int parent = nm_coll_binomial_parent(i, n, level);
  if(parent == -1)
    return -1;
  else if(i + skip > n)
    return (n - i);
  else
    return skip;
}

static inline void nm_coll_binomial_step(const struct nm_coll_tree_info_s*p_tree, int step, int*p_parent, int*p_children, int*n_children)
{
  const int child = nm_coll_binomial_child(nm_coll_r2v(p_tree->self, p_tree->n, p_tree->root), p_tree->n, p_tree->height - step - 1);
  const int parent = nm_coll_binomial_parent(nm_coll_r2v(p_tree->self, p_tree->n, p_tree->root), p_tree->n, p_tree->height - step - 1);
  if(parent != -1)
    {
      *p_parent = nm_coll_v2r(parent, p_tree->n, p_tree->root);
    }
  else
    {
      *p_parent = -1;
    }
  if(child != -1)
    {
      p_children[0] = nm_coll_v2r(child, p_tree->n, p_tree->root);
      *n_children = 1;
    }
  else
    {
      p_children[0] = -1;
      *n_children = 0;
    }
}

/* ** k-nomial tree **************************************** */

static inline void nm_coll_4_nomial_init(struct nm_coll_tree_info_s*p_tree)
{
  p_tree->max_arity = 3;
  p_tree->height = nm_coll_log_n_ceil(p_tree->n, 4);
}

static inline int nm_coll_4_nomial_child(int i, int n, int level)
{
  const int K = 4;
  const int skip = nm_coll_ipow(K, level);
  if((i % (K * skip) == 0) && (i + skip < n))
    return (i + skip);
  else
    return -1;
}

static inline void nm_coll_4_nomial_step(const struct nm_coll_tree_info_s*p_tree, int step, int*p_parent, int*p_children, int*n_children)
{
  const int K = 4;
  const int level = p_tree->height - step - 1;
  const int i = nm_coll_r2v(p_tree->self, p_tree->n, p_tree->root);
  const int skip = nm_coll_ipow(K, level);
  /* parent */
  *p_parent = -1;
  int j;
  for(j = 0; j < K - 1; j++)
    {
      if(i % (K * skip) == skip + j * skip)
        {
          *p_parent = nm_coll_v2r(i - skip - j * skip, p_tree->n, p_tree->root);
        }
    }
  /* children */
  if((i % (K * skip) == 0) && (i + skip < p_tree->n))
    {
      int c = 0;
      for(j = 0; j < K - 1; j++)
        {
          if(i + skip + skip * j < p_tree->n)
            {
              p_children[c] = nm_coll_v2r(i + skip + skip * j, p_tree->n, p_tree->root);
              c++;
            }
        }
      for(j = c; j < p_tree->max_arity; j++)
        {
          p_children[j] = -1;
        }
      *n_children = c;
    }
  else
    {
      *n_children = 0;
      for(j = 0; j < p_tree->max_arity; j++)
        {
          p_children[j] = -1;
        }
    }
}


/* ** flat tree ******************************************** */

static inline void nm_coll_flat_init(struct nm_coll_tree_info_s*p_tree)
{
  p_tree->max_arity = (p_tree->self == p_tree->root) ? p_tree->n : 0;
  p_tree->height = 1;
}

static inline int nm_coll_flat_weight(const struct nm_coll_tree_info_s*p_tree, int step, int self)
{
  assert(self >= 0 && self < p_tree->n);
  if(step == 0)
    {
      if(self == p_tree->root)
        {
          return p_tree->n - 1;
        }
      else
        {
          return 1;
        }
    }
  else
    {
      return -1;
    }
}

static inline void nm_coll_flat_step(const struct nm_coll_tree_info_s*p_tree, int step, int*p_parent, int*p_children, int*n_children)
{
  if(p_tree->self == p_tree->root)
    {
      int j, k;
      for(k = 0, j = 0; k < p_tree->n; k++)
        {
          if(k != p_tree->root)
            {
              p_children[j] = k;
              j++;
            }
        }
      p_children[p_tree->n - 1] = -1;
      *n_children = p_tree->n - 1;
      *p_parent = -1;
    }
  else
    {
      *n_children = 0;
      *p_parent = p_tree->root;
    }
}

/* ** chain ************************************************ */

static inline void nm_coll_chain_init(struct nm_coll_tree_info_s*p_tree)
{
  p_tree->max_arity = 1;
  p_tree->height = p_tree->n - 1;
}

static inline void nm_coll_chain_step(const struct nm_coll_tree_info_s*p_tree, int step, int*p_parent, int*p_children, int*n_children)
{
  if((p_tree->self == step) && (p_tree->self < p_tree->n - 1))
    {
      p_children[0] = p_tree->self + 1;
      *n_children = 1;
    }
  else
    {
      p_children[0] = -1;
      *n_children = 0;
    }
  if((p_tree->self - 1 == step) && (p_tree->self > 0))
    {
      *p_parent = p_tree->self - 1;
    }
  else
    {
      *p_parent = -1;
    }
}

/* ** binary tree ****************************************** */

static inline int int_min(const int a, const int b)
{
  return (a <= b) ? a : b;
}

static inline int nm_coll_binary_nb_nodes_in_left_subtree(const struct nm_coll_tree_info_s* const p_tree)
{
  int height = p_tree->height + 1; // we need that the height of a tree with only a root equals 1

  // Number of nodes, including root node, in a full binary tree with height height-1:
  int nb_nodes_tree_previous_height = (1 << (height-1)) - 1;

  // Maximum theoretical number of nodes (here, leafs) in last stage:
  int max_nb_nodes_last_height = 1 << (height-1);

  /*
   * The number of nodes in the left subtree is:
   * * the number of nodes in the full left subtree of height h-1
   * * + the minimum between the real number of nodes in the last stage
   *     and the maximal number of nodes in the last stage of the subtree
   * * (in other words, we need to know if nodes in the last stage overflow to right subtree)
   */
  return (nb_nodes_tree_previous_height >> 1) + int_min(p_tree->n - (1 << (height-1)) + 1, max_nb_nodes_last_height >> 1);
}

static inline void nm_coll_binary_init(struct nm_coll_tree_info_s*p_tree)
{
  p_tree->max_arity = 2;
  p_tree->height = nm_coll_log2_floor(p_tree->n);
}

static inline void nm_coll_binary_step(const struct nm_coll_tree_info_s*p_tree, int step, int*p_parent, int*p_children, int*n_children)
{
  const int parent = (p_tree->self - 1) / 2;
  const int child1 = 2 * p_tree->self + 1;
  const int child2 = 2 * p_tree->self + 2;
  /* parent */
  if(nm_coll_log2_floor(p_tree->self + 1) - 1 == step)
    {
      *p_parent = parent;
    }
  else
    {
      *p_parent = -1;
    }
  /* children */
  *n_children = 0;
  p_children[0] = -1;
  p_children[1] = -1;
  if(nm_coll_log2_floor(p_tree->self + 1) == step)
    {
      if(p_tree->n > child1)
        {
          p_children[0] = child1;
          (*n_children)++;
        }
      if(p_tree->n > child2)
        {
          p_children[1] = child2;
          (*n_children)++;
        }
    }
}

/* ** 4-ary tree ******************************************* */

static inline void nm_coll_4ary_init(struct nm_coll_tree_info_s*p_tree)
{
  p_tree->max_arity = 4;
  p_tree->height = nm_coll_log4_floor(p_tree->n);
}

/* ** generic trees **************************************** */

/** initialize a generic tree descriptor
 */
static inline void nm_coll_tree_init(struct nm_coll_tree_info_s*p_tree, nm_coll_tree_kind_t kind, int n, int self, int root)
{
  p_tree->kind = kind;
  p_tree->root = root;
  p_tree->n = n;
  p_tree->self = self;
  switch(kind)
    {
    case NM_COLL_TREE_FLAT:
      nm_coll_flat_init(p_tree);
      break;
    case NM_COLL_TREE_CHAIN:
      nm_coll_chain_init(p_tree);
      break;
    case NM_COLL_TREE_BINOMIAL:
      nm_coll_binomial_init(p_tree);
      break;
    case NM_COLL_TREE_4NOMIAL:
      nm_coll_4_nomial_init(p_tree);
      break;
    case NM_COLL_TREE_BINARY:
      nm_coll_binary_init(p_tree);
      break;
    default:
      NM_FATAL("coll- unexpected kind of tree %d.\n", kind);
      break;
    }
}

/**  get the parent & children for the local node at the given step.
 * @note by convention, steps are numbered in _broadcast_ order from 0 to tree.height - 1
 */
static inline void nm_coll_tree_step(const struct nm_coll_tree_info_s*p_tree, int step, int*p_parent, int*p_children, int*n_children)
{
  switch(p_tree->kind)
    {
    case NM_COLL_TREE_FLAT:
      nm_coll_flat_step(p_tree, step, p_parent, p_children, n_children);
      break;
    case NM_COLL_TREE_CHAIN:
      nm_coll_chain_step(p_tree, step, p_parent, p_children, n_children);
      break;
    case NM_COLL_TREE_BINOMIAL:
      nm_coll_binomial_step(p_tree, step, p_parent, p_children, n_children);
      break;
    case NM_COLL_TREE_4NOMIAL:
      nm_coll_4_nomial_step(p_tree, step, p_parent, p_children, n_children);
      break;
    case NM_COLL_TREE_BINARY:
      nm_coll_binary_step(p_tree, step, p_parent, p_children, n_children);
      break;
    default:
      NM_FATAL("coll- unexpected kind of tree %d.\n", p_tree->kind);
      break;
    }
}

static inline int nm_coll_tree_weight(const struct nm_coll_tree_info_s*p_tree, int step, int self)
{
  switch(p_tree->kind)
    {
    case NM_COLL_TREE_FLAT:
      return nm_coll_flat_weight(p_tree, step, self);
      break;
    case NM_COLL_TREE_BINOMIAL:
      {
        const int vself = nm_coll_r2v(self, p_tree->n, p_tree->root);
        return nm_coll_binomial_weight(vself, p_tree->n, p_tree->height - step - 1);
      }
      break;
    default:
      NM_FATAL("coll- unexpected kind of tree %d.\n", p_tree->kind);
      break;
    }
 return -1;
}

static inline nm_coll_tree_kind_t nm_coll_tree_heuristic(int comm_size, nm_len_t data_size)
{
  static enum nm_coll_tree_kind_e default_kind = NM_COLL_TREE_NONE;
  if(default_kind == NM_COLL_TREE_NONE)
    {
      const char*s_kind = getenv("NM_COLL_TREE_KIND");
      if((s_kind == NULL) || (strcmp("", s_kind) == 0) || (strcmp("default", s_kind) == 0))
        {
          default_kind = NM_COLL_TREE_DEFAULT;
        }
      else if(strcmp("flat", s_kind) == 0)
        {
          default_kind = NM_COLL_TREE_FLAT;
        }
      else if(strcmp("chain", s_kind) == 0)
        {
          default_kind = NM_COLL_TREE_CHAIN;
        }
      else if(strcmp("binomial", s_kind) == 0)
        {
          default_kind = NM_COLL_TREE_BINOMIAL;
        }
      else if(strcmp("4nomial", s_kind) == 0)
        {
          default_kind = NM_COLL_TREE_4NOMIAL;
        }
      else if(strcmp("binary", s_kind) == 0)
        {
          default_kind = NM_COLL_TREE_BINARY;
        }
      else
        {
          NM_FATAL("coll: cannot parse tree kind '%s'\n", s_kind);
        }
      if(default_kind != NM_COLL_TREE_DEFAULT)
        {
          NM_DISPF("coll: tree kind = '%s' forced by environment.\n", s_kind);
        }
    }
  if(default_kind == NM_COLL_TREE_DEFAULT)
    {
      if(data_size > 32 * 1024)
        {
          return NM_COLL_TREE_BINOMIAL;
        }
      else
        {
          return NM_COLL_TREE_4NOMIAL;
        }
    }
  else
    {
      return default_kind;
    }
}


/* ********************************************************* */

typedef void (*nm_coll_tree_notifier_t)(void*ref);

/** common status for a non-blocking operation that walks a tree
 */
struct nm_coll_tree_status_s
{
 nm_session_t p_session;
  nm_group_t p_group;
  nm_tag_t tag;
  struct nm_coll_tree_info_s tree;
  int n_children;
  int*children;
  nm_sr_request_t*children_requests;
  nm_sr_request_t parent_request;
  int pending_reqs;                   /**< number of pending send/recv operations */
  nm_cond_status_t status;
  nm_coll_tree_notifier_t p_notify;
  void*p_ref;
};

static inline void nm_coll_tree_status_init(struct nm_coll_tree_status_s*p_status, nm_session_t p_session,
                                            nm_coll_tree_kind_t kind, nm_group_t p_group, int root, int self, nm_tag_t tag)
{
  p_status->p_session = p_session;
  p_status->p_group = p_group;
  p_status->tag = tag;
  nm_coll_tree_init(&p_status->tree, kind, nm_group_size(p_group), self, root);
  p_status->n_children = -1;
  p_status->children_requests = malloc((1 + p_status->tree.max_arity) * sizeof(nm_sr_request_t));
  p_status->children = malloc((1 + p_status->tree.max_arity) * sizeof(int));
  nm_cond_init(&p_status->status, 0);
  p_status->pending_reqs = 0;
  p_status->p_notify = NULL;
  p_status->p_ref = NULL;
}

/** set notification function to be called upon completion
 * @note when notifier is called, status is already destroyed */
static inline void nm_coll_tree_status_set_notifier(struct nm_coll_tree_status_s*p_status,
                                                    nm_coll_tree_notifier_t p_notify, void*p_ref)
{
  p_status->p_notify = p_notify;
  p_status->p_ref = p_ref;
}

static inline void nm_coll_tree_status_destroy(struct nm_coll_tree_status_s*p_status)
{
  assert(p_status->p_session != NULL);
  p_status->p_session = NULL;
  assert(p_status->p_group != NULL);
  p_status->p_group = NULL;
  assert(p_status->pending_reqs == 0);
  free(p_status->children_requests);
  p_status->children_requests = NULL;
  free(p_status->children);
  p_status->children = NULL;
  nm_cond_destroy(&p_status->status);
}

static inline void nm_coll_tree_status_wait(struct nm_coll_tree_status_s*p_status)
{
  nm_cond_wait(&p_status->status, 1, p_status->p_session->p_core);
  nm_coll_tree_status_destroy(p_status);
}

static inline void nm_coll_tree_status_signal(struct nm_coll_tree_status_s*p_status)
{
  if(p_status->p_notify == NULL)
    {
      nm_cond_signal(&p_status->status, 1);
    }
  else
    {
      /* destroy status then call notifier, since notifier may deallocated status */
      nm_coll_tree_notifier_t p_notify = p_status->p_notify;
      void*p_ref = p_status->p_ref;
      nm_coll_tree_status_destroy(p_status);
      p_status = NULL;
      (*p_notify)(p_ref);
    }
}

static inline int nm_coll_tree_status_test(struct nm_coll_tree_status_s*p_status)
{
  int rc = nm_cond_test(&p_status->status, 1);
  if(rc)
    {
      nm_coll_tree_status_destroy(p_status);
    }
  return rc;
}

static inline void nm_coll_tree_send(struct nm_coll_tree_status_s*p_coll_tree, int dest,
                                     struct nm_data_s*p_data, nm_sr_request_t*p_req)
{
  assert(dest != -1);
  nm_gate_t p_gate = nm_group_get_gate(p_coll_tree->p_group, dest);
  nm_atomic_inc(&p_coll_tree->pending_reqs);
  nm_sr_isend_data(p_coll_tree->p_session, p_gate, p_coll_tree->tag, p_data, p_req);
}

static inline void nm_coll_tree_recv(struct nm_coll_tree_status_s*p_coll_tree, int from,
                                     struct nm_data_s*p_data, nm_sr_request_t*p_req)
{
  assert(from != -1);
  nm_gate_t p_gate = nm_group_get_gate(p_coll_tree->p_group, from);
  nm_atomic_inc(&p_coll_tree->pending_reqs);
  nm_sr_irecv_data(p_coll_tree->p_session, p_gate, p_coll_tree->tag, p_data, p_req);
}

static inline void nm_coll_tree_set_notifier(struct nm_coll_tree_status_s*p_coll_tree, nm_sr_request_t*p_req,
                                            void(*p_notifier)(nm_sr_event_t, const nm_sr_event_info_t*, void*))
{
  nm_sr_request_set_ref(p_req, p_coll_tree);
  nm_sr_request_monitor(p_coll_tree->p_session, p_req, NM_SR_EVENT_FINALIZED, p_notifier);
}

static inline int nm_coll_tree_req_index(struct nm_coll_tree_status_s*p_coll_tree, nm_sr_request_t*p_request)
{
  if( (p_request < &p_coll_tree->children_requests[0]) ||
      (p_request > &p_coll_tree->children_requests[p_coll_tree->tree.max_arity - 1]) )
    {
      NM_FATAL("given p_request = %p is not in coll_tree", p_request);
    }
  const int i = p_request - p_coll_tree->children_requests;
  assert(i >= 0);
  assert(i < p_coll_tree->tree.max_arity);
  return i;
}


/** @} */
