/** @file
 * @brief PadicoControl over minidriver
 * @note usable only as secondary controler, cannot bootstrap
 */

#include <Padico/Puk.h>
#include <Padico/Module.h>
#include <Padico/PadicoTM.h>
#include <Padico/PadicoControl.h>
#include <Padico/Topology.h>
#include <Padico/PMI2.h>

#define NB_RECV_BUF 8
#define NAMESERV_DRIVER_URL_STR     "Control-minidriver-%d"
#define NAMESERV_CONNECT_SYNCH_STR  "Control-minidriver-%d-%d-ready"
#define NAMESERV_DRIVER_URL_SIZE    31
#define NAMESERV_CONNECT_SYNCH_SIZE 37

PADICO_MODULE_DECLARE(Control_minidriver);
PADICO_MODULE_INIT(control_minidriver_init);

/* ********************************************************* */

#if defined(NMAD) && defined(PIOMAN)

#include <nm_minidriver.h>
#include <pioman.h>


/* *** PadicoComponent driver ******************************** */

static void*control_minidriver_instantiate(puk_instance_t instance, puk_context_t context);
static void control_minidriver_destroy(void*_status);

/** instanciation facet for Control_minidriver
 */
const static struct puk_component_driver_s control_minidriver_component_driver =
{
  .instantiate = &control_minidriver_instantiate,
  .destroy     = &control_minidriver_destroy
};

/* *** PadicoControl driver ******************************** */

static void control_minidriver_send_common(void*_status, padico_na_node_t node, padico_control_msg_t msg);

/** 'Control' facet for Control/PSP
 */
const static struct padico_na_control_driver_s control_minidriver_control_driver =
{
  .send_common = &control_minidriver_send_common
};

/* *** PadicoConnector driver ****************************** */

static int control_minidriver_connector_connect(void*_status, padico_req_t req, padico_na_node_t node);

/** 'Connector' facet for Control/PSP
 */
const static struct padico_control_connector_driver_s control_minidriver_connector_driver =
{
  .connect = &control_minidriver_connector_connect
};

/* *** PadicoBootstrap driver ******************************** */

static int control_minidriver_bootstrap_new(void*_status, const struct in_addr*ip, uint16_t*port);
static int control_minidriver_bootstrap_connect(void*_status, const struct in_addr*ip, const uint16_t port);

/** 'Bootstrap' facet for Control/Minidriver
 */
const static struct padico_bootstrap_driver_s minidriver_bootstrap_driver =
{
  .bootstrap_new     = &control_minidriver_bootstrap_new,
  .bootstrap_connect = &control_minidriver_bootstrap_connect
};

/* ********************************************************* */

/** context for control-minidriver */
PUK_VECT_TYPE(piom_ltask, struct piom_ltask*);
PUK_VECT_TYPE(iovec, struct iovec*);
struct control_minidriver_context_s
{
  piom_ltask_vect_t polling_ltasks;
  char*msg_buf[NB_RECV_BUF];
  int msg_size[NB_RECV_BUF];
  int next_in;
  int next_out;
  marcel_mutex_t lock;
  int polling_active;
  int nb_instances;
  int polling_done;
  marcel_cond_t done;
  struct nm_minidriver_properties_s props;
  const void*drv_url;
  size_t url_size;
};

struct control_minidriver_status_s
{
  puk_instance_t instance;
  puk_context_t context;
  struct puk_receptacle_NewMad_minidriver_s minidriver;
  padico_na_node_t peer;          /**< peer node */
  struct iovec v_size;            /** iovec used for polling */
  struct iovec v_msg;             /** iovec used for polling */
};

static struct
{
  struct piom_ltask ltask;
} control_minidriver;

/* ********************************************************* */

static void*control_minidriver_polling(void*_status);

static void*control_minidriver_instantiate(puk_instance_t instance, puk_context_t context)
{
  struct control_minidriver_status_s*status = padico_malloc(sizeof(struct control_minidriver_status_s));
  status->instance = instance;
  status->context  = context;
  status->peer     = NULL;
  puk_context_indirect_NewMad_minidriver(instance, "minidriver", &status->minidriver);

  /* instantiate control-minidriver context */
  struct control_minidriver_context_s* task_list = puk_context_get_status(context);
  if(task_list == NULL)
    {
      task_list = malloc(sizeof(struct control_minidriver_context_s));
      task_list->polling_ltasks = piom_ltask_vect_new();
      for(int i=0; i< NB_RECV_BUF; i++)
        task_list->msg_buf[i] = NULL;
      task_list->next_in = 0;
      task_list->next_out = 0;
      task_list->polling_active = 1;
      task_list->nb_instances = 0;
      task_list->polling_done = 0;
      marcel_mutex_init(&task_list->lock, NULL);
      marcel_cond_init(&task_list->done, NULL);
      puk_context_set_status(context, task_list);

      /* first control-minidriver instance inits ibverbs driver */
      puk_context_t minidriver_context = puk_context_get_subcontext(context, puk_iface_NewMad_minidriver(), NULL);
      struct puk_receptacle_NewMad_minidriver_s*r = &status->minidriver;
      (*r->driver->getprops)(minidriver_context, &task_list->props);
      (*r->driver->init)(minidriver_context, &task_list->drv_url, &task_list->url_size);

      /* activate polling */
      padico_tm_invoke_later(&control_minidriver_polling, status);
    }
  task_list->nb_instances++;
  return status;
}

static void control_minidriver_destroy(void*_status)
{
  struct control_minidriver_status_s*status = _status;
  struct control_minidriver_context_s* task_list = puk_context_get_status(status->context);
  //Stop polling
  if(task_list->polling_active == 1)
    {
      marcel_mutex_lock(&task_list->lock);
      task_list->polling_active = 0;
      marcel_mutex_unlock(&task_list->lock);

      //Wait for polling to be done
      marcel_mutex_lock(&task_list->lock);
      while(!task_list->polling_done)
        {
          marcel_cond_wait(&task_list->done, &task_list->lock);
        }
      marcel_mutex_unlock(&task_list->lock);

      //Destroy polling tasks
      while(!piom_ltask_vect_empty(task_list->polling_ltasks))
        {
          struct piom_ltask*ltask = piom_ltask_vect_pop_back(task_list->polling_ltasks);
          free(ltask);
        }
      piom_ltask_vect_delete(task_list->polling_ltasks);
    }
  task_list->nb_instances--;
  //Last instance is in charge of deallocating context
  if(task_list->nb_instances == 0)
    {
      marcel_mutex_destroy(&task_list->lock);
      free(task_list);
      puk_context_set_status(status->context, NULL);
    }
  padico_control_unplug_controler(status->instance);
  padico_free(_status);
}

/* ********************************************************* */

static void control_minidriver_send_common(void*_status, padico_na_node_t node, padico_control_msg_t msg)
{
  struct control_minidriver_status_s*status = _status;
  struct puk_receptacle_NewMad_minidriver_s*r = &status->minidriver;
  const struct iovec v_msg = padico_control_msg_get_bytes(msg);
  int len = padico_control_msg_get_size(msg);
  struct iovec v_size = { .iov_base = &len, .iov_len = sizeof(len) };
  int rc;

  //Send message size
  (*r->driver->send_iov_post)(r->_status, &v_size, 1);
  do
    {
      rc = (*r->driver->send_poll)(r->_status);
    }
  while(rc != 0);

  //Send message
  (*r->driver->send_iov_post)(r->_status, &v_msg, 1);
  do
    {
      rc = (*r->driver->send_poll)(r->_status);
    }
  while(rc != 0);
}

static int control_minidriver_task_poll(void*_status)
{
  struct control_minidriver_status_s*status = _status;
  struct puk_receptacle_NewMad_minidriver_s*r = &status->minidriver;

  int rc = (*r->driver->recv_poll_one)(r->_status);
  if(rc != 0)
    return 0;

  struct control_minidriver_context_s* task_list = puk_context_get_status(status->context);
  // Received message
  if(status->v_msg.iov_base != NULL)
    {
      struct iovec v = status->v_msg;
      task_list->msg_size[task_list->next_in] = v.iov_len;
      task_list->msg_buf[task_list->next_in] = malloc(v.iov_len);
      memcpy(task_list->msg_buf[task_list->next_in], v.iov_base, v.iov_len);

      (*r->driver->recv_iov_post)(r->_status, &status->v_size, 1);

      free(status->v_msg.iov_base);
      status->v_msg.iov_base = NULL;
      task_list->next_in = (task_list->next_in + 1 ) % NB_RECV_BUF;
    }
  //Received message size
  else
    {
      assert(task_list->msg_buf[task_list->next_in] == NULL);
      int recv_size = *(int*)status->v_size.iov_base;
      status->v_msg.iov_base = malloc(recv_size);
      status->v_msg.iov_len = recv_size;
      (*r->driver->recv_iov_post)(r->_status, &status->v_msg, 1);
    }
  return 0;
}

static void*control_minidriver_dispatcher(void*dummy)
{
  struct control_minidriver_status_s*_status = NULL;
  piom_ltask_create(&control_minidriver.ltask, &control_minidriver_task_poll, &_status, PIOM_LTASK_OPTION_REPEAT);
  for(;;)
    {
      piom_ltask_submit(&control_minidriver.ltask);
      piom_ltask_wait(&control_minidriver.ltask);
      struct control_minidriver_status_s*status = (void*)_status;
      if(status)
	{
	}
    }
  return NULL;
}

static void*control_minidriver_polling(void*_status)
{
  struct control_minidriver_status_s*status = _status;
  puk_context_t context = status->context;
  struct control_minidriver_context_s* task_list = puk_context_get_status(context);

  marcel_mutex_lock(&task_list->lock);
  int polling_active = task_list->polling_active;
  marcel_mutex_unlock(&task_list->lock);

  //Iterate on all tasks
  while(polling_active)
    {
      marcel_mutex_lock(&task_list->lock);
      piom_ltask_vect_itor_t tasks_itor;
      for(tasks_itor  = piom_ltask_vect_begin(task_list->polling_ltasks);
          tasks_itor != piom_ltask_vect_end(task_list->polling_ltasks);
          tasks_itor  = piom_ltask_vect_next(tasks_itor))
        {
          piom_ltask_submit(*tasks_itor);
          piom_ltask_wait(*tasks_itor);
          char*buf = task_list->msg_buf[task_list->next_out];
          if(buf != NULL)
            {
              int size = task_list->msg_size[task_list->next_out];
              char*recv_msg = malloc(size);
              memcpy(recv_msg, buf, size);
              padico_control_deliver_message(status->instance, status->peer, recv_msg, size);
              padico_free(buf);
              task_list->msg_buf[task_list->next_out] = NULL;
              task_list->next_out = (task_list->next_out + 1) % NB_RECV_BUF;
            }
        }
      //Check if we still need to poll
      polling_active = task_list->polling_active;
      marcel_mutex_unlock(&task_list->lock);
    }
  marcel_mutex_lock(&task_list->lock);
  task_list->polling_done = 1;
  marcel_cond_signal(&task_list->done);
  marcel_mutex_unlock(&task_list->lock);
  return NULL;
}

/* ********************************************************* */

static int control_minidriver_connector_connect(void*_status, padico_req_t req, padico_na_node_t node)
{
  return -1;
}

/* ********************************************************* */
/* *** PadicoBootstrap */
static int control_minidriver_bootstrap_new(void*_status, const struct in_addr*ip, uint16_t*port)
{
  struct control_minidriver_status_s*status = _status;
  struct control_minidriver_context_s* context = puk_context_get_status(status->context);
  //Encode url
  int size = context->url_size;
  char*enc_url = puk_hex_encode(context->drv_url, &size, NULL);

  //Publish driver url
  char service_name[NAMESERV_DRIVER_URL_SIZE];
  sprintf(service_name, NAMESERV_DRIVER_URL_STR, padico_na_rank());
  control_nameserv_pmi2_nameserv_publish(service_name, enc_url);
  control_nameserv_pmi2_nameserv_fence();
  return 0;
}

static int control_minidriver_bootstrap_connect(void*_status, const struct in_addr*ip, const uint16_t port)
{
  struct control_minidriver_status_s*status = _status;
  struct control_minidriver_context_s* context = puk_context_get_status(status->context);
  char remote_service_name[NAMESERV_DRIVER_URL_SIZE];
  //char synch_service_name[NAMESERV_CONNECT_SYNCH_SIZE];
  int enc_size = context->url_size*2;
  char remote_enc_addr[enc_size + 1];

  //Retrieve driver url
  sprintf(remote_service_name, NAMESERV_DRIVER_URL_STR, port);
  control_nameserv_pmi2_nameserv_lookup(remote_service_name, remote_enc_addr, enc_size + 1);
  void*remote_addr = puk_hex_decode(remote_enc_addr, &enc_size, NULL);

  //Connection
  struct puk_receptacle_NewMad_minidriver_s*r = &status->minidriver;
  (*r->driver->connect)(r->_status, remote_addr, enc_size);

  //Synch to ensure connection is established before sending data
  //sprintf(synch_service_name, NAMESERV_CONNECT_SYNCH_STR, padico_na_rank(), port);
  //control_nameserv_pmi2_nameserv_publish(synch_service_name, NULL);
  //control_nameserv_pmi2_nameserv_fence();
  //sprintf(synch_service_name, NAMESERV_CONNECT_SYNCH_STR, port, padico_na_rank());
  //control_nameserv_pmi2_nameserv_lookup(synch_service_name, NULL, 0);

  padico_control_plug_controler(status->instance, NULL);

  //====================================//
  /* announce local node to peer node */
  int rc;
  padico_string_t self_node = padico_na_node_serialize(padico_na_getlocalnode());
  const char*msg = padico_string_get(self_node);
  size_t len = strlen(msg) + 1;
  char*recv_msg;
  size_t recv_size;

  if(padico_na_rank()==0)
    {
      // Send message size
      struct iovec v = { .iov_base = &len, .iov_len = sizeof(len) };
      (*r->driver->send_iov_post)(r->_status, &v, 1);
      do
        {
          rc = (*r->driver->send_poll)(r->_status);
        }
      while(rc != 0);

      // Receive message size
      v.iov_base = &recv_size;
      v.iov_len = sizeof(recv_size);
      (*r->driver->recv_iov_post)(r->_status, &v, 1);
      do
        {
          rc = (*r->driver->recv_poll_one)(r->_status);
        }
      while(rc != 0);

      // Send message
      v.iov_base = msg;
      v.iov_len = len;
      (*r->driver->send_iov_post)(r->_status, &v, 1);
      do
        {
          rc = (*r->driver->send_poll)(r->_status);
        }
      while(rc != 0);

      // Receive message
      recv_msg = padico_malloc(recv_size);
      v.iov_base = recv_msg;
      v.iov_len = recv_size;
      (*r->driver->recv_iov_post)(r->_status, &v, 1);
      do
        {
          rc = (*r->driver->recv_poll_one)(r->_status);
        }
      while(rc != 0);
    }
  else
    {
      // Receive message size
      struct iovec v = { .iov_base = &recv_size, .iov_len = sizeof(recv_size) };
      (*r->driver->recv_iov_post)(r->_status, &v, 1);
      do
        {
          rc = (*r->driver->recv_poll_one)(r->_status);
        }
      while(rc != 0);

      // Send message size
      v.iov_base = &len;
      (*r->driver->send_iov_post)(r->_status, &v, 1);
      do
        {
          rc = (*r->driver->send_poll)(r->_status);
        }
      while(rc != 0);

      //Receive message
      recv_msg = padico_malloc(recv_size);
      v.iov_base = recv_msg;
      v.iov_len = recv_size;
      (*r->driver->recv_iov_post)(r->_status, &v, 1);
      do
        {
          rc = (*r->driver->recv_poll_one)(r->_status);
        }
      while(rc != 0);

      // Send message
      v.iov_base = msg;
      v.iov_len = len;
      (*r->driver->send_iov_post)(r->_status, &v, 1);
      do
        {
          rc = (*r->driver->send_poll)(r->_status);
        }
      while(rc != 0);
    }
  padico_string_delete(self_node);

  /* parse peer announce */
  padico_control_event_disable();
  struct puk_parse_entity_s e = puk_xml_parse_buffer(recv_msg, recv_size, PUK_TRUST_CLUSTER);
  padico_rc_t rc2 = puk_parse_get_rc(&e);
  if(padico_rc_iserror(rc2))
    {
      padico_rc_show(rc2);
      padico_fatal("Control-minidriver: parse error on initial handshake message.");
    }
  if(!puk_parse_is(&e, "Topology:node"))
    {
      padico_fatal("Control-minidriver: root entity is not \"Topology:node\" while initializing control channel.\n");
    }
  padico_na_node_t peer = puk_parse_get_content(&e);
  assert(peer != NULL);
  assert(status->peer == NULL);
  status->peer = peer;
  padico_control_new_node(status->instance, peer);
  padico_control_event_enable();

  assert(padico_na_getnodebyuuid(padico_na_node_getuuid(peer)) != NULL);

  /* authorize to send */
  //Not needed while this is sequential

  /* create a polling task and its correponding iovec for this driver */
  struct control_minidriver_context_s* task_list = puk_context_get_status(status->context);
  // create ltask
  struct piom_ltask*polling_task = malloc(sizeof(struct piom_ltask));
  piom_ltask_create(polling_task, &control_minidriver_task_poll, _status, PIOM_LTASK_OPTION_ONESHOT);
  // create recv iovec
  status->v_size.iov_base = malloc(sizeof(int));
  status->v_size.iov_len = sizeof(int);
  status->v_msg.iov_base = NULL;
  // post the first receive before polling starts
  (*r->driver->recv_iov_post)(r->_status, &status->v_size, 1);
  // add ltask to polling tasks
  marcel_mutex_lock(&task_list->lock);
  piom_ltask_vect_push_back(task_list->polling_ltasks, polling_task);
  marcel_mutex_unlock(&task_list->lock);

  /* notify connection establishment */
  //Not needed while this is sequential

  padico_free(recv_msg);
  return 0;
}

#endif /* NMAD && PIOMAN */

/* ********************************************************* */

static int control_minidriver_init(void)
{
#if defined(NMAD) && defined(PIOMAN)
  puk_component_declare("Control-minidriver",
			puk_component_provides("PadicoComponent", "component", &control_minidriver_component_driver),
			puk_component_provides("PadicoControl", "control", &control_minidriver_control_driver),
			puk_component_provides("PadicoConnector", "connector", &control_minidriver_connector_driver),
                        puk_component_provides("PadicoBootstrap", "bootstrap", &minidriver_bootstrap_driver),
			puk_component_uses("NewMad_minidriver", "minidriver"));
  //padico_tm_invoke_later(&control_minidriver_dispatcher, NULL);
#else /* NMAD && PIOMAN */

  padico_warning("Control-minidriver: not initialized- NewMad not included in core.");
  return -1;

#endif /* NMAD && PIOMAN */
  return 0;
}
