/*
 * NewMadeleine
 * Copyright (C) 2006-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <nm_private.h>
#include <nm_launcher.h>
#include <nm_core_interface.h>
#include "nm_private_config.h"

#include <Padico/Module.h>
#include <Padico/Puk.h>
#include <Padico/Topology.h>
#include <Padico/NetSelector.h>
#include <Padico/PadicoControl.h>
#include <Padico/PMI2.h>

#define NM_LAUNCHER_URL_SIZE 1024

/** An instance of NewMad_Launcher_madico
 */
struct nm_pmi2_launcher_status_s
{
  int size;
  int rank;
  nm_session_t p_session; /**< the nmad session */
  const char*local_url;   /**< nmad url for local node */
  struct nm_pmi2_launcher_gate_s
  {
    nm_gate_t p_gate;
    char url[NM_LAUNCHER_URL_SIZE];
    struct nm_drv_vect_s drvs;
  }*gates;                /**< remote nodes, with the url and gate, in the same order as in 'group'. */
};

/* ********************************************************* */

static void*nm_pmi2_launcher_instantiate(puk_instance_t, puk_context_t);
static void nm_pmi2_launcher_destroy(void*_status);

const static struct puk_component_driver_s nm_pmi2_launcher_component =
  {
    .instantiate = &nm_pmi2_launcher_instantiate,
    .destroy     = &nm_pmi2_launcher_destroy
  };

static void nm_pmi2_launcher_init(void*_status, int*argc, char**argv, const char*group_name);
static int  nm_pmi2_launcher_get_size(void*_status);
static int  nm_pmi2_launcher_get_rank(void*_status);
static void nm_pmi2_launcher_get_gates(void*_status, nm_gate_t*gates);
static void nm_pmi2_launcher_abort(void*_status);

const static struct nm_launcher_driver_s nm_pmi2_launcher_driver =
  {
    .init        = &nm_pmi2_launcher_init,
    .get_size    = &nm_pmi2_launcher_get_size,
    .get_rank    = &nm_pmi2_launcher_get_rank,
    .get_gates   = &nm_pmi2_launcher_get_gates,
    .abort       = &nm_pmi2_launcher_abort
  };

PADICO_MODULE_COMPONENT(NewMad_Launcher_pmi2,
  puk_component_declare("NewMad_Launcher_pmi2",
			puk_component_provides("PadicoComponent", "component",  &nm_pmi2_launcher_component),
			puk_component_provides("NewMad_Launcher", "launcher", &nm_pmi2_launcher_driver ))
			);

/* ********************************************************* */

static void*nm_pmi2_launcher_instantiate(puk_instance_t instance, puk_context_t context)
{
  struct nm_pmi2_launcher_status_s*status = padico_malloc(sizeof(struct nm_pmi2_launcher_status_s));
  /* ** init PMI2 */
  int size = padico_pmi2_getsize();
  int rank = padico_pmi2_getrank();
  NM_DISPF("NewMad_Launcher_pmi2: size = %d; rank = %d\n", size, rank);
  status->size = size;
  status->rank = rank;
  /* ** init PadicoTM default attributes */
  if( (getenv("PADICO_NETSELECTOR_inet_0") != NULL) ||
      (getenv("PADICO_NETSELECTOR_default_0") != NULL) ||
      (getenv("PADICO_NETSELECTOR_trk_small_0") != NULL) ||
      (getenv("PADICO_NETSELECTOR_trk_large_0") != NULL) )
    {
      NM_DISPF("pmi2: using user-provided NetSelector config.\n");
    }
  else
    {
      padico_setenv("PADICO_NETSELECTOR_inet_0", "NetSelector-inet-default");
      padico_setenv("PADICO_NETSELECTOR_default_0", "NetSelector-besteffort");
      padico_setenv("PADICO_NETSELECTOR_trk_small_0", "NetSelector-besteffort");
      padico_setenv("PADICO_NETSELECTOR_trk_large_0", "NetSelector-besteffort");
    }
  
  /* ** init state */
  status->gates = NULL;
  status->p_session = NULL;
  return status;
}

static void nm_pmi2_launcher_destroy(void*_status)
{
  struct nm_pmi2_launcher_status_s*status = _status;
  if(status->p_session)
    {
      nm_session_destroy(status->p_session);
      status->p_session = NULL;
    }
  if(status->gates)
    {
      padico_free(status->gates);
      status->gates = NULL;
    }
  padico_free(_status);
}

/* ********************************************************* */

static nm_drv_vect_t nm_pmi2_launcher_selector(const char*url, void*_arg)
{
  struct nm_pmi2_launcher_status_s*status = _arg;
  const int size = status->size;
  int i = 0;
  while(i < size)
    {
      if(strcmp(status->gates[i].url, url) == 0)
	{
	  nm_drv_vect_t p_drvs = nm_drv_vect_copy(&status->gates[i].drvs);
	  return p_drvs;
	}
      i++;
    }
  return NULL;
}

static void nm_pmi2_launcher_init(void*_status, int*argc, char**argv, const char*group_name)
{
  struct nm_pmi2_launcher_status_s*status = _status;
  const int size = status->size;
  const int rank = status->rank;
  status->gates = padico_malloc(size * sizeof(struct nm_pmi2_launcher_gate_s));
  status->gates[rank].p_gate = NM_GATE_NONE;
  nm_session_set_hint_size(size);
  /* ** init session */
  int rc = nm_session_create(&status->p_session, group_name);
  if(rc != NM_ESUCCESS)
    {
      NM_FATAL("NewMad_Launcher_pmi2: nm_session_create()- rc = %d\n", rc);
    }

  /* ** publish local topology */
  padico_string_t s_key = padico_string_new();
  padico_string_printf(s_key, "nm_pmi2_node_%d", rank);
  padico_string_t s_localnode = padico_na_node_serialize(padico_na_getlocalnode());
  padico_pmi2_kvs_publish_multipart(padico_string_get(s_key), padico_string_get(s_localnode));
  padico_string_printf(s_key, "nm_pmi2_host_%d", rank);
  padico_string_t s_localhost = padico_na_host_serialize(padico_na_getlocalhost());
  padico_pmi2_kvs_publish_multipart(padico_string_get(s_key), padico_string_get(s_localhost));
  padico_pmi2_kvs_fence();
  padico_string_delete(s_localnode);
  padico_string_delete(s_localhost);

  const char*nmad_driver = padico_getenv("NMAD_DRIVER");

  if(!nmad_driver)
    {
      puk_hashtable_t loaded_components = puk_hashtable_new_ptr();
      int i;
      for(i = 0; i < size; i++)
	{
          /* ** receive remote topology (node & host) */
          padico_string_printf(s_key, "nm_pmi2_node_%d", i);
          char*remote_topo_node = NULL;
          padico_pmi2_kvs_lookup_multipart(padico_string_get(s_key), &remote_topo_node);
          padico_control_event_disable();
          struct puk_parse_entity_s e = puk_xml_parse_buffer(remote_topo_node, strlen(remote_topo_node), PUK_TRUST_CLUSTER);
          padico_rc_t rc2 = puk_parse_get_rc(&e);
          if(padico_rc_iserror(rc2))
            {
              padico_rc_show(rc2);
            }
          padico_na_node_t peer = puk_parse_get_content(&e);
          padico_control_event_enable();

          padico_string_printf(s_key, "nm_pmi2_host_%d", i);
          char*remote_topo_host = NULL;
          padico_pmi2_kvs_lookup_multipart(padico_string_get(s_key), &remote_topo_host);
          padico_control_event_disable();
          e = puk_xml_parse_buffer(remote_topo_host, strlen(remote_topo_host), PUK_TRUST_CLUSTER);
          rc2 = puk_parse_get_rc(&e);
          if(padico_rc_iserror(rc2))
            {
              padico_rc_show(rc2);
            }

          padico_control_event_enable();
          
	  nm_drv_vect_init(&status->gates[i].drvs);
	  const puk_component_t component_1 =
	    padico_ns_serial_selector(peer, "trk_small", puk_iface_NewMad_minidriver());
	  const puk_component_t component_2 =
	    padico_ns_serial_selector(peer, "trk_large", puk_iface_NewMad_minidriver());
	  assert(component_1 != NULL);
	  assert(component_2 != NULL);
             
	  nm_drv_t p_drv1 = puk_hashtable_lookup(loaded_components, component_1);
	  if(p_drv1 == NULL)
	    {
	      p_drv1 = nm_session_add_driver(component_1, nm_trk_small);
	      puk_hashtable_insert(loaded_components, component_1, p_drv1);
	    }
	  nm_drv_vect_push_back(&status->gates[i].drvs, p_drv1);
	  nm_drv_t p_drv2 = puk_hashtable_lookup(loaded_components, component_2);
	  if(p_drv2 == NULL)
	    {
	      p_drv2 = nm_session_add_driver(component_2, nm_trk_large);
	      puk_hashtable_insert(loaded_components, component_2, p_drv2);
	    }
	  nm_drv_vect_push_back(&status->gates[i].drvs, p_drv2);

          free(remote_topo_node);
          free(remote_topo_host);
	}
      nm_session_set_selector(&nm_pmi2_launcher_selector, status);
      puk_hashtable_delete(loaded_components, NULL);
    }

  rc = nm_session_init(status->p_session, argc, argv, &status->local_url);
  if(rc != NM_ESUCCESS)
    {
      NM_FATAL("NewMad_Launcher_pmi2: nm_session_init()- rc = %d\n", rc);
    }
  NM_DISPF("NewMad_Launcher_pmi2: nm_session_init done- url = %s; size = %d; rank = %d\n",
           status->local_url, size, rank);
  if(strlen(status->local_url) >= NM_LAUNCHER_URL_SIZE)
    {
      NM_FATAL("NewMad_Launcher_pmi2: url is too large (%d / %d).\n", (int)strlen(status->local_url), NM_LAUNCHER_URL_SIZE);
    }

  /* ** publish local url */
  padico_string_printf(s_key, "nm_pmi2_launcher_url_%d", rank);
  padico_pmi2_kvs_publish(padico_string_get(s_key), status->local_url);
  padico_pmi2_kvs_fence();

  /* ** connect gates */
  int steps = 1; /* number of steps = size rounded to power of two */
  while(steps < size)
    {
      steps *= 2;
    }
  int s;
  for(s = 0; s < steps; s++)
    {
      const int dest = rank ^ s; /* pave the (step, rank) matrix with XOR pattern */
      assert(dest >= 0);
      assert(dest <= steps);
      if(dest >= size)
        {
          /* dest may be larger than comm size if size is not a power of two */
          continue;
        }
      else if(dest != rank)
	{
          padico_string_printf(s_key, "nm_pmi2_launcher_url_%d", dest);
          padico_pmi2_kvs_lookup(padico_string_get(s_key), status->gates[dest].url, NM_LAUNCHER_URL_SIZE);
          assert(strcmp(status->gates[dest].url, status->local_url) != 0);
	}
      else
	{
          assert(strlen(status->local_url) < NM_LAUNCHER_URL_SIZE);
	  strncpy(status->gates[dest].url, status->local_url, NM_LAUNCHER_URL_SIZE);
	  status->gates[dest].url[NM_LAUNCHER_URL_SIZE - 1] = '\0';
	}
      rc = nm_session_connect_async(status->p_session, &status->gates[dest].p_gate, status->gates[dest].url);
      if(rc != NM_ESUCCESS)
	{
	  NM_WARN("NewMad_Launcher: nm_session_connect_async()- rc = %d\n", rc);
	}
      padico_out(50, "connection to node #%d done.\n", dest);
    }
  for(s = 0; s < steps; s++)
    {
      const int dest = rank ^ s; /* pave the (step, rank) matrix with XOR pattern */
      assert(dest >= 0);
      assert(dest <= steps);
      if(dest >= size)
        {
          /* dest may be larger than comm size if size is not a power of two */
          continue;
        }
      else
        {
          rc = nm_session_connect_wait(status->p_session, status->gates[dest].p_gate);
          if(rc != NM_ESUCCESS)
            {
              NM_WARN("NewMad_Launcher: nm_session_connect_wait()- rc = %d\n", rc);
            }
        }
    }  
  padico_string_delete(s_key);
}

static int nm_pmi2_launcher_get_size(void*_status)
{
  struct nm_pmi2_launcher_status_s*status = _status;
  return status->size;
}

static int nm_pmi2_launcher_get_rank(void*_status)
{
  struct nm_pmi2_launcher_status_s*status = _status;
  return status->rank;
}

static void nm_pmi2_launcher_get_gates(void*_status, nm_gate_t*gates)
{
  struct nm_pmi2_launcher_status_s*status = _status;
  int i;
  for(i = 0; i < status->size; i++)
    {
      gates[i] = status->gates[i].p_gate;
      padico_out(50, "get_gates()- i = %d; gate = %p\n", i, gates[i]);
    }
}

static void nm_pmi2_launcher_abort(void*_status)
{
  padico_pmi2_abort("NewMad_Launcher_pmi2: abort");
  abort();
}
