/*
 * NewMadeleine
 * Copyright (C) 2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


/* Routines to reorder nodes in a tree.
 *
 * These functions reorder nodes in array according to an array of their
 * respective priority.
 *
 * One major constraint is that nodes in a subtree are contiguous in the array
 * (to easily send these subtrees through the network: it requires only a
 * pointer and a size, without extra copy).
 */

#ifndef NM_MCAST_REORDER_TREE_H
#define NM_MCAST_REORDER_TREE_H

/**
 * Breadth-first search traversal of a binomial tree to reorder nodes according to priorities.
 * - Nodes in `nodes` are sorted in priority-descending order.
 * - Nodes in `ordered_nodes` will be ordered such as our execution of the binomial tree will send
 *   data to nodes respecting priorities (ie: nodes with higher priority will receive data first)
 * - `nb_nodes` is the number of nodes, *including* src node (so nb_dest+1)
 * In `ordered_nodes`, nodes of a subtree are contiguous in the array (to easily send this subtree)
 */
static inline void binomial_reorder_nodes(int* const nodes,
                                          int* const prios,
                                          const int nb_nodes,
                                          int* const ordered_nodes,
                                          int* const ordered_prios)
{
  int child = -1;
  int parent = -1;
  int n_children = 0;
  int ordered_nodes_index = 0;
  struct nm_coll_tree_info_s tree;

  nm_coll_tree_init(&tree, NM_COLL_TREE_BINOMIAL, nb_nodes, 0, 0);

  for (int step = 0; step < tree.height; step++)
    {
      for (int node = 0; node  < nb_nodes-1; node++)
        {
          nm_coll_tree_init(&tree, NM_COLL_TREE_BINOMIAL, nb_nodes, node, 0);
          nm_coll_tree_step(&tree, step, &parent, &child, &n_children);

          if (n_children > 0)
            {
              assert(n_children == 1); // it's a *bi*nomial tree

              ordered_nodes[child-1] = nodes[ordered_nodes_index];
              ordered_prios[child-1] = prios[ordered_nodes_index];
              ordered_nodes_index++;
            }
        }
    }
}

/**
 * Depth first search traversal of a binary tree to reorder nodes according to priorities.
 * - Nodes in `unordered_nodes` are sorted in priority-descending order.
 * - Nodes in `ordered_nodes` will be ordered such as our execution of the binary tree will send
 *   data to nodes respecting priorities (ie: nodes with higher priority will receive data first)
 * - `n` is the number of recipient nodes, so *excluding* src node
 * In `ordered_nodes`, nodes of a subtree are contiguous in the array (to easily send this subtree)
 */
static inline void binary_reorder_nodes(int* const unordered_nodes,
                                        int* const ordered_nodes,
                                        int* const unordered_prios,
                                        int* const ordered_prios,
                                        const int n,
                                        int* const ordered_i,
                                        const int unordered_i)
{
  if (unordered_i > 0)
  {
    ordered_nodes[*ordered_i] = unordered_nodes[unordered_i-1];
    ordered_prios[*ordered_i] = unordered_prios[unordered_i-1];
    (*ordered_i)++;
  }

  if ((2*unordered_i+1) <= n)
  {
    binary_reorder_nodes(unordered_nodes, ordered_nodes, unordered_prios, ordered_prios, n, ordered_i, (2*unordered_i+1));

    if ((2*unordered_i+2) <= n)
    {
      binary_reorder_nodes(unordered_nodes, ordered_nodes, unordered_prios, ordered_prios, n, ordered_i, (2*unordered_i+2));
    }
  }
}

static inline void reorder_nodes(int* dests, int* prios, int* ordered_dests, int* ordered_prios, int n, nm_coll_tree_kind_t tree_type)
{
  assert(tree_type == NM_COLL_TREE_BINARY || tree_type == NM_COLL_TREE_BINOMIAL);
  if(tree_type == NM_COLL_TREE_BINOMIAL)
    {
      binomial_reorder_nodes(dests, prios, n+1, ordered_dests, ordered_prios);
    }
  else if (tree_type == NM_COLL_TREE_BINARY)
    {
      int ordered_i = 0;
      binary_reorder_nodes(dests, ordered_dests, prios, ordered_prios, n, &ordered_i, 0);
    }
  else
    {
      NM_FATAL("The given type of tree does not need reordering.\n");
    }
}

#endif /* NM_MCAST_REORDER_TREE_H */
