/*
 * NewMadeleine
 * Copyright (C) 2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <nm_launcher_interface.h>
#include <nm_coll_interface.h>
#include <nm_rpc_interface.h>

#include "nm_mcast_reorder_trees.h"
#include "nm_mcast_interface.h"

#define MCAST_VERBOSE 0
#define MCAST_INIT_CANARY 1338
#define MCAST_DEFAULT_TREE NM_COLL_TREE_BINOMIAL

#if defined(DEBUG) && (MCAST_VERBOSE == 1)
#define DEBUG_FPRINTF(fd, msg, ...) do { fprintf(fd, msg, ## __VA_ARGS__); } while(0)
#else
#define DEBUG_FPRINTF(msg, ...)
#endif

/** internal status of mcast module */
struct nm_mcast_s
{
  nm_comm_t p_comm;
  nm_rpc_service_t p_service;
};

/** header of RPC requests used for mcast */
struct nm_mcast_header_s
{
  nm_len_t hlen; /** < size of header requested by the user (like if nm_sr_send_header() was called) */
  nm_len_t len;
  nm_tag_t tag;
  nm_seq_t seq;
  nm_session_hash_t session_hash;
  int root_rank;   /** < rank of root sender, in comm_world communicator */
  int n_dests; /** < number of dests in the sub tree */
  char has_prios:1; /** < wether user provided priority for each recipient: tree will be reordered to respect them */
  nm_coll_tree_kind_t tree_type; /** < since each mcast can use a different routing algorithm, it needs to be part of the mcast attributes */
} __attribute__((packed));

/** state of a received mcast request, to share state between callbacks */
struct nm_mcast_recv_s
{
  struct nm_mcast_header_s header;
  nm_rpc_token_t rpc_token;
  struct nm_req_s* req;

  int* dests;
  int* prios;
  nm_seq_t* seqs;

  struct nm_datav_s datav;
  struct nm_data_s data;
};

/** state of a sent request belonging to an mcast, to share state between callbacks */
struct nm_mcast_sub_send_s
{
  int* dests;
  int* prios;
  nm_seq_t* seqs;
  int n_dests; /** < number of dests in the whole multicast (different than header.n_dests) */
  nm_tag_t tag;
  const struct nm_data_s* body_data;
  int tree_step;
  int tree_additionnal_info;
  nm_coll_tree_kind_t tree_type;
  nm_session_hash_t session_hash;
  int is_last; /** < a boolean to tell if this send is the last this node has to perform */
  int root_rank;
  struct nm_mcast_header_s header; /** < we keep the header in this structure because the it needs to be avaible even out of the mcast_send() function */
  int free_dests; /** < if the interface has to free dests and prios itself (the interface can allocate dests and prios itself if it reorders them) */
  struct nm_mcast_recv_s* mcast_recv; /** < when the subsend is a forward (receive, then send) we keep the mcast_recv to finalize the RPC recv */
  nm_len_t hlen; /** < size of header requested by the user (like if nm_sr_send_header() was called) */
  nm_mcast_t* user_mcast;
};



static inline void mcast_send(struct nm_mcast_sub_send_s* previous_sub_send);

/* ********************************************************* */

static const nm_tag_t mcast_tag = 0x3dc; // Multicast for IP was introduced in RFC 988 :)
static nm_coll_tree_kind_t default_tree_type = MCAST_DEFAULT_TREE;
static struct nm_mcast_s mcast = { .p_comm = NULL, .p_service = NULL };
static int my_rank = -1;

/* ********************************************************* */

/** This function is called when an injected chunk is matched with a user request, thus
 * we can actually post the receive to receive data in place.
 * This function can also by called during peeking (nm_sr_recv_peek()), so it
 * can be called several times for the same request.
 */
static void nm_mcast_notify_req(struct nm_req_s* req, const struct nm_data_s* p_data0, nm_len_t chunk_offset, nm_len_t chunk_len, void* p_ref)
{
  struct nm_mcast_recv_s* mcast_recv = (struct nm_mcast_recv_s*) p_ref;
  DEBUG_FPRINTF(stderr, "[%d] # nm_mcast_notify_req()- req = %p; chunk_offset = %ld; chunk_len = %ld (data size = %ld)\n",
          my_rank, req, chunk_offset, chunk_len, nm_data_size(p_data0));
  mcast_recv->req = req;
  struct nm_data_s data;
  const struct nm_data_s* p_data = NULL;

  if((chunk_offset > 0) || (chunk_len < nm_data_size(p_data0)))
    {
      /* adjust data to the actual chunk length & offset */
      nm_data_excerpt_build(&data, (struct nm_data_s*)p_data0, chunk_offset, chunk_len);
      p_data = &data;
    }
  else
    {
      p_data = p_data0;
    }

  if(chunk_offset + chunk_len <= mcast_recv->header.hlen)
    {
      /* receive the user header (specified by nm_rpc_req_set_hlen(), different
       * from nm_rpc_req_pack_header() !). Data is available immediately. */
      DEBUG_FPRINTF(stderr, "[%d] # nm_mcast_notify_req()- receiving as header\n", my_rank);
      nm_rpc_recv_header_data(mcast_recv->rpc_token, (struct nm_data_s*) p_data);
    }
  else
    {
      /* receive data body. Data will be available after the finalizer has been called. */
      DEBUG_FPRINTF(stderr, "[%d] # nm_mcast_notify_req()- receiving as body\n", my_rank);
      nm_rpc_irecv_body_data(mcast_recv->rpc_token, (struct nm_data_s*) p_data);
    }

  // Save all received chunks to an nm_datav to then have an nm_data to forward:
  nm_datav_add_chunk_data(&mcast_recv->datav, p_data);

  if(chunk_offset + chunk_len >= mcast_recv->header.len)
    {
      /* notify that all receive requests have been posted for this RPC incoming request.
      * (needed because we called nm_rpc_token_delay()) */
      DEBUG_FPRINTF(stderr, "[%d] # nm_mcast_notify_req()- signaling token complete\n", my_rank);
      nm_data_datav_build(&mcast_recv->data, &mcast_recv->datav);
      nm_rpc_token_complete(mcast_recv->rpc_token);
    }

  DEBUG_FPRINTF(stderr, "[%d] # nm_mcast_notify_req()- done.\n", my_rank);
}

/* RPC handler: this function is called once for each RPC message received. */
static void nm_mcast_rpc_handler(nm_rpc_token_t rpc_token)
{
  struct nm_core* p_core = nm_comm_get_session(mcast.p_comm)->p_core;
  /* initialize & fill a mcast packet descriptor */
  struct nm_mcast_recv_s* mcast_recv = malloc(sizeof(struct nm_mcast_recv_s));
  nm_rpc_token_set_ref(rpc_token, mcast_recv);
  mcast_recv->header.len = NM_LEN_UNDEFINED;
  mcast_recv->rpc_token = rpc_token;
  mcast_recv->req = NULL;
  mcast_recv->dests = NULL;
  mcast_recv->prios = NULL;
  nm_datav_init(&mcast_recv->datav);

  /* receive mcast header (set by nm_rpc_req_pack_header()) */
  nm_rpc_recv_header(mcast_recv->rpc_token, &mcast_recv->header, sizeof(mcast_recv->header));
  DEBUG_FPRINTF(stderr, "[%d] # nm_mcast_rpc_handler()- received header: msg len = %lu; seq = %d; hlen = %ld; from = %d; n_dests = %d; tag = %ld\n",
         my_rank, mcast_recv->header.len, mcast_recv->header.seq, mcast_recv->header.hlen, mcast_recv->header.root_rank, mcast_recv->header.n_dests, mcast_recv->header.tag);

  if (mcast_recv->header.n_dests > 0)
    {
      mcast_recv->dests = malloc(mcast_recv->header.n_dests * sizeof(int));
      nm_rpc_recv_header(mcast_recv->rpc_token, mcast_recv->dests, mcast_recv->header.n_dests * sizeof(int));

      mcast_recv->seqs = malloc(mcast_recv->header.n_dests * sizeof(nm_seq_t));
      nm_rpc_recv_header(mcast_recv->rpc_token, mcast_recv->seqs, mcast_recv->header.n_dests * sizeof(nm_seq_t));

      if (mcast_recv->header.has_prios)
        {
          mcast_recv->prios = malloc(mcast_recv->header.n_dests * sizeof(int));
          nm_rpc_recv_header(mcast_recv->rpc_token, mcast_recv->prios, mcast_recv->header.n_dests * sizeof(int));
        }
    }

  /* Signal this RPC handler didn't post any recv yet (and maybe won't
   * post any recv in the handler at all).
   * Warning- it needs to be before nm_core_inject_chunk(), since they
   * may trigger (or not) immediate completion and unlock the token.
   * Since we used nm_rpc_token_delay(), we will have to explicitely
   * call nm_rpc_token_complete() to notify the RPC service that we
   * posted all receives for this token.
   */
  nm_rpc_token_delay(mcast_recv->rpc_token);

  nm_gate_t p_gate = nm_comm_get_gate(mcast.p_comm, mcast_recv->header.root_rank);
  nm_core_tag_t core_tag = nm_core_tag_build(mcast_recv->header.session_hash, mcast_recv->header.tag);
  if (mcast_recv->header.hlen > 0)
    {
      /* user asked for a non-empty header; inject the header separately, to make it
      * immediately available, without waiting for the body.
      * Here, is_last_chunk is equivalent to "the whole message contains only a
      * header" */
      const int is_last_chunk = mcast_recv->header.hlen == mcast_recv->header.len;
      nm_core_inject_chunk(p_core, p_gate, core_tag, mcast_recv->header.seq,
                           0 /* chunk_offset */, mcast_recv->header.hlen, is_last_chunk,
                           &nm_mcast_notify_req, mcast_recv);
    }

  if(mcast_recv->header.len > mcast_recv->header.hlen)
    {
      /* inject the data body in the core, if present. The injected chunk is
         only a placeholder, data won't be available before the rpc finalizer
         is called */
      nm_core_inject_chunk(p_core, p_gate, core_tag, mcast_recv->header.seq,
                           mcast_recv->header.hlen, mcast_recv->header.len -  mcast_recv->header.hlen, 1 /* last chunk */,
                           &nm_mcast_notify_req, mcast_recv);
    }

  DEBUG_FPRINTF(stderr, "[%d] # nm_mcast_rpc_handler()- done.\n", my_rank);
}

/* called when we finished receiving data from a broadcast: data is received
 * and forwarded, user can do whatever he wants with it. */
static inline void mcast_recv_finalize(struct nm_mcast_recv_s* mcast_recv)
{
  /* notify the core injector that the receive is now finalized */
  nm_core_inject_finalize(nm_comm_get_session(mcast.p_comm)->p_core, mcast_recv->req);

  nm_datav_destroy(&mcast_recv->datav);

  if (mcast_recv->dests != NULL)
    {
      free(mcast_recv->dests);
      free(mcast_recv->seqs);
      if (mcast_recv->prios != NULL)
        {
          free(mcast_recv->prios);
        }
    }
  free(mcast_recv);
}

/* RPC finalizer: this function is called when the RPC is fully received: ie
 * when data from the multicast is completely received, maybe we will forward
 * to other nodes */
static void nm_mcast_rpc_finalizer(nm_rpc_token_t rpc_token)
{
  DEBUG_FPRINTF(stderr, "[%d] # nm_mcast_rpc_finalize()\n", my_rank);

  struct nm_mcast_recv_s* mcast_recv = (struct nm_mcast_recv_s*) nm_rpc_token_get_ref(rpc_token);

  /* notify the core injector that data for this request has arrived, but we
   * still need it for potential forwards (user receive will have status UNPACK_COMPLETED) */
  nm_core_inject_complete(nm_comm_get_session(mcast.p_comm)->p_core, mcast_recv->req, mcast_recv->header.len);

  if (mcast_recv->header.n_dests > 0)
    {
      struct nm_mcast_sub_send_s* sub_send = malloc(sizeof(struct nm_mcast_sub_send_s));
      sub_send->dests = mcast_recv->dests;
      sub_send->prios = mcast_recv->prios;
      sub_send->seqs = mcast_recv->seqs;
      sub_send->n_dests = mcast_recv->header.n_dests;
      sub_send->tag = mcast_recv->header.tag;
      sub_send->body_data = &mcast_recv->data;
      sub_send->tree_step = 0;
      sub_send->tree_additionnal_info = sub_send->n_dests+1;
      sub_send->tree_type = mcast_recv->header.tree_type;
      sub_send->session_hash = mcast_recv->header.session_hash;
      sub_send->root_rank = mcast_recv->header.root_rank;
      sub_send->user_mcast = NULL;
      sub_send->free_dests = 0;
      sub_send->mcast_recv = mcast_recv;
      sub_send->hlen = mcast_recv->header.hlen;

      mcast_send(sub_send);
    }
  else
    {
      mcast_recv_finalize(mcast_recv);
    }
}


/** called when an RPC send is finished: according to the routing algorithm,
 * maybe we need to perform another RPC send. */
void send_monitor(nm_rpc_req_t rpc_req, void* _ref)
{
  struct nm_mcast_sub_send_s* sub_send = (struct nm_mcast_sub_send_s*) _ref;

  DEBUG_FPRINTF(stderr, "# RPC send from %d with tag %ld is over\n", my_rank, sub_send->tag);

  if (sub_send->is_last)
    {
      if (sub_send->mcast_recv != NULL)
        {
          /* It was the last subsend of forwards of data we received from
           * another node: we need to finalize the RPC receive. */
          assert(sub_send->root_rank != my_rank);

          mcast_recv_finalize(sub_send->mcast_recv);
        }
      else
        {
          /* It was the last subsend the root node was in charge. */
          assert(sub_send->root_rank == my_rank);

          if (sub_send->user_mcast != NULL)
            {
              assert(sub_send->user_mcast->initialized == MCAST_INIT_CANARY);

              if (sub_send->user_mcast->notifier == NULL)
                {
                  nm_cond_signal(&sub_send->user_mcast->cond_mcast_ended, 1);
                }
              else
                {
                  sub_send->user_mcast->notifier(sub_send->user_mcast->notifier_ref);
                }

              /* We can do nothing about user_mcast after this, because user may have destroyed it in the above callback or after the cond_signal. */
              sub_send->user_mcast = NULL;
            }

          free(sub_send->seqs);

          if (sub_send->free_dests)
            {
              free(sub_send->dests);

              /* we free dests and prios only if we allocated them ourselves
              * becauses we reorder them according to priorities (so prio is
              * allocated !): */
              assert(sub_send->prios != NULL);

              free(sub_send->prios);
            }
        }

      free(sub_send);
    }
  else
    {
      sub_send->tree_step++;
      mcast_send(sub_send);
    }
}

// Perform one RPC send of the diffusion algorithm:
static inline void mcast_send(struct nm_mcast_sub_send_s* previous_sub_send)
{
  int dest_id = -1;
  int nb_dests_in_forward = -1;
  char is_local_last_send = -1;
  void* first_node_in_forward_list = NULL;
  void* first_prio_in_forward_list = NULL;
  void* first_seq_in_forward_list = NULL;
  int new_tree_additionnal_info = -1;

  if (previous_sub_send->tree_type == NM_COLL_TREE_CHAIN)
    {
      dest_id = 0;
      nb_dests_in_forward = previous_sub_send->n_dests-1;
      first_node_in_forward_list = (void*) (previous_sub_send->dests + 1);
      first_prio_in_forward_list = (void*) (previous_sub_send->prios + 1);
      first_seq_in_forward_list = (void*) (previous_sub_send->seqs + 1);
      is_local_last_send = 1;
      new_tree_additionnal_info = 0; // unused in this kind of tree
    }
  else
    {
      struct nm_coll_tree_info_s tree;
      nm_coll_tree_init(&tree, previous_sub_send->tree_type, previous_sub_send->n_dests+1, 0, 0);

      if (previous_sub_send->tree_type == NM_COLL_TREE_BINOMIAL)
        {
          assert(tree.max_arity <= 1);

          int parent = -1; // unused, but required by nm_coll_tree_step
          int n_children = 0;

          nm_coll_tree_step(&tree, previous_sub_send->tree_step, &parent, &dest_id, &n_children);

          assert(n_children == 1);

          dest_id--;
          // tree_additionnal_info is separator id in dests array
          nb_dests_in_forward = previous_sub_send->tree_additionnal_info - dest_id - 2;

          first_node_in_forward_list = (void*) (previous_sub_send->dests + dest_id + 1);
          first_prio_in_forward_list = (void*) (previous_sub_send->prios + dest_id + 1);
          first_seq_in_forward_list = (void*) (previous_sub_send->seqs + dest_id + 1);
          is_local_last_send = (previous_sub_send->tree_step == (tree.height-1));
          new_tree_additionnal_info = dest_id + 1; // not working when n_children > 1
        }
      else if (previous_sub_send->tree_type == NM_COLL_TREE_BINARY)
        {
          assert(tree.max_arity == 2);

          if (previous_sub_send->tree_step == 0) // left subtree
            {
              dest_id = 0;
              first_node_in_forward_list = (void*) (previous_sub_send->dests + 1);
              first_prio_in_forward_list = (void*) (previous_sub_send->prios + 1);
              first_seq_in_forward_list = (void*) (previous_sub_send->seqs + 1);
              int right_dest_id = nm_coll_binary_nb_nodes_in_left_subtree(&tree);
              nb_dests_in_forward = right_dest_id - 1;
              is_local_last_send = (right_dest_id >= previous_sub_send->n_dests);
              new_tree_additionnal_info = right_dest_id;
            }
          else if (previous_sub_send->tree_step == 1) // right subtree
            {
              dest_id = previous_sub_send->tree_additionnal_info;
              nb_dests_in_forward = previous_sub_send->n_dests - dest_id - 1;
              first_node_in_forward_list = (void*) (previous_sub_send->dests + dest_id + 1);
              first_prio_in_forward_list = (void*) (previous_sub_send->prios + dest_id + 1);
              first_seq_in_forward_list = (void*) (previous_sub_send->seqs + dest_id + 1);
              is_local_last_send = 1;
            }
          else
            {
              NM_FATAL("Binary tree cannot have more than two steps !\n");
            }
        }
      else
        {
          NM_FATAL("Unsupported type of tree\n");
        }
    }

  assert(dest_id > -1);
  assert(nb_dests_in_forward > -1);
  assert(first_node_in_forward_list != NULL);
  assert(first_prio_in_forward_list != NULL);
  assert(first_seq_in_forward_list != NULL);
  assert(is_local_last_send >= 0);
  assert(is_local_last_send == 1 || new_tree_additionnal_info >= 0);

  struct nm_mcast_sub_send_s* sub_send = malloc(sizeof(struct nm_mcast_sub_send_s));
  memcpy(sub_send, previous_sub_send, sizeof(struct nm_mcast_sub_send_s));
  sub_send->is_last = is_local_last_send;
  sub_send->tree_additionnal_info = new_tree_additionnal_info;

  const nm_session_t session = nm_session_lookup(sub_send->session_hash);
  const nm_comm_t comm = nm_comm_get_by_session(session);
  const nm_gate_t gate = nm_comm_get_gate(comm, sub_send->dests[dest_id]);
  assert(gate != NULL);
  DEBUG_FPRINTF(stderr, "[%d] sends to %d (tag %ld) with seq %d with %d nodes in the subtree type %d\n", my_rank, sub_send->dests[dest_id], sub_send->tag, sub_send->seqs[dest_id], nb_dests_in_forward, sub_send->tree_type);

  sub_send->header.hlen         = sub_send->hlen;
  sub_send->header.len          = nm_data_size(sub_send->body_data);
  sub_send->header.tag          = sub_send->tag;
  sub_send->header.seq          = sub_send->seqs[dest_id];
  sub_send->header.session_hash = session->hash_code;
  sub_send->header.root_rank    = sub_send->root_rank;
  sub_send->header.n_dests      = nb_dests_in_forward; // number of dests in the subtree
  sub_send->header.has_prios    = (sub_send->prios != NULL);
  sub_send->header.tree_type    = sub_send->tree_type;

  nm_rpc_req_t rpc_req = nm_rpc_req_init(mcast.p_service, gate, mcast_tag);
  nm_rpc_req_pack_header(rpc_req, &sub_send->header, sizeof(struct nm_mcast_header_s));
  nm_rpc_req_set_hlen(rpc_req, sub_send->header.hlen);
  if (sub_send->header.n_dests > 0)
    {
      nm_rpc_req_pack_header(rpc_req, first_node_in_forward_list, sub_send->header.n_dests*sizeof(int));
      nm_rpc_req_pack_header(rpc_req, first_seq_in_forward_list, sub_send->header.n_dests*sizeof(nm_seq_t));
      if (sub_send->header.has_prios)
        {
          nm_rpc_req_pack_header(rpc_req, first_prio_in_forward_list, sub_send->header.n_dests*sizeof(int));
        }
    }
  /* We pack only the user data as a body, to allow the user to peek on this data during the recv: */
  nm_rpc_req_pack_body_data(rpc_req, sub_send->body_data);
  if (sub_send->prios != NULL)
    {
      nm_rpc_req_set_priority(rpc_req, sub_send->prios[dest_id]);
    }
  nm_rpc_req_set_notifier(rpc_req, &send_monitor, sub_send);
  nm_rpc_req_isend(rpc_req);

  free(previous_sub_send);
}

/* ********************************************************* */

/** initialize the mcast interface */
void nm_mcast_init(void)
{
  nm_launcher_get_rank(&my_rank);

  mcast.p_comm = nm_comm_world("nm_mcast_private"); /* private communicator used for RPCs */
  mcast.p_service = nm_rpc_register(nm_comm_get_session(mcast.p_comm), mcast_tag, NM_TAG_MASK_FULL,
                                    &nm_mcast_rpc_handler, &nm_mcast_rpc_finalizer, mcast.p_comm);

  const char* s_mcast_tree = getenv("NMAD_MCAST_TREE");
  if (s_mcast_tree)
    {
      if (strcmp(s_mcast_tree, "binary") == 0)
        {
          default_tree_type = NM_COLL_TREE_BINARY;
        }
      else if (strcmp(s_mcast_tree, "chain") == 0)
        {
          default_tree_type = NM_COLL_TREE_CHAIN;
        }
      else
        {
          if (strcmp(s_mcast_tree, "binomial") != 0)
            {
              NM_WARN("Only binary, binomial and chain trees are supported for mcast. Will use binomial instead.\n");
            }
          default_tree_type = MCAST_DEFAULT_TREE;
        }
    }
}

/** destroys mcast interface */
void nm_mcast_finalize(void)
{
  nm_rpc_unregister(mcast.p_service);
  nm_comm_destroy(mcast.p_comm);
}

void nm_mcast_send_init(nm_mcast_t* _mcast)
{
  if(mcast.p_service == NULL)
    {
      NM_FATAL("mcast: interface not initialized.\n");
    }
  assert(_mcast != NULL);

  memset(_mcast, 0, sizeof(nm_mcast_t));

  nm_cond_init(&_mcast->cond_mcast_ended, 0);
  _mcast->initialized = MCAST_INIT_CANARY;
}

void nm_mcast_send_set_notifier(nm_mcast_t* _mcast, nm_mcast_send_notifier_t notifier, void* ref)
{
  assert(_mcast != NULL);
  assert(_mcast->initialized == MCAST_INIT_CANARY);

  _mcast->notifier = notifier;
  _mcast->notifier_ref = ref;
}

/** send an mcast message on the given communicator, to the list of given ranks.
 * @note only comm/ranks addressing is available instead of gates, since addresses
 * need to be serializable on the network.
 * @param hlen header length, i.e. part of length of p_data that needs to be
 * sent eagerly so as to be immediately available in the receiver handler. Set
 * to 0 if not needed.
 */
void nm_mcast_isend(nm_mcast_t* _mcast,
                   nm_comm_t p_comm,
                   int* dests,
                   int* prios,
                   int n_dests,
                   nm_tag_t tag,
                   struct nm_data_s* p_data,
                   nm_len_t hlen,
                   nm_coll_tree_kind_t tree_type)
{
  if (tree_type == NM_COLL_TREE_DEFAULT)
    {
      tree_type = default_tree_type;
    }
  else if (tree_type != NM_COLL_TREE_BINOMIAL && tree_type != NM_COLL_TREE_BINARY && tree_type != NM_COLL_TREE_CHAIN)
    {
      NM_WARN("Unsupported type of tree for mcast, will fallback to default.");
      tree_type = default_tree_type;
    }

  int free_dests = 0;

  if (prios != NULL && (tree_type == NM_COLL_TREE_BINARY || tree_type == NM_COLL_TREE_BINOMIAL))
    {
      /** We reorder node list in order to nodes with higher priority will
       * receive their data first.
       * The only requirement is that array `prios` has to be sorted in
       * descending order (and array `nodes` ordered such as prios[i] is the
       * priority of the send to the node nodes[i])
       * In `ordered_nodes`, nodes of a subtree are contiguous in the array (to
       * easily send this subtree) */
      int* ordered_dests = malloc(n_dests * sizeof(int));
      int* ordered_prios = malloc(n_dests * sizeof(int));
      reorder_nodes(dests, prios, ordered_dests, ordered_prios, n_dests, tree_type);
      dests = ordered_dests;
      prios = ordered_prios;
      free_dests = 1;
    }

  struct nm_mcast_sub_send_s* sub_send = malloc(sizeof(struct nm_mcast_sub_send_s));
  sub_send->dests = dests;
  sub_send->prios = prios;
  sub_send->n_dests = n_dests;
  sub_send->tag = tag;
  sub_send->body_data = p_data;
  sub_send->tree_step = 0;
  sub_send->tree_additionnal_info = sub_send->n_dests+1;
  sub_send->tree_type = tree_type;
  sub_send->session_hash = nm_comm_get_session(p_comm)->hash_code;
  sub_send->root_rank = my_rank;
  sub_send->user_mcast = NULL;
  sub_send->free_dests = free_dests;
  sub_send->mcast_recv = NULL;
  sub_send->hlen = hlen;
  sub_send->seqs = malloc(n_dests * sizeof(nm_seq_t));

  // Sequence numbers has to be generated by the root rank of the mutlicast:
  struct nm_core* p_core = nm_comm_get_session(mcast.p_comm)->p_core;
  int i = 0;
  for (i = 0; i < n_dests; i++)
    {
      const nm_core_tag_t core_tag = nm_core_tag_build(sub_send->session_hash, sub_send->tag);
      const nm_gate_t gate = nm_comm_get_gate(p_comm, sub_send->dests[i]);
      sub_send->seqs[i] = nm_core_send_seq_get(p_core, gate, core_tag);
    }

  if (_mcast != NULL)
    {
      assert(_mcast->initialized == MCAST_INIT_CANARY);
      sub_send->user_mcast = _mcast;
    }

  mcast_send(sub_send);
}

void nm_mcast_send(nm_comm_t p_comm,
                   int* dests,
                   int* prios,
                   int n_dests,
                   nm_tag_t tag,
                   struct nm_data_s* p_data,
                   nm_len_t hlen,
                   nm_coll_tree_kind_t tree_type)
{
  nm_mcast_t this_mcast;
  nm_mcast_send_init(&this_mcast);
  nm_mcast_isend(&this_mcast, p_comm, dests, prios, n_dests, tag, p_data, hlen, tree_type);
  nm_mcast_wait(&this_mcast);
  nm_mcast_send_destroy(&this_mcast);
}

void nm_mcast_wait(nm_mcast_t* _mcast)
{
  assert(_mcast != NULL);
  assert(_mcast->initialized == MCAST_INIT_CANARY);

  if (_mcast->notifier != NULL)
    {
      // To be consistent with the behavior of the sendrecv interface:
      NM_FATAL("Cannot wait on this mcast: a notifier is defined.\n");
    }

  nm_cond_wait(&_mcast->cond_mcast_ended, 1, nm_comm_get_session(mcast.p_comm)->p_core);
}

void nm_mcast_send_destroy(nm_mcast_t* _mcast)
{
  assert(_mcast != NULL);
  assert(_mcast->initialized == MCAST_INIT_CANARY);

  nm_cond_destroy(&_mcast->cond_mcast_ended);
  _mcast->initialized = -1;
}

void nm_mcast_set_default_tree_kind(nm_coll_tree_kind_t kind)
{
  if (kind != NM_COLL_TREE_BINARY && kind != NM_COLL_TREE_BINOMIAL && kind != NM_COLL_TREE_CHAIN)
    {
      NM_WARN("Only binary, binomial and chain trees are supported for mcast. Will use binomial instead.\n");
      default_tree_type = NM_COLL_TREE_BINOMIAL;
    }
  else
    {
      default_tree_type = kind;
    }
}
