/*
 * NewMadeleine
 * Copyright (C) 2006-2020 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#include "nm_mpi_private.h"
#include <assert.h>
#include <errno.h>

#include <Padico/Puk.h>
#include <Padico/Module.h>
PADICO_MODULE_BUILTIN(MadMPI, NULL, NULL, NULL);

static int init_done = 0, finalize_done = 0;
static int thread_level = 0;
#ifdef NMAD_PROFILE
__PUK_SYM_INTERNAL struct nm_mpi_profiling_s nm_mpi_profile = { 0 };
#endif /* NMAD_PROFILE */

NM_MPI_HANDLE_TYPE(info, struct nm_mpi_info_s, _NM_MPI_INFO_OFFSET, 8);
static struct nm_mpi_handle_info_s nm_mpi_infos;

/* ********************************************************* */
/* Aliases */

NM_MPI_ALIAS(MPI_Init, mpi_init);
NM_MPI_ALIAS(MPI_Init_thread, mpi_init_thread);
NM_MPI_ALIAS(MPI_Initialized, mpi_initialized);
NM_MPI_ALIAS(MPI_Query_thread, mpi_query_thread);
NM_MPI_ALIAS(MPI_Finalize, mpi_finalize);
NM_MPI_ALIAS(MPI_Finalized, mpi_finalized);
NM_MPI_ALIAS(MPI_Abort, mpi_abort);
NM_MPI_ALIAS(MPI_Get_processor_name, mpi_get_processor_name);
NM_MPI_ALIAS(MPI_Wtime, mpi_wtime);
NM_MPI_ALIAS(MPI_Wtick, mpi_wtick);
NM_MPI_ALIAS(MPI_Get_version, mpi_get_version);
NM_MPI_ALIAS(MPI_Get_library_version, mpi_get_library_version);
NM_MPI_ALIAS(MPI_Get_count, mpi_get_count);
NM_MPI_ALIAS(MPI_Get_elements, mpi_get_elements);
NM_MPI_ALIAS(MPI_Get_elements_x, mpi_get_elements_x);
NM_MPI_ALIAS(MPI_Get_address, mpi_get_address);
NM_MPI_ALIAS(MPI_Address, mpi_address);
NM_MPI_ALIAS(MPI_Alloc_mem, mpi_alloc_mem);
NM_MPI_ALIAS(MPI_Free_mem, mpi_free_mem);
NM_MPI_ALIAS(MPI_Info_create, mpi_info_create);
NM_MPI_ALIAS(MPI_Info_free, mpi_info_free);
NM_MPI_ALIAS(MPI_Info_set, mpi_info_set);
NM_MPI_ALIAS(MPI_Info_get, mpi_info_get);
NM_MPI_ALIAS(MPI_Info_delete, mpi_info_delete);
NM_MPI_ALIAS(MPI_Info_get_nkeys, mpi_info_get_nkeys);
NM_MPI_ALIAS(MPI_Info_get_nthkey, mpi_info_get_nthkey);
NM_MPI_ALIAS(MPI_Info_get_valuelen, mpi_info_get_valuelen);
NM_MPI_ALIAS(MPI_Info_dup, mpi_info_dup);

/** stub implementation to be overriden by profiling library */
int MPI_Pcontrol(const int level, ...) __attribute__((weak));
NM_MPI_ALIAS(MPI_Pcontrol, mpi_pcontrol);

/* ********************************************************* */


__PUK_SYM_INTERNAL
struct nm_mpi_info_s*nm_mpi_info_alloc(void)
{
  struct nm_mpi_info_s*p_info = nm_mpi_handle_info_alloc(&nm_mpi_infos);
  nm_mpi_info_hashtable_init(&p_info->content);
  return p_info;
}

 __PUK_SYM_INTERNAL
 void nm_mpi_info_destructor(char*p_key, char*p_data)
{
  free(p_key);
  free(p_data);
}

__PUK_SYM_INTERNAL
void nm_mpi_info_free(struct nm_mpi_info_s*p_info)
{
  nm_mpi_info_hashtable_destroy(&p_info->content);
  nm_mpi_handle_info_free(&nm_mpi_infos, p_info);
}

__PUK_SYM_INTERNAL
struct nm_mpi_info_s*nm_mpi_info_get(MPI_Info info)
{
  struct nm_mpi_info_s*p_info = nm_mpi_handle_info_get(&nm_mpi_infos, info);
  return p_info;
}

__PUK_SYM_INTERNAL
void nm_mpi_info_define(struct nm_mpi_info_s*p_info, char*p_key, char*p_value)
{
  if(nm_mpi_info_hashtable_probe(&p_info->content, p_key))
    {
      char*oldkey = NULL, *oldvalue = NULL;
      nm_mpi_info_hashtable_lookup2(&p_info->content, p_key, &oldkey, &oldvalue);
      nm_mpi_info_hashtable_remove(&p_info->content, p_key);
      free(oldkey);
      free(oldvalue);
    }
  nm_mpi_info_hashtable_insert(&p_info->content, strdup(p_key), strdup(p_value));
}

__PUK_SYM_INTERNAL
struct nm_mpi_info_s*nm_mpi_info_dup(struct nm_mpi_info_s*p_info)
{
  int i;
  struct nm_mpi_info_s*p_info_dup = nm_mpi_info_alloc();
  nm_mpi_info_hashtable_enumerator_t e = nm_mpi_info_hashtable_enumerator_new(&p_info->content);
  char*key, *data;
  for(i = 0; i < nm_mpi_info_hashtable_size(&p_info->content); ++i)
    {
      nm_mpi_info_hashtable_enumerator_next2(e, &key, &data);
      nm_mpi_info_hashtable_insert(&p_info_dup->content, strdup(key), strdup(data));
    }
  return p_info_dup;
}

__PUK_SYM_INTERNAL
void nm_mpi_info_update(struct nm_mpi_info_s*p_info_up, struct nm_mpi_info_s*p_info_origin)
{
  assert(p_info_origin);
  if(NULL == p_info_up) return;
  int i;
  nm_mpi_info_hashtable_enumerator_t e = nm_mpi_info_hashtable_enumerator_new(&p_info_up->content);
  char*key, *data;
  for(i = 0; i < nm_mpi_info_hashtable_size(&p_info_up->content); ++i)
    {
      nm_mpi_info_hashtable_enumerator_next2(e, &key, &data);
      if(nm_mpi_info_hashtable_probe(&p_info_origin->content, key))
	{
	  char*oldkey = NULL, *oldvalue = NULL;
	  nm_mpi_info_hashtable_lookup2(&p_info_origin->content, key, &oldkey, &oldvalue);
	  nm_mpi_info_hashtable_remove(&p_info_origin->content, key);
	  free(oldkey);
	  free(oldvalue);
	  nm_mpi_info_hashtable_insert(&p_info_origin->content, strdup(key), strdup(data));
	}
    }
}

/* ********************************************************* */

int mpi_init(int*argc, char***argv)
{
  nm_mpi_handle_info_init(&nm_mpi_infos);
  nm_mpi_err_init();
  nm_launcher_init(argc, argv ? *argv : NULL);
  nm_mpi_attrs_init();
  nm_mpi_group_init();
  nm_mpi_comm_init();
  nm_mpi_datatype_init();
  nm_mpi_op_init();
  nm_mpi_request_init();
  nm_mpi_io_init();
  nm_mpi_win_init();
  nm_mpi_rma_init();
#if defined(PIOMAN) && defined(PIOMAN_MULTITHREAD)
  thread_level = MPI_THREAD_MULTIPLE;
#else
  thread_level = MPI_THREAD_FUNNELED;
#endif
  init_done = 1;
  return MPI_SUCCESS;
}

int mpi_init_thread(int*argc, char***argv, int required, int*provided)
{
  int err;
  err = mpi_init(argc, argv);
  *provided = (thread_level > required) ? required : thread_level;
  return err;
}

int mpi_query_thread(int*provided)
{
  if(!init_done)
    {
      *provided = MPI_UNDEFINED;
      return MPI_ERR_OTHER;
    }
  *provided = thread_level;
  return MPI_SUCCESS;
}

int mpi_initialized(int*flag)
{
  *flag = init_done;
  return MPI_SUCCESS;
}

int mpi_finalized(int*flag)
{
  *flag = finalize_done;
  return MPI_SUCCESS;
}

int mpi_finalize(void)
{
  if(finalize_done)
    {
      NM_MPI_WARNING("MPI_Finalize() called multiple time.\n");
      return MPI_ERR_OTHER;
    }
#ifdef NMAD_PROFILE
  int rank = -1;
  nm_launcher_get_rank(&rank);
  fprintf(stderr, "# ## MadMPI profiling for rank #%d__________\n", rank);
  fprintf(stderr, "# max_req_send = %d; max_req_recv = %d; max_req_total = %d\n",
          nm_mpi_profile.max_req_send, nm_mpi_profile.max_req_recv, nm_mpi_profile.max_req_total);
#endif /* NMAD_PROFILE */
  mpi_barrier(MPI_COMM_WORLD);
  nm_mpi_rma_exit();
  nm_mpi_win_exit();
  nm_mpi_io_exit();
  nm_mpi_request_exit();
  nm_mpi_datatype_exit();
  nm_mpi_op_exit();
  nm_mpi_comm_exit();
  nm_mpi_group_exit();
  nm_mpi_attrs_exit();
  nm_mpi_err_exit();
  nm_mpi_handle_info_finalize(&nm_mpi_infos, NULL);
  init_done = 0;
  finalize_done = 1;
  nm_launcher_exit();
  return MPI_SUCCESS;
}

int mpi_abort(MPI_Comm comm __attribute__((unused)), int errorcode)
{
  fprintf(stderr, "# madmpi: FATAL- MPI_Abort err = %d\n", errorcode);
  nm_launcher_abort();
  abort();
  return errorcode;
}


int mpi_get_processor_name(char *name, int *resultlen)
{
  int err;
  err = gethostname(name, MPI_MAX_PROCESSOR_NAME);
  if (!err)
    {
      *resultlen = strlen(name);
      return MPI_SUCCESS;
    }
  else
    {
      return errno;
    }
}

double mpi_wtime(void)
{
  static puk_tick_t orig_time;
  static int orig_done = 0;
  if(!orig_done)
    {
      PUK_GET_TICK(orig_time);
      orig_done = 1;
    }
  puk_tick_t time;
  PUK_GET_TICK(time);
  const double usec = PUK_TIMING_DELAY(orig_time, time);
  return usec / 1000000.0;
}

double mpi_wtick(void) {
  return 1e-7;
}

int mpi_get_version(int *version, int *subversion)
{
  *version = MADMPI_VERSION;
  *subversion = MADMPI_SUBVERSION;
  return MPI_SUCCESS;
}

int mpi_get_library_version(char *version, int *resultlen)
{
  sprintf(version,"%s",NMAD_VERSION_STRING);
  *resultlen = strlen(version);

  return MPI_SUCCESS;
}

int mpi_get_count(MPI_Status*status, MPI_Datatype datatype, int*count)
{
  nm_mpi_datatype_t*p_datatype = nm_mpi_datatype_get(datatype);
  if(p_datatype == NULL)
    {
      *count = MPI_UNDEFINED;
    }
  else
    {
      const size_t size = nm_mpi_datatype_size(p_datatype);
      if(size == 0)
	{
	  *count = 0;
	}
      else if(status->size % size != 0)
	{
	  *count = MPI_UNDEFINED;
	}
      else
	{
	  *count = status->size / size;
	}
    }
  return MPI_SUCCESS;
}

int mpi_get_elements_x(const MPI_Status*status, MPI_Datatype datatype, MPI_Count*count)
{
  nm_mpi_datatype_t*p_datatype = nm_mpi_datatype_get(datatype);
  if(p_datatype == NULL)
    return MPI_ERR_TYPE;
  const size_t datatype_size = nm_mpi_datatype_size(p_datatype);
  if(datatype_size == 0)
    {
      *count = 0;
    }
  else
    {
      *count = nm_mpi_datatype_get_elements(p_datatype, status->size);
    }
  return MPI_SUCCESS;
}

int mpi_get_elements(const MPI_Status*status, MPI_Datatype datatype, int*count)
{
  MPI_Count large_count;
  int err = mpi_get_elements_x(status, datatype, &large_count);
  if(err == MPI_SUCCESS)
    {
      if(large_count < INT_MAX || large_count == MPI_UNDEFINED)
	{
	  *count = large_count;
	  return err;
	}
      else
        {
          *count = MPI_UNDEFINED;
          return MPI_ERR_OTHER;
        }
    }
  else
    {
      *count = MPI_UNDEFINED;
      return err;
    }
}

int mpi_get_address(void *location, MPI_Aint *address)
{
  /* This is the "portable" way to generate an address.
     The difference of two pointers is the number of elements
     between them, so this gives the number of chars between location
     and ptr.  As long as sizeof(char) == 1, this will be the number
     of bytes from 0 to location */
  *address = (uintptr_t)location;
  return MPI_SUCCESS;
}

int mpi_address(void *location, MPI_Aint *address)
{
  return MPI_Get_address(location, address);
}

int mpi_alloc_mem(MPI_Aint size, MPI_Info info, void*_ptr)
{
  void**ptr = _ptr;
  *ptr = malloc(size);
  return MPI_SUCCESS;
}

int mpi_free_mem(void*ptr)
{
  free(ptr);
  return MPI_SUCCESS;
}

int mpi_info_create(MPI_Info*info)
{
  struct nm_mpi_info_s*p_info = nm_mpi_info_alloc();
  *info = p_info->id;
  return MPI_SUCCESS;
}

int mpi_info_free(MPI_Info*info)
{
  struct nm_mpi_info_s*p_info = nm_mpi_handle_info_get(&nm_mpi_infos, *info);
  if(p_info == NULL)
    return MPI_ERR_INFO;
  nm_mpi_info_free(p_info);
  *info = MPI_INFO_NULL;
  return MPI_SUCCESS;
}

int mpi_info_dup(MPI_Info info, MPI_Info*newinfo)
{
  struct nm_mpi_info_s*p_info = nm_mpi_handle_info_get(&nm_mpi_infos, info);
  if(p_info == NULL)
    return MPI_ERR_INFO;
  if(newinfo == NULL)
    return MPI_ERR_OTHER;
  struct nm_mpi_info_s*p_newinfo = nm_mpi_info_dup(p_info);
  *newinfo = p_newinfo->id;
  return MPI_SUCCESS;
}

int mpi_info_set(MPI_Info info, char*key, char*value)
{
  struct nm_mpi_info_s*p_info = nm_mpi_handle_info_get(&nm_mpi_infos, info);
  if(p_info == NULL)
    return MPI_ERR_INFO;
  nm_mpi_info_define(p_info, key, value);
  return MPI_SUCCESS;
}

int mpi_info_get(MPI_Info info, char*key, int valuelen, char*value, int*flag)
{
  struct nm_mpi_info_s*p_info = nm_mpi_handle_info_get(&nm_mpi_infos, info);
  if(p_info == NULL)
    return MPI_ERR_INFO;
  char*v = nm_mpi_info_hashtable_lookup(&p_info->content, key);
  if(v == NULL)
    {
      *flag = 0;
    }
  else
    {
      if(strlen(v) > valuelen)
	{
	  return MPI_ERR_OTHER;
	}
      strcpy(value, v);
      *flag = 1;
    }
  return MPI_SUCCESS;
}

int mpi_info_delete(MPI_Info info, char*key)
{
  struct nm_mpi_info_s*p_info = nm_mpi_handle_info_get(&nm_mpi_infos, info);
  if(p_info == NULL)
    return MPI_ERR_INFO;
  if(!nm_mpi_info_hashtable_probe(&p_info->content, key))
      return MPI_ERR_INFO_NOKEY;
  char*oldkey = NULL, *oldvalue = NULL;
  nm_mpi_info_hashtable_lookup2(&p_info->content, key, &oldkey, &oldvalue);
  nm_mpi_info_hashtable_remove(&p_info->content, key);
  free(oldkey);
  free(oldvalue);
  return MPI_SUCCESS;
}

int mpi_info_get_nkeys(MPI_Info info, int*nkeys)
{
  struct nm_mpi_info_s*p_info = nm_mpi_handle_info_get(&nm_mpi_infos, info);
  if(p_info == NULL)
    return MPI_ERR_INFO;
  *nkeys = nm_mpi_info_hashtable_size(&p_info->content);
  return MPI_SUCCESS;
}

int mpi_info_get_nthkey(MPI_Info info, int n, char*key)
{
  struct nm_mpi_info_s*p_info = nm_mpi_handle_info_get(&nm_mpi_infos, info);
  if(p_info == NULL)
    return MPI_ERR_INFO;
  nm_mpi_info_hashtable_enumerator_t e = nm_mpi_info_hashtable_enumerator_new(&p_info->content);
  char*ikey;
  int i;
  for(i = 0; i <= n; i++)
    {
      ikey = nm_mpi_info_hashtable_enumerator_next_key(e);
      if(ikey == NULL)
        break;
    }
  if(ikey != NULL)
    {
      strncpy(key, ikey, MPI_MAX_INFO_KEY);
    }
  else
    {
      return MPI_ERR_OTHER;
    }
  return MPI_SUCCESS;
}

int mpi_info_get_valuelen(MPI_Info info, char*key, int*valuelen, int*flag)
{
  struct nm_mpi_info_s*p_info = nm_mpi_handle_info_get(&nm_mpi_infos, info);
  if(p_info == NULL)
    return MPI_ERR_INFO;
  char*v = nm_mpi_info_hashtable_lookup(&p_info->content, key);
  if(v == NULL)
    {
      *flag = 0;
    }
  else
    {
      *valuelen = strlen(v);
      *flag = 1;
    }
  return MPI_SUCCESS;
}


int mpi_pcontrol(const int level, ...)
{
  return MPI_SUCCESS;
}

