
/*
 * NewMadeleine
 * Copyright (C) 2014-2020 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <nm_public.h>
#include <nm_private.h>
#include <nm_sendrecv_interface.h>
#include <nm_launcher_interface.h>
#include <Padico/Puk.h>
#include "nm_coll_private.h"
#include "nm_mpi_private.h"

/* ** reduce *********************************************** */

__PUK_SYM_INTERNAL
int nm_mpi_coll_reduce(nm_mpi_communicator_t*p_comm, int root,
                       int count, nm_mpi_datatype_t*p_datatype,
                       const void*sendbuf, void*recvbuf,
                       nm_tag_t tag, nm_mpi_operator_t*p_operator)
{
  const int size = nm_comm_size(p_comm->p_nm_comm);
  const int self = nm_comm_rank(p_comm->p_nm_comm);
  const int slot_size = count * p_datatype->true_extent;
  const nm_coll_tree_kind_t kind = nm_coll_tree_heuristic(nm_comm_size(p_comm->p_nm_comm), slot_size);
  struct nm_coll_tree_info_s tree;
  nm_coll_tree_init(&tree, kind, size, self, root);
  int*children = malloc(sizeof(int) * tree.max_arity);
  nm_mpi_request_t**requests = malloc(size * sizeof(nm_mpi_request_t*));
  void*gather_buf = malloc(tree.max_arity * slot_size);
  void*local_buf = NULL;
  const void*next_sbuf = (sendbuf == MPI_IN_PLACE) ? recvbuf : sendbuf; /* next data to send */
  /* input buffer from self */
  if(self == root)
    {
      assert(recvbuf != NULL);
      local_buf = recvbuf;
      if(sendbuf != MPI_IN_PLACE)
        {
          memcpy(recvbuf, sendbuf, slot_size);
        }
    }
  else
    {
      if(sendbuf == MPI_IN_PLACE)
        {
          local_buf = recvbuf;
        }
      else
        {
          local_buf = malloc(slot_size);
          memcpy(local_buf, sendbuf, slot_size);
        }
    }
  int s;
  for(s = tree.height - 1; s >= 0; s--)
    {
      int parent = -1;
      int n_children = 0;
      nm_coll_tree_step(&tree, s, &parent, children, &n_children);
      assert((parent == -1) || (n_children == 0) || (children[0] == -1));
      if(n_children > 0)
        {
          /* receive from children */
          int i;
          for(i = 0; i < n_children; i++)
            {
              assert(children[i] != -1);
              requests[i] = nm_mpi_coll_irecv(gather_buf + i * slot_size, count, p_datatype, children[i], tag, p_comm);
            }
          for(i = 0; i < n_children; i++)
            {
              nm_mpi_coll_wait(requests[i]);
            }
          /* Do the reduction operation */
          for(i = n_children - 1; i >= 0; i--)
            {
              void*inbuf = gather_buf + i * slot_size;
              void*inoutbuf = (i > 0) ? (gather_buf + (i - 1) * slot_size) : local_buf;
              (*p_operator->function)(inbuf, inoutbuf, &count, &p_datatype->id);
              next_sbuf = inoutbuf;
            }
        }
      /* send to parent */
      if(parent != -1)
        {
          nm_mpi_request_t*p_req = nm_mpi_coll_isend(next_sbuf, count, p_datatype, parent, tag, p_comm);
          nm_mpi_coll_wait(p_req);
        }
    }
  FREE_AND_SET_NULL(gather_buf);
  if(local_buf != recvbuf)
    {
      FREE_AND_SET_NULL(local_buf);
    }
  FREE_AND_SET_NULL(requests);
  FREE_AND_SET_NULL(children);
  return MPI_SUCCESS;

}

/* ** ireduce *********************************************** */

/** status of a reduce */
struct nm_mpi_coll_reduce_s
{
  struct nm_coll_tree_status_s coll_tree;
  const void*sendbuf; /**< send buffer given by user; may be NULL in case of MPI_IN_PLACE */
  void*recvbuf;       /**< recv buffer given by use; may be NULL if self != root */
  void*local_buf;     /**< local accumulator buffer; either malloced or a pointer to recvbuf if available */
  void*gather_buf;    /**< buffers used to receive data from toher nodes */
  void*next_sbuf;     /**< next buffer to send; either sendbuf or local_buf (or recvbuf if in-place) */
  nm_mpi_datatype_t*p_datatype;
  int count;
  int self;
  int root;
  nm_mpi_operator_t*p_operator;
  int step;           /**< current step in tree */
  int n_children;     /**< number of children in this step */
};

static void nm_mpi_coll_ireduce_step(struct nm_mpi_coll_reduce_s*p_reduce);
static void nm_mpi_coll_ireduce_send_notifier(nm_sr_event_t event, const nm_sr_event_info_t*event_info, void*_ref);
static void nm_mpi_coll_ireduce_recv_notifier(nm_sr_event_t event, const nm_sr_event_info_t*event_info, void*_ref);

static void nm_mpi_coll_ireduce_free(struct nm_mpi_coll_reduce_s*p_reduce)
{
  if(p_reduce->local_buf != p_reduce->recvbuf)
    free(p_reduce->local_buf);
  free(p_reduce->gather_buf);
  free(p_reduce);
}

__PUK_SYM_INTERNAL
void nm_mpi_coll_ireduce_wait(struct nm_mpi_coll_reduce_s*p_reduce)
{
  nm_coll_tree_status_wait(&p_reduce->coll_tree);
  nm_mpi_coll_ireduce_free(p_reduce);
}

__PUK_SYM_INTERNAL
int nm_mpi_coll_ireduce_test(struct nm_mpi_coll_reduce_s*p_reduce)
{
  int rc = nm_coll_tree_status_test(&p_reduce->coll_tree);
  if(rc)
    {
      nm_mpi_coll_ireduce_free(p_reduce);
    }
  return rc;
}


static void nm_mpi_coll_ireduce_step(struct nm_mpi_coll_reduce_s*p_reduce)
{
 restart:
  assert(p_reduce->coll_tree.pending_reqs == 0);
  if(p_reduce->step >= 0)
    {
      int parent = -1;
      nm_coll_tree_step(&p_reduce->coll_tree.tree, p_reduce->step, &parent,
                        p_reduce->coll_tree.children, &p_reduce->n_children);
      p_reduce->step--;
      assert(!((parent != -1) && (p_reduce->n_children > 0)));
      if(parent != -1)
        {
          /* send to parent */
          struct nm_data_s send_data;
          nm_mpi_data_build(&send_data, p_reduce->next_sbuf, p_reduce->p_datatype, p_reduce->count);
          nm_coll_tree_send(&p_reduce->coll_tree, parent, &send_data, &p_reduce->coll_tree.parent_request);
          nm_coll_tree_set_notifier(&p_reduce->coll_tree, &p_reduce->coll_tree.parent_request,
                                    &nm_mpi_coll_ireduce_send_notifier);
        }
      else if(p_reduce->n_children > 0)
        {
          /* receive from children */
          const size_t slot_size = p_reduce->count * p_reduce->p_datatype->true_extent;
          int i;
          for(i = 0; i < p_reduce->n_children; i++)
            {
              assert(p_reduce->coll_tree.children[i] != -1);
              struct nm_data_s recv_data;
              nm_mpi_data_build(&recv_data, (void*)p_reduce->gather_buf + i * slot_size, p_reduce->p_datatype, p_reduce->count);
              nm_coll_tree_recv(&p_reduce->coll_tree, p_reduce->coll_tree.children[i], &recv_data,
                                &p_reduce->coll_tree.children_requests[i]);
            }
          for(i = 0; i < p_reduce->n_children; i++)
            {
              nm_coll_tree_set_notifier(&p_reduce->coll_tree, &p_reduce->coll_tree.children_requests[i],
                                        &nm_mpi_coll_ireduce_recv_notifier);
            }
        }
      else
        {
          assert((parent == -1) && (p_reduce->n_children == 0));
          goto restart; /* no parent nor children; go to next step */
        }
    }
  else
    {
      /* notify completion */
      nm_coll_tree_status_signal(&p_reduce->coll_tree);
    }
}

static struct nm_mpi_coll_reduce_s*
nm_mpi_coll_ireduce_init(nm_mpi_communicator_t*p_comm, int root,
                         int count, nm_mpi_datatype_t*p_datatype,
                         const void*sendbuf, void*recvbuf,
                         nm_tag_t tag, nm_mpi_operator_t*p_operator,
                         nm_coll_tree_notifier_t p_notify, void*ref)
{
  nm_group_t p_group = nm_comm_group(p_comm->p_nm_comm);
  const int self = nm_comm_rank(p_comm->p_nm_comm);
  nm_session_t p_session = nm_comm_get_session(p_comm->p_nm_comm);
  const size_t slot_size = count * p_datatype->true_extent;

  struct nm_mpi_coll_reduce_s*p_reduce = malloc(sizeof(struct nm_mpi_coll_reduce_s));
  const nm_coll_tree_kind_t kind = nm_coll_tree_heuristic(nm_group_size(p_group), slot_size);
  nm_coll_tree_status_init(&p_reduce->coll_tree, p_session, kind, p_group, root, self, tag);
  nm_coll_tree_status_set_notifier(&p_reduce->coll_tree, p_notify, ref);
  p_reduce->step = p_reduce->coll_tree.tree.height - 1;
  p_reduce->p_operator = p_operator;
  p_reduce->sendbuf = sendbuf;
  p_reduce->recvbuf = recvbuf;
  p_reduce->gather_buf = malloc(p_reduce->coll_tree.tree.max_arity * slot_size);
  p_reduce->local_buf = NULL;
  p_reduce->next_sbuf = (sendbuf == MPI_IN_PLACE) ? recvbuf : (void*)sendbuf;
  p_reduce->p_datatype = p_datatype;
  p_reduce->count = count;
  p_reduce->self = self;
  p_reduce->root = root;
  p_reduce->n_children = 0;
  if(p_reduce->gather_buf == NULL)
    {
      NM_FATAL("madmpi: malloc failed in ireduce.\n");
    }
  /* input buffer from self */
  if(self == root)
    {
      assert(recvbuf != NULL);
      p_reduce->local_buf = recvbuf;
      if(sendbuf != MPI_IN_PLACE)
        {
          memcpy(recvbuf, sendbuf, slot_size);
        }
    }
  else
    {
      if(sendbuf == MPI_IN_PLACE)
        {
          p_reduce->local_buf = recvbuf;
        }
      else
        {
          p_reduce->local_buf = malloc(slot_size);
          if(p_reduce->local_buf == NULL)
            {
              NM_FATAL("madmpi: malloc failed in ireduce.\n");
            }
          memcpy(p_reduce->local_buf, sendbuf, slot_size);
        }
    }
  return p_reduce;
}

static void nm_mpi_coll_ireduce_start(struct nm_mpi_coll_reduce_s*p_reduce)
{
  /* start tree algorithm */
  nm_mpi_coll_ireduce_step(p_reduce);
}

__PUK_SYM_INTERNAL
struct nm_mpi_coll_reduce_s*nm_mpi_coll_ireduce(nm_mpi_communicator_t*p_comm, int root,
                                                int count, nm_mpi_datatype_t*p_datatype,
                                                const void*sendbuf, void*recvbuf,
                                                nm_tag_t tag, nm_mpi_operator_t*p_operator)
{
struct nm_mpi_coll_reduce_s*p_reduce = nm_mpi_coll_ireduce_init(p_comm, root, count, p_datatype, sendbuf, recvbuf, tag, p_operator, NULL, NULL);
 nm_mpi_coll_ireduce_start(p_reduce);
 return p_reduce;
}

static void nm_mpi_coll_ireduce_recv_notifier(nm_sr_event_t event, const nm_sr_event_info_t*p_event_info, void* _ref)
{
  struct nm_coll_tree_status_s*p_coll_tree = _ref;
  struct nm_mpi_coll_reduce_s*p_reduce = nm_container_of(p_coll_tree, struct nm_mpi_coll_reduce_s, coll_tree);
  nm_atomic_dec(&p_reduce->coll_tree.pending_reqs);
  if(p_reduce->coll_tree.pending_reqs == 0)
    {
      const size_t slot_size = p_reduce->count * nm_mpi_datatype_size(p_reduce->p_datatype);
      int i;
      for(i = p_reduce->n_children - 1; i >= 0; i--)
        {
          void*inbuf = p_reduce->gather_buf + i * slot_size;
          void*inoutbuf = (i > 0) ? (p_reduce->gather_buf + (i - 1) * slot_size) : p_reduce->local_buf;
          (*p_reduce->p_operator->function)(inbuf, inoutbuf, &p_reduce->count, &p_reduce->p_datatype->id);
          p_reduce->next_sbuf = inoutbuf;
        }
      nm_mpi_coll_ireduce_step(p_reduce);
    }
}

static void nm_mpi_coll_ireduce_send_notifier(nm_sr_event_t event, const nm_sr_event_info_t*p_event_info, void* _ref)
{
  struct nm_coll_tree_status_s*p_coll_tree = _ref;
  struct nm_mpi_coll_reduce_s*p_reduce = nm_container_of(p_coll_tree, struct nm_mpi_coll_reduce_s, coll_tree);
  const int p = nm_atomic_dec(&p_reduce->coll_tree.pending_reqs);
  if(p == 0)
    {
      nm_mpi_coll_ireduce_step(p_reduce);
    }
}


/* ** iallreduce ******************************************* */

struct nm_mpi_coll_allreduce_s
{
  struct nm_mpi_coll_reduce_s*p_reduce;
  struct nm_coll_bcast_s*p_bcast;
  nm_mpi_communicator_t*p_comm;
  enum
    {
     NM_MPI_COLL_IALLREDUCE_REDUCING,
     NM_MPI_COLL_IALLREDUCE_BROADCASTING,
     NM_MPI_COLL_IALLREDUCE_COMPLETED
    } state;
  nm_cond_status_t status;
};

static void nm_mpi_coll_iallreduce_bcast_notifier(void*ref)
{
  struct nm_mpi_coll_allreduce_s*p_allreduce = ref;
  nm_cond_signal(&p_allreduce->status, 1);
}

static void nm_mpi_coll_iallreduce_reduce_notifier(void*ref)
{
  struct nm_mpi_coll_allreduce_s*p_allreduce = ref;
  const int root = 0;
  const nm_tag_t tag =  NM_MPI_TAG_PRIVATE_ALLREDUCE;
  assert(p_allreduce->state == NM_MPI_COLL_IALLREDUCE_REDUCING);
  p_allreduce->state = NM_MPI_COLL_IALLREDUCE_BROADCASTING;
  struct nm_data_s data;
  nm_mpi_data_build(&data, p_allreduce->p_reduce->recvbuf, p_allreduce->p_reduce->p_datatype, p_allreduce->p_reduce->count);
  nm_comm_t p_comm = p_allreduce->p_comm->p_nm_comm;
  nm_group_t p_group = nm_comm_group(p_comm);
  const int self = nm_comm_rank(p_comm);
  nm_session_t p_session = nm_comm_get_session(p_comm);
  p_allreduce->p_bcast = nm_coll_group_data_ibcast_init(p_session, p_group, root, self, &data, tag,
                                                        &nm_mpi_coll_iallreduce_bcast_notifier, p_allreduce);
  nm_coll_ibcast_start(p_allreduce->p_bcast);
}

__PUK_SYM_INTERNAL
struct nm_mpi_coll_allreduce_s*nm_mpi_coll_iallreduce(nm_mpi_communicator_t*p_comm,
                                                      int count, nm_mpi_datatype_t*p_datatype,
                                                      const void*sendbuf, void*recvbuf,
                                                      nm_tag_t tag, nm_mpi_operator_t*p_operator)
{
  const int root = 0;
  struct nm_mpi_coll_allreduce_s*p_allreduce = malloc(sizeof(struct nm_mpi_coll_allreduce_s));
  nm_cond_init(&p_allreduce->status, 0);
  p_allreduce->state = NM_MPI_COLL_IALLREDUCE_REDUCING;
  p_allreduce->p_comm = p_comm;
  p_allreduce->p_reduce = nm_mpi_coll_ireduce_init(p_comm, root, count, p_datatype, sendbuf, recvbuf,
                                                   tag, p_operator, &nm_mpi_coll_iallreduce_reduce_notifier, p_allreduce);
  p_allreduce->p_bcast = NULL;
  nm_mpi_coll_ireduce_start(p_allreduce->p_reduce);
  return p_allreduce;
}

static void nm_mpi_coll_iallreduce_free(struct nm_mpi_coll_allreduce_s*p_allreduce)
{
  nm_mpi_coll_ireduce_free(p_allreduce->p_reduce);
  nm_coll_ibcast_free(p_allreduce->p_bcast);
  nm_cond_destroy(&p_allreduce->status);
  free(p_allreduce);
}

__PUK_SYM_INTERNAL
void nm_mpi_coll_iallreduce_wait(struct nm_mpi_coll_allreduce_s*p_allreduce)
{
  nm_cond_wait(&p_allreduce->status, 1, nm_comm_get_session(p_allreduce->p_comm->p_nm_comm)->p_core);
  nm_mpi_coll_iallreduce_free(p_allreduce);
}

__PUK_SYM_INTERNAL
int nm_mpi_coll_iallreduce_test(struct nm_mpi_coll_allreduce_s*p_allreduce)
{
  int rc = nm_cond_test(&p_allreduce->status, 1);
  if(rc)
    {
      nm_mpi_coll_iallreduce_free(p_allreduce);
      return MPI_SUCCESS;
    }
  else
    {
      return -NM_EAGAIN;
    }
}


/* ** ialltoall ******************************************** */

/** status of an alltoall operation (blocking/non-blocking) */
struct nm_mpi_coll_alltoall_s
{
  nm_cond_status_t status;       /**< cond status for the whole operation; set to 1 when finished */
  nm_mpi_communicator_t*p_comm;  /**< communicator for the alltoall */
  int seq;                       /**< next request to send */
  int pending;                   /**< number of pendings requests (send+recv) */
  nm_sr_request_t*send_requests;
  nm_sr_request_t*recv_requests;
  const void*sendbuf;
  int sendcount;
  nm_mpi_datatype_t*p_send_datatype;
};

static void nm_mpi_coll_ialltoall_step(struct nm_mpi_coll_alltoall_s*p_alltoall);
static void nm_mpi_coll_ialltoall_send_notifier(nm_sr_event_t event, const nm_sr_event_info_t*p_info, void*_ref);
static void nm_mpi_coll_ialltoall_recv_notifier(nm_sr_event_t event, const nm_sr_event_info_t*p_info, void*_ref);

__PUK_SYM_INTERNAL
struct nm_mpi_coll_alltoall_s*nm_mpi_coll_ialltoall(nm_mpi_communicator_t*p_comm,
                                                    const void*sendbuf, int sendcount, nm_mpi_datatype_t*p_send_datatype,
                                                    void*recvbuf, int recvcount, nm_mpi_datatype_t*p_recv_datatype)
{
  const int nports = 4; /* send at most to 4 nodes at the same time. TODO- tune value */
  const nm_tag_t tag = NM_MPI_TAG_PRIVATE_ALLTOALL;
  nm_session_t p_session = nm_mpi_communicator_get_session(p_comm);
  const int self = nm_comm_rank(p_comm->p_nm_comm);
  struct nm_mpi_coll_alltoall_s*p_alltoall = malloc(sizeof(struct nm_mpi_coll_alltoall_s));
  nm_cond_init(&p_alltoall->status, 0);
  p_alltoall->p_comm = p_comm;
  p_alltoall->seq = 0;
  p_alltoall->pending = 2 * (nm_comm_size(p_comm->p_nm_comm) - 1);
  p_alltoall->send_requests = malloc(nm_comm_size(p_comm->p_nm_comm) * sizeof(nm_sr_request_t));
  p_alltoall->recv_requests = malloc(nm_comm_size(p_comm->p_nm_comm) * sizeof(nm_sr_request_t));
  p_alltoall->sendbuf = sendbuf;
  p_alltoall->sendcount = sendcount;
  p_alltoall->p_send_datatype = p_send_datatype;
  /* pre-post all receive requests */
  int i;
  for(i = 0; i < nm_comm_size(p_comm->p_nm_comm); i++)
    {
      if(i != self)
	{
          nm_gate_t p_gate = nm_comm_get_gate(p_comm->p_nm_comm, i);
          struct nm_data_s data;
          nm_mpi_data_build(&data, (void*)nm_mpi_datatype_get_ptr(recvbuf, i * recvcount, p_recv_datatype),
                            p_recv_datatype, recvcount);
          nm_sr_irecv_data(p_session, p_gate, tag, &data, &p_alltoall->recv_requests[i]);
          nm_sr_request_set_ref(&p_alltoall->recv_requests[i], p_alltoall);
          nm_sr_request_monitor(p_session, &p_alltoall->recv_requests[i], NM_SR_EVENT_FINALIZED,
                                &nm_mpi_coll_ialltoall_recv_notifier);
	}
    }
  /* pre-load nports (=4) send requests */
  for(i = 0; i < nports; i++)
    {
      nm_mpi_coll_ialltoall_step(p_alltoall);
    }
  nm_mpi_datatype_copy(nm_mpi_datatype_get_ptr((void*)sendbuf, self * sendcount, p_send_datatype), p_send_datatype, sendcount,
                       nm_mpi_datatype_get_ptr(recvbuf, self * recvcount, p_recv_datatype), p_recv_datatype, recvcount);
  return p_alltoall;
}

static inline void nm_mpi_coll_ialltoall_free(struct nm_mpi_coll_alltoall_s*p_alltoall)
{
  free(p_alltoall->send_requests);
  free(p_alltoall->recv_requests);
  nm_cond_destroy(&p_alltoall->status);
  free(p_alltoall);
}

__PUK_SYM_INTERNAL
int nm_mpi_coll_ialltoall_test(struct nm_mpi_coll_alltoall_s*p_alltoall)
{
  int rc = nm_cond_test(&p_alltoall->status, 1);
  if(rc)
    {
      nm_mpi_coll_ialltoall_free(p_alltoall);
      return MPI_SUCCESS;
    }
  else
    {
      return -NM_EAGAIN;
    }
}

__PUK_SYM_INTERNAL
void nm_mpi_coll_ialltoall_wait(struct nm_mpi_coll_alltoall_s*p_alltoall)
{
  nm_cond_wait(&p_alltoall->status, 1, nm_comm_get_session(p_alltoall->p_comm->p_nm_comm)->p_core);
  nm_mpi_coll_ialltoall_free(p_alltoall);
}

/** try to start a new send request on the current alltoall */
static void nm_mpi_coll_ialltoall_step(struct nm_mpi_coll_alltoall_s*p_alltoall)
{
  const nm_tag_t tag = NM_MPI_TAG_PRIVATE_ALLTOALL;
  nm_session_t p_session = nm_mpi_communicator_get_session(p_alltoall->p_comm);
  int s = nm_atomic_inc(&p_alltoall->seq);
  if(s == 0)
    {
      /* skip self */
      s = nm_atomic_inc(&p_alltoall->seq);
    }
  if(s < nm_comm_size(p_alltoall->p_comm->p_nm_comm))
    {
      /* send to rank = seq + self */
      const int dest = (s + nm_comm_rank(p_alltoall->p_comm->p_nm_comm)) % nm_comm_size(p_alltoall->p_comm->p_nm_comm);
      nm_gate_t p_gate = nm_comm_get_gate(p_alltoall->p_comm->p_nm_comm, dest);
      struct nm_data_s data;
      nm_mpi_data_build(&data, (void*)nm_mpi_datatype_get_ptr((void*)p_alltoall->sendbuf, dest * p_alltoall->sendcount, p_alltoall->p_send_datatype),
                        p_alltoall->p_send_datatype, p_alltoall->sendcount);

      nm_sr_isend_data(p_session, p_gate, tag, &data, &p_alltoall->send_requests[dest]);
      nm_sr_request_set_ref(&p_alltoall->send_requests[dest], p_alltoall);
      nm_sr_request_monitor(p_session, &p_alltoall->send_requests[dest], NM_SR_EVENT_FINALIZED,
                            &nm_mpi_coll_ialltoall_send_notifier);
    }
}

/** handler for complete send requests in alltoall */
static void nm_mpi_coll_ialltoall_send_notifier(nm_sr_event_t event, const nm_sr_event_info_t*p_info, void*_ref)
{
  struct nm_mpi_coll_alltoall_s*p_alltoall = _ref;
  const int p = nm_atomic_dec(&p_alltoall->pending);
  if(p == 0)
    {
      nm_cond_signal(&p_alltoall->status, 1);
    }
  else
    {
      nm_mpi_coll_ialltoall_step(p_alltoall);
    }
}

/** handler for complete receive requests in alltoall */
static void nm_mpi_coll_ialltoall_recv_notifier(nm_sr_event_t event, const nm_sr_event_info_t*p_info, void*_ref)
{
  struct nm_mpi_coll_alltoall_s*p_alltoall = _ref;
  const int p = nm_atomic_dec(&p_alltoall->pending);
  if(p == 0)
    {
      nm_cond_signal(&p_alltoall->status, 1);
    }
}


/* ** iallgather ******************************************* */

struct nm_mpi_coll_allgather_s
{
  struct nm_coll_gather_s*p_gather;
  struct nm_coll_bcast_s*p_bcast;
  nm_mpi_communicator_t*p_comm;
  enum
    {
     NM_MPI_COLL_IALLGATHER_GATHERING,
     NM_MPI_COLL_IALLGATHER_BROADCASTING,
     NM_MPI_COLL_IALLGATHER_COMPLETED
    } state;
  nm_cond_status_t status;
  struct nm_data_s bcast_data;
};

static void nm_mpi_coll_iallgather_free(struct nm_mpi_coll_allgather_s*p_allgather)
{
  nm_coll_igather_free(p_allgather->p_gather);
  nm_coll_ibcast_free(p_allgather->p_bcast);
  nm_cond_destroy(&p_allgather->status);
  free(p_allgather);
}

__PUK_SYM_INTERNAL
int nm_mpi_coll_iallgather_test(struct nm_mpi_coll_allgather_s*p_allgather)
{
  int rc = nm_cond_test(&p_allgather->status, 1);
  if(rc)
    {
      nm_mpi_coll_iallgather_free(p_allgather);
      return MPI_SUCCESS;
    }
  else
    {
      return -NM_EAGAIN;
    }
}

__PUK_SYM_INTERNAL
void nm_mpi_coll_iallgather_wait(struct nm_mpi_coll_allgather_s*p_allgather)
{
  nm_cond_wait(&p_allgather->status, 1, nm_comm_get_session(p_allgather->p_comm->p_nm_comm)->p_core);
  nm_mpi_coll_iallgather_free(p_allgather);
}

static void nm_mpi_coll_iallgather_bcast_notifier(void*ref)
{
  struct nm_mpi_coll_allgather_s*p_allgather = ref;
  nm_cond_signal(&p_allgather->status, 1);
}

static void nm_mpi_coll_iallgather_gather_notifier(void*ref)
{
  struct nm_mpi_coll_allgather_s*p_allgather = ref;
  assert(p_allgather->state == NM_MPI_COLL_IALLGATHER_GATHERING);
  p_allgather->state = NM_MPI_COLL_IALLGATHER_BROADCASTING;

  const int root = 0;
  const nm_tag_t tag = NM_MPI_TAG_PRIVATE_ALLGATHER;
  nm_comm_t p_comm = p_allgather->p_comm->p_nm_comm;
  nm_group_t p_group = nm_comm_group(p_comm);
  const int self = nm_comm_rank(p_comm);
  nm_session_t p_session = nm_comm_get_session(p_comm);
  p_allgather->p_bcast =
    nm_coll_group_data_ibcast_init(p_session, p_group, root, self, &p_allgather->bcast_data, tag,
                                   &nm_mpi_coll_iallgather_bcast_notifier, p_allgather);
  nm_coll_ibcast_start(p_allgather->p_bcast);
}

__PUK_SYM_INTERNAL
struct nm_mpi_coll_allgather_s*nm_mpi_coll_iallgather(nm_mpi_communicator_t*p_comm,
                                                      const void*sendbuf, int sendcount, nm_mpi_datatype_t*p_send_datatype,
                                                      void*recvbuf, int recvcount, nm_mpi_datatype_t*p_recv_datatype)
{
  const int root = 0;
  const int self = nm_comm_rank(p_comm->p_nm_comm);
  const nm_tag_t tag = NM_MPI_TAG_PRIVATE_ALLGATHER;
  struct nm_mpi_coll_allgather_s*p_allgather = malloc(sizeof(struct nm_mpi_coll_allgather_s));
  nm_cond_init(&p_allgather->status, 0);
  p_allgather->state = NM_MPI_COLL_IALLGATHER_GATHERING;
  p_allgather->p_comm = p_comm;
  p_allgather->p_bcast = NULL;

  /* build send & recv data descriptor for igather from iallgather info.
   * There is no direct matching because of MPI_IN_PLACE.
   */
  struct nm_data_s send_data;
  struct nm_data_s*p_recv_data = NULL;
  if(sendbuf == MPI_IN_PLACE)
    {
      if(root == self)
        {
          nm_data_null_build(&send_data);
        }
      else
        {
          nm_mpi_data_build(&send_data,
                            nm_mpi_datatype_get_ptr(recvbuf, (self * recvcount), p_recv_datatype),
                            p_recv_datatype, recvcount);
        }
    }
  else
    {
      nm_mpi_data_build(&send_data, (void*)sendbuf, p_send_datatype, sendcount);
    }
  if(root == self)
    {
      const int size = nm_comm_size(p_comm->p_nm_comm);
      p_recv_data = malloc(size * sizeof(struct nm_data_s));
      int i;
      for(i = 0; i < size; i++)
        {
          nm_mpi_data_build(&p_recv_data[i],
                            nm_mpi_datatype_get_ptr(recvbuf, (i * recvcount), p_recv_datatype),
                            p_recv_datatype, recvcount);
        }
    }
  else
    {
      p_recv_data = NULL;
    }

  nm_mpi_data_build(&p_allgather->bcast_data, recvbuf, p_recv_datatype, recvcount * nm_comm_size(p_comm->p_nm_comm));

  p_allgather->p_gather = nm_coll_igather_data_init(nm_comm_get_session(p_comm->p_nm_comm),
                                                    nm_comm_group(p_comm->p_nm_comm),
                                                    root, self, &send_data, p_recv_data, tag,
                                                    &nm_mpi_coll_iallgather_gather_notifier, p_allgather);
  nm_coll_igather_start(p_allgather->p_gather);
  return p_allgather;
}
