/*
 * NewMadeleine
 * Copyright (C) 2006-2020 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#include "nm_mpi_private.h"

#include <Padico/Puk.h>
#include <Padico/Module.h>
PADICO_MODULE_HOOK(MadMPI);

NM_MPI_HANDLE_TYPE(errhandler, struct nm_mpi_errhandler_s, _NM_MPI_ERRHANDLER_OFFSET, 8);
static struct nm_mpi_handle_errhandler_s nm_mpi_errhandlers;

/* ********************************************************* */
/* Aliases */

NM_MPI_ALIAS(MPI_Error_string,            mpi_error_string);
NM_MPI_ALIAS(MPI_Error_class,             mpi_error_class);
NM_MPI_ALIAS(MPI_Errhandler_create,       mpi_errhandler_create);
NM_MPI_ALIAS(MPI_Errhandler_set,          mpi_errhandler_set);
NM_MPI_ALIAS(MPI_Errhandler_get,          mpi_errhandler_get);
NM_MPI_ALIAS(MPI_Errhandler_free,         mpi_errhandler_free);
NM_MPI_ALIAS(MPI_Comm_create_errhandler,  mpi_comm_create_errhandler);
NM_MPI_ALIAS(MPI_Comm_set_errhandler,     mpi_comm_set_errhandler);
NM_MPI_ALIAS(MPI_Comm_get_errhandler,     mpi_comm_get_errhandler);
NM_MPI_ALIAS(MPI_Comm_call_errhandler,    mpi_comm_call_errhandler);

/* ********************************************************* */

__PUK_SYM_INTERNAL
void nm_mpi_err_init(void)
{
  nm_mpi_handle_errhandler_init(&nm_mpi_errhandlers);
  struct nm_mpi_errhandler_s*p_errhandler_return =
    nm_mpi_handle_errhandler_store(&nm_mpi_errhandlers, MPI_ERRORS_RETURN);
  p_errhandler_return->kind = NM_MPI_ERRHANDLER_NONE;
  struct nm_mpi_errhandler_s*p_errhandler_fatal =
    nm_mpi_handle_errhandler_store(&nm_mpi_errhandlers, MPI_ERRORS_ARE_FATAL);
  p_errhandler_fatal->kind = NM_MPI_ERRHANDLER_NONE;
}

__PUK_SYM_INTERNAL
void nm_mpi_err_exit(void)
{
  nm_mpi_handle_errhandler_finalize(&nm_mpi_errhandlers, NULL);
}

__PUK_SYM_INTERNAL
struct nm_mpi_errhandler_s*nm_mpi_errhandler_get(int errhandler, enum nm_mpi_errhandler_kind_e kind)
{
  struct nm_mpi_errhandler_s*p_errhandler = nm_mpi_handle_errhandler_get(&nm_mpi_errhandlers, errhandler);
  if((p_errhandler->kind != kind) &&
     (kind != NM_MPI_ERRHANDLER_NONE) &&
     (p_errhandler->id >= _NM_MPI_ERRHANDLER_OFFSET))
    {
      NM_MPI_WARNING("wrong type for errhandler = %d; expected = %d; found = %d", errhandler, kind, p_errhandler->kind);
      return NULL;
    }
  return p_errhandler;
}

__PUK_SYM_INTERNAL
struct nm_mpi_errhandler_s*nm_mpi_errhandler_alloc(enum nm_mpi_errhandler_kind_e kind, MPI_Handler_function*func)
{
  struct nm_mpi_errhandler_s*p_errhandler = nm_mpi_handle_errhandler_alloc(&nm_mpi_errhandlers);
  p_errhandler->kind = kind;
  p_errhandler->function.legacy = func;
  return p_errhandler;
}

__PUK_SYM_INTERNAL
void nm_mpi_errhandler_exec(struct nm_mpi_errhandler_s*p_errhandler, enum nm_mpi_errhandler_kind_e kind, int handle, int err)
{
  assert((p_errhandler->kind == kind) || (p_errhandler->id < _NM_MPI_ERRHANDLER_OFFSET));
  const int id = p_errhandler->id;
  if(id == MPI_ERRHANDLER_NULL)
    {
      NM_MPI_FATAL_ERROR("trying to invocate MPI_ERRHANDLER_NULL.\n");
    }
  switch(kind)
    {
    case NM_MPI_ERRHANDLER_COMM:
      if(id == MPI_ERRORS_ARE_FATAL)
        {
          NM_MPI_FATAL_ERROR("error %d on communicator %d\n", err, handle);
        }
      else if(id == MPI_ERRORS_RETURN)
        {
          NM_MPI_WARNING("error %d on communicator %d", err, handle);
        }
      else
        {
          (*p_errhandler->function.comm_func)(&handle, &err);
        }
      break;
    case NM_MPI_ERRHANDLER_WIN:
      if(id == MPI_ERRORS_ARE_FATAL)
        {
          NM_MPI_FATAL_ERROR("error %d on window %d\n", err, handle);
        }
      else if(id == MPI_ERRORS_RETURN)
        {
          NM_MPI_WARNING("error %d on window %d", err, handle);
        }
      else
        {
          (*p_errhandler->function.win_func)(&handle, &err);
        }
      break;
    case NM_MPI_ERRHANDLER_FILE:
      if(id == MPI_ERRORS_ARE_FATAL)
        {
          NM_MPI_FATAL_ERROR("error %d on file %d\n", err, handle);
        }
      else if(id == MPI_ERRORS_RETURN)
        {
          NM_MPI_WARNING("error %d on file %d", err, handle);
        }
      else
        {
          (*p_errhandler->function.file_func)(&handle, &err);
        }
      break;
    default:
      NM_MPI_FATAL_ERROR("undefined errhandler kind %d\n", p_errhandler->kind);
      break;
    }
}


/* ********************************************************* */
/* ** MPI API */

int mpi_error_class(int errorcode, int*errorclass)
{
  if(errorcode < MPI_ERR_LASTCODE)
    *errorclass = errorcode;
  else
    *errorclass = MPI_ERR_UNKNOWN;
  return MPI_SUCCESS;
}

int mpi_error_string(int errorcode, char *string, int *resultlen)
{
  *resultlen = 100;
  string = malloc(*resultlen * sizeof(char));
  switch (errorcode) {
    case MPI_SUCCESS : {
      snprintf(string, *resultlen, "Error MPI_SUCCESS\n");
      break;
    }
    case MPI_ERR_BUFFER : {
      snprintf(string, *resultlen, "Error MPI_ERR_BUFFER\n");
      break;
    }
    case MPI_ERR_COUNT : {
      snprintf(string, *resultlen, "Error MPI_ERR_COUNT\n");
      break;
    }
    case MPI_ERR_TYPE : {
      snprintf(string, *resultlen, "Error MPI_ERR_TYPE\n");
      break;
    }
    case MPI_ERR_TAG : {
      snprintf(string, *resultlen, "Error MPI_ERR_TAG\n");
      break;
    }
    case MPI_ERR_COMM : {
      snprintf(string, *resultlen, "Error MPI_ERR_COMM\n");
      break;
    }
    case MPI_ERR_RANK : {
      snprintf(string, *resultlen, "Error MPI_ERR_RANK\n");
      break;
    }
    case MPI_ERR_ROOT : {
      snprintf(string, *resultlen, "Error MPI_ERR_ROOT\n");
      break;
    }
    case MPI_ERR_GROUP : {
      snprintf(string, *resultlen, "Error MPI_ERR_GROUP\n");
      break;
    }
    case MPI_ERR_OP : {
      snprintf(string, *resultlen, "Error MPI_ERR_OP\n");
      break;
    }
    case MPI_ERR_TOPOLOGY : {
      snprintf(string, *resultlen, "Error MPI_ERR_TOPOLOGY\n");
      break;
    }
    case MPI_ERR_DIMS : {
      snprintf(string, *resultlen, "Error MPI_ERR_DIMS\n");
      break;
    }
    case MPI_ERR_ARG : {
      snprintf(string, *resultlen, "Error MPI_ERR_ARG\n");
      break;
    }
    case MPI_ERR_UNKNOWN : {
      snprintf(string, *resultlen, "Error MPI_ERR_UNKNOWN\n");
      break;
    }
    case MPI_ERR_TRUNCATE : {
      snprintf(string, *resultlen, "Error MPI_ERR_TRUNCATE\n");
      break;
    }
    case MPI_ERR_OTHER : {
      snprintf(string, *resultlen, "Error MPI_ERR_OTHER\n");
      break;
    }
    case MPI_ERR_INTERN : {
      snprintf(string, *resultlen, "Error MPI_ERR_INTERN\n");
      break;
    }
    case MPI_ERR_IN_STATUS : {
      snprintf(string, *resultlen, "Error MPI_ERR_IN_STATUS\n");
      break;
    }
    case MPI_ERR_PENDING : {
      snprintf(string, *resultlen, "Error MPI_ERR_PENDING\n");
      break;
    }
    case MPI_ERR_REQUEST : {
      snprintf(string, *resultlen, "Error MPI_ERR_REQUEST\n");
      break;
    }
    case MPI_ERR_LASTCODE : {
      snprintf(string, *resultlen, "Error MPI_ERR_LASTCODE\n");
      break;
    }
    default : {
      snprintf(string, *resultlen, "Error %d unknown\n", errorcode);
    }
  }
  return MPI_SUCCESS;
}


int mpi_comm_create_errhandler(MPI_Comm_errhandler_function*function, MPI_Errhandler*errhandler)
{
  struct nm_mpi_errhandler_s*p_errhandler = nm_mpi_errhandler_alloc(NM_MPI_ERRHANDLER_COMM, function);
  *errhandler = p_errhandler->id;
  return MPI_SUCCESS;
}

int mpi_comm_set_errhandler(MPI_Comm comm, MPI_Errhandler errhandler)
{
  struct nm_mpi_errhandler_s*p_errhandler = nm_mpi_errhandler_get(errhandler, NM_MPI_ERRHANDLER_COMM);
  if(p_errhandler == NULL)
    return MPI_ERR_ARG;
  nm_mpi_communicator_t*p_comm = nm_mpi_communicator_get(comm);
  if(p_comm == NULL)
    return MPI_ERR_COMM;
  p_comm->p_errhandler = p_errhandler;
  return MPI_SUCCESS;
}

int mpi_comm_get_errhandler(MPI_Comm comm, MPI_Errhandler*errhandler)
{
  nm_mpi_communicator_t*p_comm = nm_mpi_communicator_get(comm);
  if(p_comm == NULL)
    return MPI_ERR_COMM;
  *errhandler = p_comm->p_errhandler->id;
  return MPI_SUCCESS;
}

int mpi_comm_call_errhandler(MPI_Comm comm, int errorcode)
{
  nm_mpi_communicator_t*p_comm = nm_mpi_communicator_get(comm);
  if(p_comm == NULL)
    return MPI_ERR_COMM;
  nm_mpi_errhandler_exec(p_comm->p_errhandler, NM_MPI_ERRHANDLER_COMM, comm, errorcode);
  return MPI_SUCCESS;
}

/* legacy MPI-1 functions */
int mpi_errhandler_create(MPI_Handler_function*function, MPI_Errhandler*errhandler)
{
  return mpi_comm_create_errhandler(function, errhandler);
}

int mpi_errhandler_set(MPI_Comm comm, MPI_Errhandler errhandler)
{
  return mpi_comm_set_errhandler(comm, errhandler);
}

int mpi_errhandler_get(MPI_Comm comm, MPI_Errhandler*errhandler)
{
  return mpi_comm_get_errhandler(comm, errhandler);
}

int mpi_errhandler_free(MPI_Errhandler*errhandler)
{
  struct nm_mpi_errhandler_s*p_errhandler = nm_mpi_errhandler_get(*errhandler, NM_MPI_ERRHANDLER_NONE);
  if(p_errhandler == NULL)
    return MPI_ERR_ARG;
  nm_mpi_handle_errhandler_free(&nm_mpi_errhandlers, p_errhandler);
  *errhandler = MPI_ERRHANDLER_NULL;
  return MPI_SUCCESS;
}
