/*
 * NewMadeleine
 * Copyright (C) 2006-2019 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <nm_onesided_interface.h>
#include <nm_sendrecv_interface.h>
#include <nm_core_interface.h>
#ifdef PIOMAN
#include <pioman.h>
#endif /* PIOMAN */
#include <nm_log.h>
#include <stdlib.h>

/** tag used to send op requests */
#define NM_ONESIDED_TAG 0x01
/** tag used for replies, with explict recv */
#define NM_ONESIDED_TAG_REPLY 0x02

/* ********************************************************* */

struct nm_onesided_data_content_s
{
  struct nm_onesided_header_s*p_header;
  const struct nm_data_s*p_data;
};

static void nm_onesided_data_traversal(const void*_content, nm_data_apply_t apply, void*_context);
const struct nm_data_ops_s nm_onesided_data_ops =
  {
    .p_traversal = &nm_onesided_data_traversal
  };
NM_DATA_TYPE(onesided, struct nm_onesided_data_content_s, &nm_onesided_data_ops);

static void nm_onesided_data_traversal(const void*_content, nm_data_apply_t apply, void*_context)
{
  const struct nm_onesided_data_content_s*p_content = _content;
  (*apply)(p_content->p_header, sizeof(struct nm_onesided_header_s), _context);
  nm_data_traversal_apply(p_content->p_data, apply, _context);
}

static inline void nm_onesided_data_build(struct nm_data_s*p_onesided_data,
                                          struct nm_onesided_header_s*p_header, const struct nm_data_s*p_data)
{
  nm_data_onesided_set(p_onesided_data, (struct nm_onesided_data_content_s)
                       {
                         .p_header = p_header,
                         .p_data = p_data
                       });
}

/* ********************************************************* */

static void nm_onesided_handler(nm_sr_event_t event, const nm_sr_event_info_t*p_info, void*_ref);

static void nm_onesided_req_reload(nm_onesided_t p_onesided)
{
  if(p_onesided->running)
    {
      nm_sr_recv_init(p_onesided->p_session, &p_onesided->request.sr_req);
      nm_sr_request_set_ref(&p_onesided->request.sr_req, p_onesided);
      nm_sr_request_monitor(p_onesided->p_session, &p_onesided->request.sr_req,
                            NM_SR_EVENT_FINALIZED | NM_SR_EVENT_RECV_DATA | NM_SR_EVENT_RECV_CANCELLED,
                            &nm_onesided_handler);
      nm_sr_recv_irecv(p_onesided->p_session, &p_onesided->request.sr_req, NM_ANY_GATE, NM_ONESIDED_TAG, NM_TAG_MASK_FULL);
    }
  else
    {
      /* acknowledge termination */
      nm_cond_signal(&p_onesided->done, 1);
    }
}

static void nm_onesided_ack_handler(nm_sr_event_t event, const nm_sr_event_info_t*p_info, void*_ref)
{
  nm_sr_request_t*p_req = p_info->req.p_request;
  if(event & NM_SR_EVENT_FINALIZED)
    {
      struct nm_onesided_s*p_onesided = _ref;
      nm_sr_req_free(p_onesided->req_allocator, p_req);
    }
}

static void nm_onesided_handler(nm_sr_event_t event, const nm_sr_event_info_t*p_info, void*_ref)
{
  assert(! ((event & NM_SR_EVENT_FINALIZED) &&
            (event & NM_SR_EVENT_RECV_DATA)));
  nm_sr_request_t*pp_req = p_info->req.p_request;
  struct nm_onesided_s*p_onesided = NULL;
  nm_sr_request_get_ref(pp_req, (void**)&p_onesided);
  assert(p_onesided != NULL);
  if(!(event & NM_SR_EVENT_RECV_CANCELLED))
    {
      if(event & NM_SR_EVENT_FINALIZED)
        {
          nm_onesided_req_reload(p_onesided);
        }
      if(event & NM_SR_EVENT_RECV_DATA)
        {
          struct nm_data_s data_header;
          nm_data_contiguous_build(&data_header, &p_onesided->request.header, sizeof(struct nm_onesided_header_s));
          int rc = nm_sr_recv_peek(p_onesided->p_session, &p_onesided->request.sr_req, &data_header);
          if(rc != NM_ESUCCESS)
            {
              NM_FATAL("# nm_onesided: rc = %d in nm_sr_recv_peek()\n", rc);
            }
          const nm_len_t offset = sizeof(struct nm_onesided_header_s);
          nm_sr_recv_offset(p_onesided->p_session, &p_onesided->request.sr_req, offset);
          switch(p_onesided->request.header.flags)
            {
            case NM_ONESIDED_OP_PUT:
              {
                const nm_len_t size = p_onesided->request.header.size;
                void*ptr = (void*)p_onesided->request.header.addr;
                nm_data_contiguous_build(&p_onesided->data_body, ptr, size);
                nm_onesided_data_build(&p_onesided->data, &p_onesided->request.header, &p_onesided->data_body);
                nm_sr_recv_unpack_data(p_onesided->p_session, &p_onesided->request.sr_req, &p_onesided->data);
              }
              break;
            case NM_ONESIDED_OP_GET:
              {
                const nm_len_t size = p_onesided->request.header.size;
                const void*ptr = (const void*)p_onesided->request.header.addr;
                nm_gate_t p_gate = NM_GATE_NONE;
                nm_sr_request_get_gate(pp_req, &p_gate);
                nm_tag_t tag = 0;
                nm_sr_request_get_tag(pp_req, &tag);
                nm_sr_recv_unpack_data(p_onesided->p_session, &p_onesided->request.sr_req, &data_header);
                nm_sr_request_t*p_req_ack = nm_sr_req_malloc(p_onesided->req_allocator);
                nm_sr_isend(p_onesided->p_session, p_gate, NM_ONESIDED_TAG_REPLY, ptr, size, p_req_ack);
                nm_sr_request_set_ref(p_req_ack, p_onesided);
                nm_sr_request_monitor(p_onesided->p_session, p_req_ack, NM_SR_EVENT_FINALIZED, &nm_onesided_ack_handler);
              }
              break;
            case NM_ONESIDED_OP_FENCE:
              {
                nm_sr_request_t*pp_req = p_info->req.p_request;
                nm_gate_t p_gate = NM_GATE_NONE;
                nm_sr_request_get_gate(pp_req, &p_gate);
                nm_tag_t tag = 0;
                nm_sr_request_get_tag(pp_req, &tag);
                nm_sr_recv_unpack_data(p_onesided->p_session, &p_onesided->request.sr_req, &data_header);
                nm_sr_request_t*p_req_ack = nm_sr_req_malloc(p_onesided->req_allocator);
                nm_sr_isend(p_onesided->p_session, p_gate, NM_ONESIDED_TAG_REPLY, NULL, 0, p_req_ack);
                nm_sr_request_set_ref(p_req_ack, p_onesided);
                nm_sr_request_monitor(p_onesided->p_session, p_req_ack, NM_SR_EVENT_FINALIZED, &nm_onesided_ack_handler);
              }
              break;
            default:
              fprintf(stderr, "# nm_onesided: unknown op = %d\n", p_onesided->request.header.flags);
              break;
            }
        }
    }
}

void nm_onesided_init(nm_session_t p_session, nm_onesided_t*pp_onesided)
{
#if !defined(PIOMAN_MULTITHREAD)
  NM_WARN("nm_onesided: a multi-threaded flavor of pioman is recommended for 'onesided' interface. Please rebuild with a multi-threaded flavor of pioman for background progression.");
#endif /* PIOMAN_MULTITHREAD*/
  struct nm_onesided_s*p_onesided = malloc(sizeof(struct nm_onesided_s));
  padico_string_t s_name = padico_string_new();
  padico_string_printf(s_name, "%s-onesided", nm_session_get_name(p_session));
  nm_session_open(&p_onesided->p_session, padico_string_get(s_name));
  padico_string_delete(s_name);
  p_onesided->req_allocator = nm_sr_req_allocator_new(4);
  p_onesided->running = 1;
  nm_cond_init(&p_onesided->done, 1);
  nm_onesided_req_reload(p_onesided);
  *pp_onesided = p_onesided;
}

void nm_onesided_finalize(nm_onesided_t p_onesided)
{
  nm_sr_flush(p_onesided->p_session);
  nm_atomic_dec(&p_onesided->running); /* disallow recv reloaad */
  int rc = nm_sr_rcancel(p_onesided->p_session, &p_onesided->request.sr_req);
  if(rc != NM_ESUCCESS)
    {
      /* concurrent handler detected; wait for completion */
      nm_cond_wait(&p_onesided->done, 1, p_onesided->p_session->p_core);
    }
  nm_sr_req_allocator_delete(p_onesided->req_allocator);
  nm_session_destroy(p_onesided->p_session);
  p_onesided->p_session = NULL;
  free(p_onesided);
}

/* ********************************************************* */

void nm_onesided_iput(nm_onesided_t p_onesided, struct nm_gate_s*p_gate,
                      struct nm_data_s*p_data, uintptr_t dest_addr, nm_onesided_request_t*p_req)
{
  p_req->header.addr  = dest_addr;
  p_req->header.size  = nm_data_size(p_data);
  p_req->header.flags = NM_ONESIDED_OP_PUT;
  struct nm_data_s data;
  nm_onesided_data_build(&data, &p_req->header, p_data);
  nm_sr_send_init(p_onesided->p_session, &p_req->sr_req);
  nm_sr_send_pack_data(p_onesided->p_session, &p_req->sr_req, &data);
  nm_sr_send_dest(p_onesided->p_session, &p_req->sr_req, p_gate, NM_ONESIDED_TAG);
  nm_sr_send_header(p_onesided->p_session, &p_req->sr_req, sizeof(struct nm_onesided_header_s));
  nm_sr_send_submit(p_onesided->p_session, &p_req->sr_req);
}

void nm_onesided_iget(nm_onesided_t p_onesided, struct nm_gate_s*p_gate,
                      struct nm_data_s*p_data, uintptr_t dest_addr, nm_onesided_request_t*p_req)
{
  struct nm_onesided_header_s header = { .addr = dest_addr, .size = nm_data_size(p_data), .flags = NM_ONESIDED_OP_GET };
  nm_sr_send(p_onesided->p_session, p_gate, NM_ONESIDED_TAG, &header, sizeof(header));
  nm_sr_recv_init(p_onesided->p_session, &p_req->sr_req);
  nm_sr_recv_unpack_data(p_onesided->p_session, &p_req->sr_req, p_data);
  nm_sr_recv_irecv(p_onesided->p_session, &p_req->sr_req, p_gate, NM_ONESIDED_TAG_REPLY, NM_TAG_MASK_FULL);
}

void nm_onesided_ifence(nm_onesided_t p_onesided, struct nm_gate_s*p_gate, nm_onesided_request_t*p_req)
{
  struct nm_onesided_header_s header = { .addr = (uintptr_t)0, .flags = NM_ONESIDED_OP_FENCE };
  nm_sr_send(p_onesided->p_session, p_gate, NM_ONESIDED_TAG, &header, sizeof(header));
  nm_sr_irecv(p_onesided->p_session, p_gate, NM_ONESIDED_TAG_REPLY, NULL, 0, &p_req->sr_req);
}

void nm_onesided_req_wait(nm_onesided_request_t*p_req)
{
  nm_sr_request_wait(&p_req->sr_req);
}

void nm_onesided_put(nm_onesided_t p_onesided, struct nm_gate_s*p_gate, 
                      struct nm_data_s*p_data, uintptr_t dest_addr)
{
  nm_onesided_request_t req;
  nm_onesided_iput(p_onesided, p_gate, p_data, dest_addr, &req);
  nm_onesided_req_wait(&req);
}

void nm_onesided_get(nm_onesided_t p_onesided, struct nm_gate_s*p_gate,
                      struct nm_data_s*p_data, uintptr_t dest_addr)
{
  nm_onesided_request_t req;
  nm_onesided_iget(p_onesided, p_gate, p_data, dest_addr, &req);
  nm_onesided_req_wait(&req);
}

void nm_onesided_fence(nm_onesided_t p_onesided, struct nm_gate_s*p_gate)
{
  nm_onesided_request_t req;
  nm_onesided_ifence(p_onesided, p_gate, &req);
  nm_onesided_req_wait(&req);
}

