/*
 * NewMadeleine
 * Copyright (C) 2006-2021 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#include <stdint.h>
#include <sys/uio.h>
#include <assert.h>

#include <nm_private.h>

PADICO_MODULE_HOOK(NewMad_Core);

static void nm_pw_post_recv(struct nm_pkt_wrap_s*p_pw);
static void nm_pw_poll_recv(struct nm_pkt_wrap_s*p_pw);
static void nm_pw_poll_recv_any(struct nm_pkt_wrap_s*p_pw);
static void nm_pw_wait_recv_any(struct nm_pkt_wrap_s*p_pw);

/** Post a pw for recv on a driver.
 * either p_gate or p_drv may be NULL, but not both at the same time.
 */
void nm_core_post_recv(struct nm_pkt_wrap_s*p_pw, nm_gate_t p_gate,
		       nm_trk_id_t trk_id, nm_drv_t p_drv)
{
  nm_pw_assign(p_pw, trk_id, p_drv, p_gate);
  p_pw->flags |= NM_PW_RECV;
  if(p_gate)
    {
      struct nm_trk_s*p_trk = p_pw->p_trk;
      assert(p_trk != NULL);
      assert(p_trk->p_pw_recv == NULL);
      p_trk->p_pw_recv = p_pw;
    }
  else
    {
      p_pw->flags |= NM_PW_POLL_ANY;
    }
#ifdef PIOMAN
  nm_ltask_submit_pw_recv(p_pw);
#else /* PIOMAN */
  /* always put the request in the list of pending requests,
   * even if it completes immediately. It will be removed from
   * the list by nm_pw_process_complete_recv().
   */
  nm_pkt_wrap_list_push_back(&p_pw->p_drv->p_core->pending_recv_list, p_pw);
#endif /* PIOMAN */
}

/** Make progress on an active incoming request.
 */
void nm_pw_recv_progress(struct nm_pkt_wrap_s*p_pw)
{
  const nm_pw_flag_t flags = p_pw->flags;
  assert(flags & NM_PW_RECV);
  if(flags & NM_PW_POLL_ANY)
    {
      /* polling on any before posting */
      nm_pw_poll_recv_any(p_pw);
    }
  else if(flags & NM_PW_POSTED)
    {
      /* already posted- poll */
      nm_pw_poll_recv(p_pw);
    }
  else
    {
      /* not posted yet */
      nm_pw_post_recv(p_pw);
    }
}

/** Blocking wait on an active incoming request.
 */
void nm_pw_recv_wait(struct nm_pkt_wrap_s*p_pw)
{
  assert(p_pw->flags & NM_PW_RECV);
  if(p_pw->flags & NM_PW_POLL_ANY)
    {
      /* regular wait on recv_any */
      nm_pw_wait_recv_any(p_pw);
    }
  else
    {
      /* recv_any was successfull just before switch from polling to
       * blocking wait; continue busy polling */
      while(!(p_pw->flags & NM_PW_COMPLETED))
        {
          nm_pw_poll_recv(p_pw);
        }
    }

}

static void nm_pw_poll_recv_any(struct nm_pkt_wrap_s*p_pw)
{
  struct nm_core*p_core = p_pw->p_drv->p_core;
  nm_core_nolock_assert(p_core);
  assert(p_pw->p_trk == NULL);
  const struct nm_minidriver_iface_s*p_driver = p_pw->p_drv->driver;
  assert(p_driver->recv_probe_any != NULL);
  void*_status = NULL;
  int rc = (*p_driver->recv_probe_any)(p_pw->p_drv->minidriver_context, &_status);
  if(rc == NM_ESUCCESS)
    {
      assert(_status != NULL);
      p_pw->flags &= ~NM_PW_POLL_ANY;
      /* reverse resolution status -> gate */
      struct nm_trk_s*p_trk = nm_trk_hashtable_lookup(&p_core->trk_table, _status);
      assert(p_trk != NULL);
      assert(p_trk->trk_id == p_pw->trk_id);
      nm_pw_assign(p_pw, p_pw->trk_id, p_trk->p_drv, p_trk->p_gate);
      assert(p_trk->p_pw_recv == NULL);
      p_trk->p_pw_recv = p_pw;
      assert(p_pw->p_gate != NULL);
      nm_pw_post_recv(p_pw);
    }
  else if(rc == -NM_ECANCELED)
    {
      p_pw->flags |= NM_PW_CANCELED;
#ifndef PIOMAN
      nm_pkt_wrap_list_remove(&p_core->pending_recv_list, p_pw);
#endif /* !PIOMAN */
      nm_pw_completed_enqueue(p_core, p_pw);
    }
  else if(rc == -NM_ECLOSED)
    {
      p_pw->flags |= NM_PW_CLOSED;
#ifndef PIOMAN
      nm_pkt_wrap_list_remove(&p_core->pending_recv_list, p_pw);
#endif /* !PIOMAN */
      nm_pw_completed_enqueue(p_core, p_pw);
    }
}

static void nm_pw_wait_recv_any(struct nm_pkt_wrap_s*p_pw)
{
  struct nm_core*p_core = p_pw->p_drv->p_core;
  nm_core_nolock_assert(p_core);
  assert(p_pw->flags & NM_PW_POLL_ANY);
  assert(p_pw->p_trk == NULL);
  const struct nm_minidriver_iface_s*p_driver = p_pw->p_drv->driver;
  assert(p_driver->recv_wait_any != NULL);
  void*_status = NULL;
  int rc = (*p_driver->recv_wait_any)(p_pw->p_drv->minidriver_context, &_status);
  if(rc == NM_ESUCCESS)
    {
      assert(_status != NULL);
      assert(p_pw->p_drv->p_pw_recv_any == p_pw);
      p_pw->flags &= ~NM_PW_POLL_ANY;
      /* reverse resolution status -> gate */
      struct nm_trk_s*p_trk = nm_trk_hashtable_lookup(&p_core->trk_table, _status);
      assert(p_trk != NULL);
      nm_pw_assign(p_pw, p_pw->trk_id, p_trk->p_drv, p_trk->p_gate);
      assert(p_trk->p_pw_recv == NULL);
      p_trk->p_pw_recv = p_pw;
      assert(p_pw->p_gate != NULL);
      nm_pw_post_recv(p_pw);
      while(!(p_pw->flags & NM_PW_COMPLETED))
        {
          nm_pw_poll_recv(p_pw);
        }
    }
  else if((rc == -NM_ECLOSED) || (rc == -NM_ECANCELED))
    {
      if(rc == -NM_ECLOSED)
        {
          p_pw->flags |= NM_PW_CLOSED;
        }
      if(rc == -NM_ECANCELED)
        {
          p_pw->flags |= NM_PW_CANCELED;
        }
#ifndef PIOMAN
      nm_pkt_wrap_list_remove(&p_core->pending_recv_list, p_pw);
#endif /* !PIOMAN */
      nm_pw_completed_enqueue(p_core, p_pw);
    }
  else
    {
      NM_WARN("nm_pw_wait_any()- rc = %d", rc);
    }
}

/** Poll an active incoming request.
 */
static void nm_pw_poll_recv(struct nm_pkt_wrap_s*p_pw)
{
  int err = NM_ESUCCESS;
  struct nm_core*p_core = p_pw->p_drv->p_core;
  static struct timespec next_poll = { .tv_sec = 0, .tv_nsec = 0 };

  nm_core_nolock_assert(p_core);

  NM_TRACEF("polling inbound request: gate %p, drv %p, trk %d",
	    p_pw->p_gate,
	    p_pw->p_drv,
	    p_pw->trk_id);
  if(p_pw->p_drv->props.capabilities.min_period > 0)
    {
      struct timespec t;
#ifdef CLOCK_MONOTONIC_RAW
      const clockid_t clock = CLOCK_MONOTONIC_RAW;
#else
      const clockid_t clock = CLOCK_MONOTONIC;
#endif
      clock_gettime(clock, &t);
      if(t.tv_sec < next_poll.tv_sec ||
	 (t.tv_sec == next_poll.tv_sec && t.tv_nsec < next_poll.tv_nsec))
	return;
      t.tv_nsec += 100 * 1000;
      if(t.tv_nsec > 1000000000)
	{
	  t.tv_nsec -= 1000000000;
	  t.tv_sec += 1;
	}
      next_poll = t;
      NM_WARN("nm_pw_poll_recv()- non-zero min_period.\n");
    }
  assert(p_pw->p_trk != NULL);
  assert(p_pw->p_trk == &p_pw->p_gate->trks[p_pw->trk_id]);
  const struct puk_receptacle_NewMad_minidriver_s*r = &p_pw->p_trk->receptacle;
  if(p_pw->flags & NM_PW_BUF_RECV)
    {
      void*buf = NULL;
      nm_len_t len = NM_LEN_UNDEFINED;
      err = (*r->driver->recv_buf_poll)(r->_status, &buf, &len);
      if(err == NM_ESUCCESS)
        {
          assert(len != NM_LEN_UNDEFINED);
          assert(len <= p_pw->max_len);
          p_pw->v[0].iov_base = buf;
          p_pw->v[0].iov_len = len;
          p_pw->length = len;
        }
    }
  else
    {
      err = (*r->driver->recv_poll_one)(r->_status);
    }
#ifdef DEBUG
  if((err == NM_ESUCCESS) && (p_pw->p_data == NULL) &&
     (p_pw->p_drv->driver->recv_iov_post == NULL) &&
     (p_pw->p_drv->driver->recv_buf_poll == NULL) &&
     (p_pw->p_drv->props.capabilities.supports_data))
    {
      struct nm_data_s*p_data = &p_pw->p_trk->rdata;
      assert(!nm_data_isnull(p_data));
      nm_data_null_build(p_data);
    }
#endif /* DEBUG */

  if((err == NM_ESUCCESS) && (p_pw->flags & NM_PW_DYNAMIC_V0))
    {
      /* dynamic v0- data was received in an intermediate buffer */
      struct nm_req_s*p_unpack = p_pw->p_unpack;
      assert(p_unpack != NULL);
      nm_data_copy_to(&p_unpack->data, p_pw->chunk_offset, p_pw->length, p_pw->v[0].iov_base);
    }

  if((err == NM_ESUCCESS) || (err == -NM_ECLOSED) || (err == -NM_ECANCELED))
    {
      if(err == -NM_ECLOSED)
        p_pw->flags |= NM_PW_CLOSED;
      if(err == -NM_ECANCELED)
        p_pw->flags |= NM_PW_CANCELED;
#ifdef NMAD_PROFILE
      {
        /* no lock needed since we are in the drv ltask */
        struct nm_drv_s*p_drv = p_pw->p_trk->p_drv;
        p_drv->n_recv_packets++;
        p_drv->n_recv_bytes += p_pw->length;
      }
#endif /* NMAD_PROFILE */
#ifndef PIOMAN
      nm_pkt_wrap_list_remove(&p_core->pending_recv_list, p_pw);
#endif /* !PIOMAN */
      nm_pw_completed_enqueue(p_core, p_pw);
    }
  else if (err != -NM_EAGAIN)
    {
      NM_WARN("drv->poll_recv returned %d", err);
    }
}

/** Actually post a recv request to the driver
 */
static void nm_pw_post_recv(struct nm_pkt_wrap_s*p_pw)
{
  struct nm_core*p_core = p_pw->p_drv->p_core;
  /* no lock needed; only this ltask is allowed to touch the pw */
  nm_core_nolock_assert(p_core);

  assert(p_pw->p_gate);
  assert(p_pw->p_trk == &p_pw->p_gate->trks[p_pw->trk_id]);

  const struct nm_minidriver_capabilities_s*p_capabilities = &p_pw->p_trk->p_drv->props.capabilities;
  struct puk_receptacle_NewMad_minidriver_s*r = &p_pw->p_trk->receptacle;

  if(p_pw->p_data)
    {
      /* pw content is only p_data */
      assert(p_pw->v_nb == 0);
      if(p_capabilities->supports_data)
        {
          assert(r->driver->recv_data_post != NULL);
          int rc = (*r->driver->recv_data_post)(r->_status, p_pw->p_data, p_pw->chunk_offset, p_pw->length);
          if(rc != NM_ESUCCESS)
            {
              NM_WARN("drv->recv_data_post() rc = %d (%s)\n", rc, nm_strerror(rc));
            }
        }
      else
        {
          struct nm_data_properties_s props;
          nm_data_chunk_properties_compute(p_pw->p_data, p_pw->chunk_offset, p_pw->length, &props);
          const int density = props.size / props.blocks;
          if(props.is_contig)
            {
              struct iovec*v0 = nm_pw_grow_iovec(p_pw);
              v0->iov_base = nm_data_chunk_baseptr_get(p_pw->p_data, p_pw->chunk_offset, p_pw->length);
              v0->iov_len = p_pw->length;
            }
          else if((!p_capabilities->supports_iovec) || (density < NM_LARGE_MIN_DENSITY))
            {
              void*buf = malloc(p_pw->length);
	      if(buf == NULL)
		{
		  NM_FATAL("out of memory for recv.\n");
		}
              struct iovec*v0 = nm_pw_grow_iovec(p_pw);
              v0->iov_base = buf;
              v0->iov_len = p_pw->length;
              p_pw->flags |= NM_PW_DYNAMIC_V0;
            }
          else
            {
              nm_pw_data_to_iovec(p_pw);
            }
          int rc = (*r->driver->recv_iov_post)(r->_status, p_pw->v, p_pw->v_nb);
          if(rc != NM_ESUCCESS)
            {
              NM_WARN("drv->recv_iov_post() rc = %d (%s)\n", rc, nm_strerror(rc));
            }
        }
    }
  else
    {
      /* no p_data, content is iovec */
      if(p_pw->p_drv->props.capabilities.supports_buf_recv)
        {
          assert(r->driver->recv_buf_poll != NULL);
          p_pw->flags |= NM_PW_BUF_RECV;
        }
      else if(r->driver->recv_iov_post)
        {
          int rc = (*r->driver->recv_iov_post)(r->_status, p_pw->v, p_pw->v_nb);
          if(rc != NM_ESUCCESS)
            {
              NM_WARN("drv->recv_iov_post() rc = %d (%s)\n", rc, nm_strerror(rc));
            }
        }
      else
        {
          assert(r->driver->recv_data_post);
          struct nm_data_s*p_data = &p_pw->p_trk->rdata;
          assert(nm_data_isnull(p_data));
          nm_data_iov_set(p_data, (struct nm_data_iov_s){ .v = p_pw->v, .n = p_pw->v_nb });
          int rc = (*r->driver->recv_data_post)(r->_status, p_data, 0 /* chunk_offset */, p_pw->length);
          if(rc != NM_ESUCCESS)
            {
              NM_WARN("drv->recv_data_post() rc = %d (%s)\n", rc, nm_strerror(rc));
            }
        }
    }
  p_pw->flags |= NM_PW_POSTED;
  nm_pw_poll_recv(p_pw);
}

void nm_drv_refill_recv(nm_drv_t p_drv, nm_gate_t p_gate)
{
  struct nm_core*p_core = p_drv->p_core;
  nm_core_lock_assert(p_core);
  if(!p_core->enable_schedopt)
    return;
  if(p_drv->props.capabilities.has_recv_any)
    {
      /* recv any available- single pw with no gate */
      if(p_drv->p_pw_recv_any == NULL)
	{
	  struct nm_pkt_wrap_s*p_pw = nm_pw_alloc_buffer(p_core);
	  nm_core_post_recv(p_pw, NM_GATE_NONE, NM_TRK_SMALL, p_drv);
	}
    }
  else
    {
      /* no recv any- post a pw per gate */
      assert(p_gate != NULL);
      assert(p_gate->status == NM_GATE_STATUS_CONNECTED);
#ifdef DEBUG
      struct nm_trk_s*p_trk = nm_gate_trk_get(p_gate, p_drv);
      assert(p_trk != NULL && !p_trk->p_pw_recv);
      assert(p_trk->kind == nm_trk_small);
#endif /* DEBUG */
      struct nm_pkt_wrap_s*p_pw = nm_pw_alloc_buffer(p_core);
      nm_core_post_recv(p_pw, p_gate, NM_TRK_SMALL, p_drv);
    }
}
